// by Alexander Ilyin

#include "dialog_file_waves.h"
#include "wx/button.h"
#include "types.h"
#include "console.h"
#include "wx/listctrl.h"


// + base sizes +
// I'll perhaps leave 100; probably will be bigger fonts in the future
#define DIALOG_ADD_WAVES__SIZE__LIST_X (90 + SIZE_SCROLL_BAR_WIDTH)
#define DIALOG_ADD_WAVES__SIZE__LIST_Y 300
// test
//#define DIALOG_ADD_WAVES__SIZE__LIST_Y 100
#define DIALOG_ADD_WAVES__SIZE__SPACE_BETWEEN_LISTS_X 70
// - base sizes -

#define DIALOG_ADD_WAVES__SIZE__BUTTON_ADD_X 50
#define DIALOG_ADD_WAVES__SIZE__BUTTON_DELETE_X 40


BEGIN_EVENT_TABLE(DialogFileAddRemoveWaves, wxDialog)
	EVT_BUTTON(wxID_ADD, DialogFileAddRemoveWaves::OnAdd)
	EVT_BUTTON(wxID_REMOVE, DialogFileAddRemoveWaves::OnDelete)
	EVT_BUTTON(wxID_OK, DialogFileAddRemoveWaves::OnOk)
// actually this one, but seems wxListBox doesn't support it
	EVT_COMMAND_ENTER(wxID_ANY, DialogFileAddRemoveWaves::OnEnter)
//	EVT_LISTBOX(wxID_ANY, DialogFileAddRemoveWaves::OnEnter)
//	EVT_LISTBOX_DCLICK(wxID_ANY, DialogFileAddRemoveWaves::OnEnter)
//	EVT_CHOICE(wxID_ANY, DialogFileAddRemoveWaves::OnEnter)
//	EVT_LIST_ITEM_ACTIVATED(wxID_ANY, DialogFileAddRemoveWaves::OnActivated)
//	EVT_LISTBOX_ITEM_ACTIVATED(wxID_ANY, DialogFileAddRemoveWaves::OnActivated)
	EVT_CHAR(DialogFileAddRemoveWaves::OnChar)
//	EVT_KEY_DOWN(DialogFileAddRemoveWaves::OnKeyDown)
	EVT_HELP(wxID_ANY, DialogFileAddRemoveWaves::OnHelp)
END_EVENT_TABLE()


// by default the 'wxDialog' exits on 'wxID_OK' and 'wxID_CANCEL', and (seems) does nothing on 'wxID_ADD' and 'wxID_REMOVE'

DialogFileAddRemoveWaves::DialogFileAddRemoveWaves (wxWindow* window, File* file, Project* project)
:
//wxDialog (window, wxID_ANY, _T("Add waves to file"), wxDefaultPosition, wxSize (DIALOG_ADD_WAVES__SIZE__X, DIALOG_ADD_WAVES__SIZE__Y))
//wxDialog (window, wxID_ANY, _T("Add/remove file waves"), wxDefaultPosition,
//	wxSize (DIALOG_ADD_WAVES__SIZE__X + SIZE_BORDER, DIALOG_ADD_WAVES__SIZE__Y + SIZE_BORDER + SIZE_CAPTION),
//	wxDEFAULT_DIALOG_STYLE | wxWANTS_CHARS)
//DialogWithButtons (window, _T("Add waves to file"))
DialogWithButtons (window, _T("Manage file waves"), wxHORIZONTAL, "DialogFileAddRemoveWaves")
{
	int i_result;
	int size_list_x;
	int size_list_y;
	int size_space_between_lists_x;

	m_file = file;
	m_proj = project;
	m_waves_num_proj = m_proj->m_waves_num;
	m_waves_num_file = m_file->m_containing_waves_num;

// == parameters ==
	size_list_x = DIALOG_ADD_WAVES__SIZE__LIST_X;
	size_list_y = DIALOG_ADD_WAVES__SIZE__LIST_Y;
	size_space_between_lists_x = DIALOG_ADD_WAVES__SIZE__SPACE_BETWEEN_LISTS_X;

// create in order of key events handling rotating (by Tab key)

//	PosOnPanel (int x, int y);
// 'wxListCtrl' does not offers necessary style; 'wxListBox' is exactly it
// perhaps I'd better use 'wxListCtrl', which allows captions :)
	m_list_box_left = new ListBoxM (this, wxID_ANY, wxDefaultPosition, wxSize (size_list_x, size_list_y),
// 'wxLB_EXTENDED' to allow fast deselection of all
// not using 'wxWANTS_CHARS', otherwise will have to handle Tab work self
//		0, 0, wxLB_NEEDED_SB | wxLB_EXTENDED | wxBORDER_SUNKEN | wxWANTS_CHARS);
		0, 0, wxLB_NEEDED_SB | wxLB_EXTENDED | wxBORDER_SUNKEN);
	m_list_box_id_left = m_list_box_left->GetId ();
	m_list_box_right = new ListBoxM (this, wxID_ANY, wxDefaultPosition, wxSize (size_list_x, size_list_y),
		0, 0, wxLB_NEEDED_SB | wxLB_EXTENDED | wxBORDER_SUNKEN);
	m_list_box_id_right = m_list_box_right->GetId ();

	wxButton* button_copy;
	wxButton* button_delete;
// use wx predefined (unhandled by wx itself) ids for that
//	pos_x = m_x_middle - DIALOG_ADD_WAVES__SIZE__BUTTON_ADD_X / 2;
//	pos_y = m_y_middle - 20 - m_defparam_size_button_y;
	button_copy = new wxButton (this, wxID_ADD, _T("--->"), wxDefaultPosition,
		wxSize (DIALOG_ADD_WAVES__SIZE__BUTTON_ADD_X, m_defparam_size_button_y));
//	pos_y = m_y_middle + 20;
	button_delete = new wxButton (this, wxID_REMOVE, _T("Delete"), wxDefaultPosition,
		wxSize (DIALOG_ADD_WAVES__SIZE__BUTTON_DELETE_X, m_defparam_size_button_y));

// + sizers +
//	wxBoxSizer* sizer_column0;
	wxBoxSizer* sizer_column0;
	wxBoxSizer* sizer_column1;
//	wxBoxSizer* sizer_column2;
	wxBoxSizer* sizer_column2;
//	int spacer_between_textctrls = DIALOG_RESAMPLE__SIZE_X_SPACE;
//	wxSizerItem* sizer_item_space0;
//	wxSizerItem* sizer_item_chkbox_auto_get_samplerate;

//	sizer_column0 = new wxBoxSizer (wxVERTICAL);
	CreateSizerWithControlAndLabel_Vertically (&sizer_column0, m_list_box_left, "Available waves");
	sizer_column1 = new wxBoxSizer (wxVERTICAL);
//	sizer_column2 = new wxBoxSizer (wxVERTICAL);
	CreateSizerWithControlAndLabel_Vertically (&sizer_column2, m_list_box_right, "Waves in file");

	sizer_column1->Add (button_copy, 0, wxALIGN_CENTRE_HORIZONTAL);
// make the sizer also wider
	sizer_column1->Add (size_space_between_lists_x, 40);
	sizer_column1->Add (button_delete, 0, wxALIGN_RIGHT);
// - sizers -

	m_sizer_contain->Add (sizer_column0, 0, wxALIGN_CENTRE_VERTICAL);
	m_sizer_contain->Add (sizer_column1, 0, wxALIGN_CENTRE_VERTICAL);
	m_sizer_contain->Add (sizer_column2, 0, wxALIGN_CENTRE_VERTICAL);

	SetSizerAndButtons ();

	ListLeft_Fill ();
	ListRight_Fill ();

//	CentreOnParent ();
// *not exits on 'wxADD' or 'wxREMOVE'
	i_result = ShowModal ();
//	if (i_result == wxID_OK)
//		List_Copy ();
// 'delete' should not be used, but 'Destroy ()' instead; but hmmm, maybe not when wxID_OK (?)
//	Destroy ();
}

void DialogFileAddRemoveWaves::ListLeft_Fill ()
{
	int i;

	for (i = 0; i < m_waves_num_proj; i++)
	{
//		char buffer[64];
		wxString str;

//		sprintf (buffer, "Wave %02d", i);
		str.Printf ("Wave %02d", i);
//		m_list_ctrl_left->InsertItem (i, buffer);
		m_list_box_left->InsertItems (1, &str, i);
// for the case, if want add in some non-growing order - put wave index along
		m_list_box_left->SetClientData (i, (void*)i);
	}
}

void DialogFileAddRemoveWaves::ListRight_Fill ()
{
	int i;

	for (i = 0; i < m_waves_num_file; i++)
	{
		Wave* wave;
		wxString str;

		wave = m_file->m_containing_waves[i];
		str.Printf ("Wave %02d", wave->m_ix);
		m_list_box_right->InsertItems (1, &str, i);
// wave index
		m_list_box_right->SetClientData (i, (void*)wave->m_ix);
	}
}

void DialogFileAddRemoveWaves::List_Add ()
{
	int i;
	int wave_ix;
	int insert_at;
	wxArrayInt selection;
	uint selected_num;
	uint items_num_in_right;

	items_num_in_right = m_list_box_right->GetCount ();
	selected_num = m_list_box_right->GetSelections (selection);
// todo :  deselect all in the right list
// todo :  add on "enter"
	if (selected_num)
		insert_at = selection[0];
	else
		insert_at = items_num_in_right;

	for (i = 0; i < m_waves_num_proj; i++)
	{
		bool selected;

		selected = m_list_box_left->IsSelected (i);
		if (selected)
		{
			wxString str;
			int waves_num;

			waves_num = m_list_box_right->GetCount ();
			if (waves_num >= PROJECT_WAVES_MAX)
			{
				Console_Warning ("can't add more waves to file, %d max possible", PROJECT_WAVES_MAX);
				break;
			}
// get wave index
			wave_ix = (int)m_list_box_left->GetClientData (i);
			str.Printf ("Wave %02d", wave_ix);
			m_list_box_right->InsertItems (1, &str, insert_at);
			m_list_box_right->SetClientData (insert_at, (void*)wave_ix);
		}
	}
}

void DialogFileAddRemoveWaves::List_Delete ()
{
	wxArrayInt selection;
	int selected_num;
//	uint items_num_in_right;
	int i;

	selected_num = m_list_box_right->GetSelections (selection);

	if (! selected_num)
		return;
// need signed
	for (i = selected_num - 1; i >= 0; i--)
	{
//		m_list_ctrl_right;
		m_list_box_right->Delete (selection[i]);
	}
}

// may be needs to be placed in 'File'
void DialogFileAddRemoveWaves::FileWaves_RemoveAll ()
{
	int i;

	for (i = m_waves_num_file - 1; i >= 0; i--)
	{
		m_file->Wave_Remove (i);
	}
}

void DialogFileAddRemoveWaves::FileWaves_AddFromList ()
{
	int i;
	int waves_num;

	waves_num = m_list_box_right->GetCount ();
	for (i = 0; i < waves_num; i++)
	{
		int wave_ix;

		wave_ix = (int)m_list_box_right->GetClientData (i);
		wxASSERT(wave_ix >= 0);
		wxASSERT(wave_ix < PROJECT_WAVES_MAX);
		m_file->Wave_Add (i, m_proj->m_waves[wave_ix]);
	}
}

void DialogFileAddRemoveWaves::OnAdd (wxCommandEvent& event)
{
	List_Add ();
}

void DialogFileAddRemoveWaves::OnDelete (wxCommandEvent& event)
{
	List_Delete ();
}

void DialogFileAddRemoveWaves::OnOk (wxCommandEvent& event)
{
	FileWaves_RemoveAll ();
	FileWaves_AddFromList ();
// allow to exit
	event.Skip ();
}

// not comes at 'wxControl::MSWOnNotify ()'
void DialogFileAddRemoveWaves::OnEnter (wxCommandEvent& event)
{
//	event.Skip ();
}

// ListBox doesn't have that
//void DialogFileAddRemoveWaves::OnActivated (wxCommandEvent& event)
void DialogFileAddRemoveWaves::OnChar (wxKeyEvent& event)
{
	if (event.GetId () != m_list_box_id_left)
		goto check_id2;
//	if (event.GetKeyCode () == WXK_RETURN)
	if (event.GetKeyCode () == WXK_INSERT)
	{
		List_Add ();
		return;
	}
	goto finish;

check_id2:
	if (event.GetId () != m_list_box_id_right)
		goto finish;
	if (event.GetKeyCode () == WXK_DELETE)
	{
		List_Delete ();
		return;
	}

finish:
	event.Skip ();
}

void DialogFileAddRemoveWaves::OnHelp (wxHelpEvent& event)
{
}

