// by Alexander Ilyin

#include "dialog_resample.h"
#include "wx/button.h"
#include "wx/valtext.h"
#include "wx/stattext.h"
#include "wx/tooltip.h"
#include "console.h"
//#include <windows.h>
#include "error.h"
#include "wx/sound.h"
#include "wx/statbox.h"
//#include "wx/sizer.h"
#include "help.h"


// must be moved out from here
wxSound snd_chord ("chord.wav");
wxSound snd_xp_ding ("Windows XP Ding.wav");

#define DIALOG_RESAMPLE__USE_PERCENTS 0

//#define DIALOG_RESAMPLE__HEIGHT_STATIC_TEXT 16
#define DIALOG_RESAMPLE__HEIGHT_CONTROL 20

//#define DIALOG_RESAMPLE__SIZE_X_CONTROLS 100
//#define DIALOG_RESAMPLE__SIZE_X_STATICTEXT 100
#define DIALOG_RESAMPLE__SIZE_X_SPACE 30
//#define DIALOG_RESAMPLE__SIZE_X (DIALOG_RESAMPLE__SIZE_X_CONTROLS + DIALOG_RESAMPLE__SIZE_X_DISTANCE)

//#define DIALOG_RESAMPLE__SIZER_H 50

#define DIALOG_RESAMPLE__BUTTON_W 100
#define DIALOG_RESAMPLE__BUTTON_H 20
#define DIALOG_RESAMPLE__BUTTON_INDENT 30

enum
{
	DLG_RESAMPLE_BUTTON_DIVIDE = wxID_HIGHEST,
	DLG_RESAMPLE_BUTTON_SIMPLIFY,
	DLG_RESAMPLE_BUTTON_APPROXIMATE,
	DLG_RESAMPLE_BUTTON_MORE_EXACT,
	DLG_RESAMPLE_CHECKBOX_AUTO_GET_RATE,
	DLG_RESAMPLE_BUTTON_GET_SAMPLERATE,
	DLG_RESAMPLE_CHECKBOX_CHANGE_RATE,
};

BEGIN_EVENT_TABLE(DialogResample, wxDialog)
	EVT_BUTTON(wxID_OK, DialogResample::OnBtnOk)
	EVT_BUTTON(DLG_RESAMPLE_BUTTON_DIVIDE, DialogResample::OnBtnDivide)
	EVT_BUTTON(DLG_RESAMPLE_BUTTON_SIMPLIFY, DialogResample::OnBtnSimplify)
	EVT_BUTTON(DLG_RESAMPLE_BUTTON_APPROXIMATE, DialogResample::OnBtnBegin)
	EVT_BUTTON(DLG_RESAMPLE_BUTTON_MORE_EXACT, DialogResample::OnBtnApproximate)
	EVT_CHECKBOX(DLG_RESAMPLE_CHECKBOX_AUTO_GET_RATE, DialogResample::OnChkboxAutoGetSamplerate)
	EVT_BUTTON(DLG_RESAMPLE_BUTTON_GET_SAMPLERATE, DialogResample::OnBtnGetSamplerate)
	EVT_CHECKBOX(DLG_RESAMPLE_CHECKBOX_CHANGE_RATE, DialogResample::OnChkboxChangeRate)
//	EVT_CHOICE(wxID_ANY, DialogResample::OnChoice)
	EVT_TEXT(wxID_ANY, DialogResample::OnText)
	EVT_HELP(wxID_ANY, DialogResample::OnHelp)
//	EVT_BUTTON(DLG_RESAMPLE_BUTTON_DIVIDE, DialogResample::OnBtnHelp)
END_EVENT_TABLE()


//DialogResample::DialogResample (wxWindow* window_parent, File* file)
DialogResample::DialogResample (wxWindow* window_parent)
:
DialogWithButtons (window_parent, _T("Resample wave(s)"), wxVERTICAL, "DialogResample")
//DialogWithButtons (window_parent, _T("Resample wave(s)"), wxDEFAULT_DIALOG_STYLE | wxWS_EX_CONTEXTHELP)
{
//	bool result;
	wxSize ctrl_size_button;

// + values +
	m_start_samplerate = 0;
	m_val_rate_out = 1;
	m_val_rate_in = 1;
	m_val_relation = 1.0;
	m_val_tolerance = 0.1;
	m_val_cycle_samples_out = 1;
	m_val_cycle_samples_in = 1;
	m_val_auto_get_samplerate = 1;
	m_val_change_samplerate = 1;
// - values -

//	result = snd_xp_ding.IsOk ();
// hardly useful with the question mark on the title bar; mostly normally works by F1
	SetExtraStyle (wxWS_EX_CONTEXTHELP);

// + sizes +
	ctrl_size_button = wxSize (DIALOG_RESAMPLE__BUTTON_W, DIALOG_RESAMPLE__BUTTON_H);
// - sizes -

// + text input/output controls +
	wxBoxSizer* sizer_rate_out;
	wxBoxSizer* sizer_rate_in;
	wxBoxSizer* sizer_relation0;
	wxBoxSizer* sizer_tolerance;
	wxBoxSizer* sizer_cycle_samples_out;
	wxBoxSizer* sizer_cycle_samples_in;
	wxBoxSizer* sizer_relation1;
	wxBoxSizer* sizer_error;

	CreateSizerWithTextAndLabel_Vertically (&sizer_rate_out, &m_textctrl_rate_out, _T("Rate out :"));
	CreateSizerWithTextAndLabel_Vertically (&sizer_rate_in, &m_textctrl_rate_in, _T("Rate in :"));
	CreateSizerWithTextAndLabel_Vertically (&sizer_relation0, &m_textctrl_relation0, _T("Relation :"));
	CreateSizerWithTextAndLabel_Vertically (&sizer_tolerance, &m_textctrl_tolerance, _T("Tolerance :"));
// *theoretically these two also must be in a static box sizer
	CreateSizerWithTextAndLabel_Vertically (&sizer_cycle_samples_out, &m_textctrl_cycle_samples_out, _T("Cycle samples out :"));
	CreateSizerWithTextAndLabel_Vertically (&sizer_cycle_samples_in, &m_textctrl_cycle_samples_in, _T("Cycle samples in :"));
	CreateSizerWithTextAndLabel_Vertically (&sizer_relation1, &m_textctrl_relation1, _T("Relation :"));
	CreateSizerWithTextAndLabel_Vertically (&sizer_error, &m_textctrl_error, _T("Error :"));
// - text input/output controls -

// + buttons +
	wxButton* button_divide;
	wxButton* button_simplify;
	wxButton* button_approximate;
	wxButton* button_more_exact;
	wxStaticBox* static_box;
	wxStaticBoxSizer* sizer_by_rates;
	wxStaticBoxSizer* sizer_by_relation;
	int buttons_border = 4;
	int buttons_indent = DIALOG_RESAMPLE__BUTTON_INDENT;
	wxBoxSizer* sizer_divide;

// *parent must remain sizer parent, as I see in examples
	button_divide = new wxButton (this, DLG_RESAMPLE_BUTTON_DIVIDE, _T("--->"), wxDefaultPosition,
		wxSize (DIALOG_RESAMPLE__SIZE_X_SPACE, DIALOG_RESAMPLE__HEIGHT_CONTROL));
	button_simplify = new wxButton (this, DLG_RESAMPLE_BUTTON_SIMPLIFY, _T("Simplify"),
		wxDefaultPosition, ctrl_size_button);
	button_approximate = new wxButton (this, DLG_RESAMPLE_BUTTON_APPROXIMATE, _T("Approximate"),
		wxDefaultPosition, ctrl_size_button);
	button_more_exact = new wxButton (this, DLG_RESAMPLE_BUTTON_MORE_EXACT, _T("More exact"),
		wxDefaultPosition, ctrl_size_button);

// *no sense to set sizes here, it will not work
	static_box = new wxStaticBox (this, wxID_ANY, _T("From rates"));
	sizer_by_rates = new wxStaticBoxSizer (static_box, wxHORIZONTAL);
	static_box = new wxStaticBox (this, wxID_ANY, _T("From relation with tolerance"));
	sizer_by_relation = new wxStaticBoxSizer (static_box, wxHORIZONTAL);

	sizer_by_rates->Add (buttons_indent, 0);
	sizer_by_rates->Add (button_simplify, 0, wxALL | wxALIGN_CENTRE_HORIZONTAL, buttons_border);
	sizer_by_rates->Add (buttons_indent, 0);
	sizer_by_relation->Add (buttons_indent, 0);
	sizer_by_relation->Add (button_approximate, 0, wxALL | wxALIGN_CENTRE_HORIZONTAL, buttons_border);
//	sizer_by_relation->Add (m_defparam_space_between_buttons_x, 0);
	sizer_by_relation->Add (buttons_indent, 0);
	sizer_by_relation->Add (buttons_indent, 0);
	sizer_by_relation->Add (button_more_exact, 0, wxALL | wxALIGN_CENTRE_HORIZONTAL, buttons_border);
	sizer_by_relation->Add (buttons_indent, 0);

// create also sizer, because have to place the button lower
	CreateSizerWithControlAndLabel_Vertically (&sizer_divide, button_divide, "", wxTOP);
// - buttons -

// + configuration controls +
	wxButton* button_get_samplerate;

	m_chkbox_auto_get_samplerate = new wxCheckBox (this, DLG_RESAMPLE_CHECKBOX_AUTO_GET_RATE, _T("Auto get samplerate"));
//		wxDefaultPosition, wxSize (120, DIALOG_RESAMPLE__HEIGHT_CONTROL));
	button_get_samplerate = new wxButton (this, DLG_RESAMPLE_BUTTON_GET_SAMPLERATE, _T("Get samplerate"),
//		PosOnPanel (DIALOG_RESAMPLE__X4 - 20, pos_y), wxSize (DIALOG_RESAMPLE__BUTTON_W, DIALOG_RESAMPLE__HEIGHT_CONTROL));
		wxDefaultPosition, ctrl_size_button);
	m_chkbox_change_samplerate = new wxCheckBox (this, DLG_RESAMPLE_CHECKBOX_CHANGE_RATE, _T("Change samplerate"));
//		PosOnPanel (DIALOG_RESAMPLE__X1 + 0, pos_y), wxSize (120, DIALOG_RESAMPLE__HEIGHT_CONTROL));
// - configuration controls -

// + sizers +
	wxBoxSizer* sizer_row0;
	wxBoxSizer* sizer_row1;
	wxBoxSizer* sizer_row2;
	wxBoxSizer* sizer_row3;
	wxBoxSizer* sizer_row4;
	int spacer_between_textctrls = DIALOG_RESAMPLE__SIZE_X_SPACE;
	wxSizerItem* sizer_item_space0;
//	wxSizerItem* sizer_item_chkbox_auto_get_samplerate;

	sizer_row0 = new wxBoxSizer (wxHORIZONTAL);
	sizer_row1 = new wxBoxSizer (wxHORIZONTAL);
	sizer_row2 = new wxBoxSizer (wxHORIZONTAL);
	sizer_row3 = new wxBoxSizer (wxHORIZONTAL);
	sizer_row4 = new wxBoxSizer (wxHORIZONTAL);

// == start settings ==
// move down a bit
	sizer_row0->Add (m_chkbox_auto_get_samplerate, 0, wxTOP, 5);
//	sizer_item_chkbox_auto_get_samplerate = sizer_row0->Add (m_chkbox_auto_get_samplerate);
	sizer_row0->Add (10, 0);
	sizer_row0->Add (button_get_samplerate);

// == values in ==
	sizer_row1->Add (sizer_rate_out);
	sizer_row1->AddSpacer (spacer_between_textctrls);
	sizer_row1->Add (sizer_rate_in);
//	sizer_row1->Add (button_divide);
	sizer_row1->Add (sizer_divide);
	sizer_row1->Add (sizer_relation0);
	sizer_row1->AddSpacer (spacer_between_textctrls);
	sizer_row1->Add (sizer_tolerance);

// == convertion buttons ==
	sizer_row2->Add (sizer_by_rates);
	sizer_row2->AddSpacer (20);
//	sizer_row2->Add (m_defparam_space_between_buttons_x, 0);
	sizer_row2->Add (sizer_by_relation);

// == values out ==
	sizer_row3->Add (sizer_cycle_samples_out);
	sizer_item_space0 = sizer_row3->AddSpacer (spacer_between_textctrls);
	sizer_row3->Add (sizer_cycle_samples_in);
	sizer_row3->AddSpacer (spacer_between_textctrls);
	sizer_row3->Add (sizer_relation1);
	sizer_row3->AddSpacer (spacer_between_textctrls);
	sizer_row3->Add (sizer_error);

// == additional configuration ==
	sizer_row4->Add (m_chkbox_change_samplerate);

//	m_sizer_contain->Add (sizer_row0, 0, wxALIGN_CENTRE_HORIZONTAL);
	m_sizer_contain->Add (sizer_row0, 0, wxALIGN_LEFT);
	m_sizer_contain->Add (sizer_row1, 0, wxALIGN_CENTRE_HORIZONTAL);
	m_sizer_contain->Add (sizer_row2, 0, wxALIGN_CENTRE_HORIZONTAL);
	m_sizer_contain->Add (sizer_row3, 0, wxALIGN_CENTRE_HORIZONTAL);
//	m_sizer_contain->Add (sizer_row4, 0, wxALIGN_CENTRE_HORIZONTAL);
	m_sizer_contain->Add (sizer_row4, 0, wxALIGN_LEFT);
// - sizers -

	SetSizerAndButtons (10);

// + offsets' correction +
// 'Move ()' will not help
//	int pos_x;
//	int pos_y;
//	int pos_x0;
	int pos_x1;
	int space;
	wxPoint pos;

//	m_textctrl_rate_out->GetPosition (&pos_x0, 0);
//	pos_x0 = pos.x;
//	space = pos_x1 - pos_x0;
	pos = sizer_row1->GetPosition ();
	m_textctrl_rate_in->GetPosition (&pos_x1, 0);
	space = pos_x1 - pos.x;
//	m_chkbox_auto_get_samplerate->GetPosition (&pos_x, &pos_y);
//	pos_x = 0;
//	pos_y += 3;
//	m_chkbox_auto_get_samplerate->Move (pos_x, pos_y);
	sizer_row0->Prepend (space, 0);
//	sizer_item_chkbox_auto_get_samplerate->SetInitSize (pos_x, pos_y);
//	button_get_samplerate->GetPosition (&pos_x, &pos_y);
//	button_get_samplerate->Move (pos_x, pos_y);
//	m_chkbox_change_samplerate->GetPosition (&pos_x, &pos_y);
//	m_chkbox_change_samplerate->Move (pos_x, pos_y);
//	Refresh ();
	pos = sizer_item_space0->GetPosition ();
	sizer_row4->Prepend (pos.x, 0);
// - offsets' correction -
}

void DialogResample::Config_Read (wxConfigBase* config)
{
// absolute
	config->SetPath (_T("/Dialogs/Resample"));
	config->Read (_T("rate_out"), &m_val_rate_out, m_val_rate_out);
	config->Read (_T("rate_in"), &m_val_rate_in, m_val_rate_in);
	config->Read (_T("relation"), &m_val_relation, m_val_relation);
	config->Read (_T("tolerance"), &m_val_tolerance, m_val_tolerance);
	config->Read (_T("cycle_samples_out"), &m_val_cycle_samples_out, m_val_cycle_samples_out);
	config->Read (_T("cycle_samples_in"), &m_val_cycle_samples_in, m_val_cycle_samples_in);
	config->Read (_T("auto_get_samplerate"), &m_val_auto_get_samplerate, m_val_auto_get_samplerate);
	config->Read (_T("change_samplerate"), &m_val_change_samplerate, m_val_change_samplerate);
	config->SetPath (_T("../../"));
}

void DialogResample::Config_Write (wxConfigBase* config)
{
	config->SetPath (_T("/Dialogs/Resample"));
	config->Write (_T("rate_out"), m_val_rate_out);
	config->Write (_T("rate_in"), m_val_rate_in);
	config->Write (_T("relation"), m_val_relation);
	config->Write (_T("tolerance"), m_val_tolerance);
	config->Write (_T("cycle_samples_out"), m_val_cycle_samples_out);
	config->Write (_T("cycle_samples_in"), m_val_cycle_samples_in);
	config->Write (_T("auto_get_samplerate"), m_val_auto_get_samplerate);
	config->Write (_T("change_samplerate"), m_val_change_samplerate);
	config->SetPath (_T("../../"));
}

int DialogResample::ShowModal ()
{
	int i_result;

	m_rate_out = m_val_rate_out;
	m_rate_in = m_val_rate_in;
	m_relation0 = m_val_relation;
	m_tolerance = m_val_tolerance;
	m_cycle_samples_out = m_val_cycle_samples_out;
	m_cycle_samples_in = m_val_cycle_samples_in;
	m_relation1 = 0.0;
	m_error = 0.0;
	if (m_val_auto_get_samplerate)
//		if (m_start_samplerate)
			m_rate_in = m_start_samplerate;

	Values_Write ();
//	Values_ValidateAndRead (0);
	Values_ValidateAndRead ();
	i_result = DialogWithButtons::ShowModal ();
	return i_result;
}

void DialogResample::Values_Write ()
{
	wxString str;

// print invalid empty string where need to show an invalid value
	str.Printf ("%u", m_rate_out);
// 'SetValue ()' requests 'OnText ()'
	m_textctrl_rate_out->ChangeValue (str);
	if (m_rate_in)
		str.Printf ("%u", m_rate_in);
	else
		str.Printf ("");
	m_textctrl_rate_in->ChangeValue (str);
	str.Printf ("%.09lf", m_relation0);
	m_textctrl_relation0->ChangeValue (str);
	str.Printf ("%.09lf", m_tolerance);
	m_textctrl_tolerance->ChangeValue (str);
	if (m_cycle_samples_out)
		str.Printf ("%u", m_cycle_samples_out);
	else
		str.Printf ("");
	m_textctrl_cycle_samples_out->ChangeValue (str);
	if (m_cycle_samples_in)
		str.Printf ("%u", m_cycle_samples_in);
	else
		str.Printf ("");
	m_textctrl_cycle_samples_in->ChangeValue (str);
//	str.Printf ("-----");
	str.Printf ("%.09lf", m_relation1);
	m_textctrl_relation1->ChangeValue (str);
	str.Printf ("%.09lf", m_error);
	m_textctrl_error->ChangeValue (str);

	m_chkbox_auto_get_samplerate->SetValue (m_val_auto_get_samplerate);
	m_chkbox_change_samplerate->SetValue (m_val_change_samplerate);
}

// validate data
void DialogResample::Values_ValidateAndRead ()
{
//	bool result;
//	bool wrong_value;
	bool is_valid_value;

	m_is_data_valid = 1;

//	wrong_value = ! CtrlReadUint (m_textctrl_rate_out, (uint*)&m_rate_out);
	is_valid_value = CtrlReadUInt (m_textctrl_rate_out, (uint*)&m_rate_out);
//	if (wrong_value)
//		m_is_data_valid = 0;
//	ShowWrongValue (m_textctrl_rate_out, wrong_value);
	SetBackgroundValidnessColour (m_textctrl_rate_out, is_valid_value);

//	wrong_value = ! CtrlReadUint (m_textctrl_rate_in, (uint*)&m_rate_in);
	is_valid_value = CtrlReadUInt (m_textctrl_rate_in, (uint*)&m_rate_in);
//	if (wrong_value)
//		m_is_data_valid = 0;
//	ShowWrongValue (m_textctrl_rate_in, wrong_value);
	SetBackgroundValidnessColour (m_textctrl_rate_in, is_valid_value);

//	wrong_value = ! CtrlReadDouble (m_textctrl_relation0, &m_relation0, 0);
	is_valid_value = CtrlReadDouble (m_textctrl_relation0, &m_relation0, 0);
//	if (wrong_value)
//		m_is_data_valid = 0;
//	ShowWrongValue (m_textctrl_relation0, wrong_value);
	SetBackgroundValidnessColour (m_textctrl_relation0, is_valid_value);

//	wrong_value = ! CtrlReadDouble (m_textctrl_tolerance, &m_tolerance, 1);
	is_valid_value = CtrlReadDouble (m_textctrl_tolerance, &m_tolerance, 1);
//	if (wrong_value)
//		m_is_data_valid = 0;
//	ShowWrongValue (m_textctrl_tolerance, wrong_value);
	SetBackgroundValidnessColour (m_textctrl_tolerance, is_valid_value);

//	wrong_value = ! CtrlReadUint (m_textctrl_cycle_samples_out, (uint*)&m_cycle_samples_out);
	is_valid_value = CtrlReadUInt (m_textctrl_cycle_samples_out, (uint*)&m_cycle_samples_out);
//	if (wrong_value)
//		m_is_data_valid = 0;
	m_is_data_valid &= is_valid_value;
//	ShowWrongValue (m_textctrl_cycle_samples_out, wrong_value);
	SetBackgroundValidnessColour (m_textctrl_cycle_samples_out, is_valid_value);

//	wrong_value = ! CtrlReadUint (m_textctrl_cycle_samples_in, (uint*)&m_cycle_samples_in);
	is_valid_value = CtrlReadUInt (m_textctrl_cycle_samples_in, (uint*)&m_cycle_samples_in);
//	if (wrong_value)
//		m_is_data_valid = 0;
	m_is_data_valid &= is_valid_value;
//	ShowWrongValue (m_textctrl_cycle_samples_in, wrong_value);
	SetBackgroundValidnessColour (m_textctrl_cycle_samples_in, is_valid_value);
}

// at once puts results
// GetRationals (), or better - GetRates ()
bool DialogResample::GetFractionNumerators (bool approximate)
{
	int numerator;
	int denominator;
	double relation_base;
	double relation;
	double precision;
	double tolerance_ratio;
	double tolerance;

	denominator = approximate ? m_cycle_samples_in : 1;
	relation_base = m_relation0;
	tolerance_ratio = approximate ? m_error : m_tolerance;
#if DIALOG_RESAMPLE__USE_PERCENTS
	tolerance = tolerance_ratio / 100.0;
	tolerance += 1.0;
#else
	tolerance = tolerance_ratio + 1.0;
#endif
	if (approximate)
		if (! tolerance_ratio)
			denominator++;
// stop on high value; such approximation looses it's sense on big values
// *10mln loop is executed for about 2 seconds on my 1.2GHZ proc.
	while (denominator < 1000000)
	{
		numerator = denominator * relation_base + 0.5;
		relation = (double)numerator / denominator;
		precision = relation / relation_base;
		if (precision < 1.0)
			precision = 1.0 / precision;
// no, allow 0 tolerance
//		if (precision < tolerance)
//		if (precision <= tolerance)
		if (tolerance_ratio ? precision < tolerance : precision <= tolerance)
			goto found;
		denominator++;
	}
	return 0;

found:
	m_cycle_samples_out = numerator;
	m_cycle_samples_in = denominator;
	m_relation1 = relation;
// back to 0 based value
	precision -= 1.0;
#if DIALOG_RESAMPLE__USE_PERCENTS
// back to percents
	precision *= 100.0;
#endif
	m_error = precision;
	return 1;
}

// +++ events +++

void DialogResample::OnText (wxCommandEvent& event)
{
// validate data
//	Values_ValidateAndRead (0);
	Values_ValidateAndRead ();
}

#if 0
void DialogResample::OnChoice (wxCommandEvent& event)
{
	int id;

	id = event.GetId ();

	event.Skip ();
}
#endif

void DialogResample::OnBtnOk (wxCommandEvent& event)
{
	if (! m_is_data_valid)
		return;
//	Values_ValidateAndRead (1);
//	Values_ValidateAndRead ();

	m_val_rate_out = m_rate_out;
	m_val_rate_in = m_rate_in;
	m_val_relation = m_relation0;
	m_val_tolerance = m_tolerance;
	m_val_cycle_samples_out = m_cycle_samples_out;
	m_val_cycle_samples_in = m_cycle_samples_in;
	m_val_auto_get_samplerate = m_chkbox_auto_get_samplerate->GetValue ();
	m_val_change_samplerate = m_chkbox_change_samplerate->GetValue ();

// allow to exit
	event.Skip ();
}

void DialogResample::OnBtnDivide (wxCommandEvent& event)
{
	double value;

	if (! m_rate_out)
		goto indicate_error;
	if (! m_rate_in)
		goto indicate_error;
	value = m_rate_out;
	value /= m_rate_in;
	m_relation0 = value;
	Values_Write ();
	Values_ValidateAndRead ();
	return;

indicate_error:
// bad, this sound sounds first time with relatively big delay; and next times with still sensitive delay
//	snd_chord.Play ();
	snd_xp_ding.Play ();
}

// perhaps move into "math.cpp"
int GCD (int a, int b)
{
	if (! b)
		return a;
	return GCD (b, a % b);
}

// input values :  'm_rate_out', 'm_rate_in'; output :  'm_relation0'
void DialogResample::OnBtnSimplify (wxCommandEvent& event)
{
	int gcd;

	if (! m_rate_out)
		goto indicate_error;
	if (! m_rate_in)
		goto indicate_error;
	gcd = GCD (m_rate_out, m_rate_in);
	m_cycle_samples_out = m_rate_out / gcd;
	m_cycle_samples_in = m_rate_in / gcd;
	Values_Write ();
	Values_ValidateAndRead ();
	return;

indicate_error:
	snd_xp_ding.Play ();
}

// input values :  'm_relation0', 'm_tolerance'
void DialogResample::OnBtnBegin (wxCommandEvent& event)
{
	bool result;

	if (! m_relation0)
		goto indicate_error;
	result = GetFractionNumerators (0);
	if (! result)
	{
		m_cycle_samples_out = 0;
		m_cycle_samples_in = 0;
	}
	Values_Write ();
	Values_ValidateAndRead ();
	if (! result)
		goto indicate_error;
	return;

indicate_error:
	snd_xp_ding.Play ();
}

// input values :  'm_cycle_samples_in', 'm_relation1', 'm_error'
void DialogResample::OnBtnApproximate (wxCommandEvent& event)
{
	bool result;

	if (! m_cycle_samples_in)
		goto indicate_error;
	if (! m_relation1)
		goto indicate_error;
//	if (! m_error)
//		goto indicate_error;
	result = GetFractionNumerators (1);
	if (! result)
	{
		m_cycle_samples_out = 0;
		m_cycle_samples_in = 0;
	}
	Values_Write ();
	Values_ValidateAndRead ();
	if (! result)
		goto indicate_error;
	return;

indicate_error:
	snd_xp_ding.Play ();
}

void DialogResample::OnChkboxAutoGetSamplerate (wxCommandEvent& event)
{
}

void DialogResample::OnBtnGetSamplerate (wxCommandEvent& event)
{
	m_rate_in = m_start_samplerate;
	Values_Write ();
	Values_ValidateAndRead ();
}

void DialogResample::OnChkboxChangeRate (wxCommandEvent& event)
{
}

void DialogResample::OnHelp (wxHelpEvent& event)
{
	Help_Display (wxString ("Dialog - resample"));
}

// --- events ---

