// by Alexander Ilyin
// waves display

#include "display.h"
#include "wx/image.h"
#include "wx/artprov.h"
#include "statusbar.h"
#include "error.h"
#include "console.h"
//#pragma warning(default:4018)
//#pragma warning(default:4244)
//#pragma warning(default:4284)
//#pragma warning(default:4786)


#define CURSOR_INDENT_X (0.25)


BEGIN_EVENT_TABLE(Display, wxScrolledWindow)
	EVT_PAINT(Display::OnPaint)
	EVT_LEFT_DOWN(Display::OnLeftDown)
	EVT_MOTION(Display::OnMotion)
	EVT_LEAVE_WINDOW(Display::OnLeaveWindow)
	EVT_CHAR(Display::OnChar)
END_EVENT_TABLE()


//    MyCanvas(wxWindow *parent) :
//       wxScrolledWindow(parent,wxID_ANY,wxDefaultPosition,wxDefaultSize,wxNO_FULL_REPAINT_ON_RESIZE) { }
// in wxWidgets "drawing" sample the frame and canvas are created parallel (not one-in-one)
//Display::Display (wxFrame* frame, Project* project)
Display::Display (wxWindow* parent, Project* project)
	:
//	wxScrolledWindow (frame, wxID_ANY, wxDefaultPosition, wxDefaultSize, wxNO_FULL_REPAINT_ON_RESIZE)
//	wxScrolledWindow (parent, wxID_ANY, wxDefaultPosition, wxDefaultSize, wxNO_FULL_REPAINT_ON_RESIZE)
//	wxScrolledWindow (parent, wxID_ANY, wxDefaultPosition, wxDefaultSize)
	wxScrolledWindow (parent, wxID_ANY, wxDefaultPosition, wxDefaultSize, wxHSCROLL | wxVSCROLL)
//	wxScrolledWindow (parent, wxID_ANY, wxDefaultPosition, wxDefaultSize,
//	wxScrolledWindow (parent, wxID_ANY, wxDefaultPosition, wxSize (400, 300),
//		wxHSCROLL | wxVSCROLL | wxNO_FULL_REPAINT_ON_RESIZE)
{
	m_window_parent = parent;
	m_proj = project;

	m_wave_height = 100;
// must be not here
	m_wave_height_half = m_wave_height / 2;
	m_indent_x = 3;
	m_indent_y = 3;
//	m_space_x = 1;
	m_space_y = 5;
	m_time_per_pixel = F1_0;
	m_view_wave_blend_mode = WAVE_VIEW__NONE;
//	m_cursor_time_pos = 0;
	m_cursor_pos_x = 0;

//	SetBackgroundColour (*wxLIGHT_GREY);
//	wxColour colour;
// + common colours +
// *hard to say why the colours below aren't set with numbers, like this one
// not black; frame around wave have black
	m_colour_background = wxColour (0x80, 0x80, 0x80);
//	m_colour_background = wxColour (0x40, 0x40, 0x40);
// frame around a wave
//	m_colour_space = db_colours.Find ("grey");
//	m_colour_space = db_colours.Find ("dark grey");
	m_colour_space = db_colours.Find ("black");
	if (! m_colour_space.Ok ())
		wxFAIL_MSG("'m_colour_space' is invalid");
	m_colour_wave_main_pen = db_colours.Find ("black");
	if (! m_colour_wave_main_pen.Ok ())
		wxFAIL_MSG("'m_colour_wave_main_pen' is invalid");
	m_colour_wave_indent = db_colours.Find ("dark grey");
	if (! m_colour_wave_indent.Ok ())
		wxFAIL_MSG("'m_colour_wave_indent' is invalid");
// not exist
//	m_colour_wave_indent_selected = db_colours.Find ("dark blue");
//	m_colour_wave_indent_selected = db_colours.Find ("dark slate blue");
//	m_colour_wave_indent_selected = db_colours.Find ("midnight blue");
//	m_colour_wave_indent_selected = db_colours.Find ("blue");
//	m_colour_wave_indent_selected = db_colours.Find ("medium blue");
//	m_colour_wave_indent_selected = db_colours.Find ("light blue");
	m_colour_wave_indent_selected = db_colours.Find ("slate blue");
	if (! m_colour_wave_indent_selected.Ok ())
		wxFAIL_MSG("'m_colour_wave_indent_selected' is invalid");
	m_colour_wave_middle_line = db_colours.Find ("white");
	if (! m_colour_wave_middle_line.Ok ())
		wxFAIL_MSG("'m_colour_wave_middle_line' is invalid");
// - common colours -
// + density view +
	m_colour_wave_background = db_colours.Find ("light grey");
	if (! m_colour_wave_background.Ok ())
		wxFAIL_MSG("'m_colour_wave_background' is invalid");
// - density view -
// + detailed view +
	m_colour_wave_background_detailed = db_colours.Find ("grey");
	if (! m_colour_wave_background_detailed.Ok ())
		wxFAIL_MSG("'m_colour_wave_background_detailed' is invalid");
//	m_colour_wave_connection_lines = db_colours.Find ("grey");
	m_colour_wave_connection_lines = db_colours.Find ("black");
	if (! m_colour_wave_connection_lines.Ok ())
		wxFAIL_MSG("'m_colour_wave_connection_lines' is invalid");
	m_colour_wave_dots_detailed = db_colours.Find ("white");
	if (! m_colour_wave_dots_detailed.Ok ())
		wxFAIL_MSG("'m_colour_wave_dots_detailed' is invalid");
// - detailed view -
//	m_colour_cursor = wxColour (0xa0, 0x80, 0x80);
	m_colour_cursor = wxColour (0xff, 0x80, 0x80, 0x20);

//	SetBackgroundColour (m_colour_space);
	SetBackgroundColour (m_colour_background);

// not intended to use any grid class, it even will not make less code

	DefineSize ();

//	ClearBackground ();
//	Refresh ();
}

void Display::Config_Read (wxConfigBase* config)
{
//	m_dialog_resample->Config_Read (config);
// absolute
	config->SetPath (_T("/Display"));
	config->Read (_T("wave_view"), &m_view_wave_blend_mode, WAVE_VIEW__MIPMAP);
}

void Display::Config_Write (wxConfigBase* config)
{
//	m_dialog_resample->Config_Write (config);
	config->SetPath (_T("/Display"));
	config->Write (_T("wave_view"), m_view_wave_blend_mode);
}

// gets position before indent
//int Display::GetWaveOffsetY (int ix, bool without_last_space)
//int Display::GetWaveOffsetY (int ix, bool after_indent)
int Display::GetWaveOffsetY (int ix)
{
	int wave_height_with_space;
	int offset_y;

	Assert(m_wave_height >= 2);
	Assert(m_wave_height_half == m_wave_height / 2);
	if (! ix)
		return 0;
	wave_height_with_space = m_wave_height + m_space_y;
	wave_height_with_space += m_indent_y * 2;
	offset_y = wave_height_with_space * ix;
//	if (without_last_space)
//		offset_y -= m_space_y;
//	if (after_indent)
//		offset_y += m_indent_y;
	return offset_y;
}

int Display::GetWaveIxByOffsetY (int pos_y)
{
	int wave_height_with_indent;
// or 'wave_repeat_period' :)
	int wave_height_with_indent_and_space;
	int pos_y_of_wave;
	int ix;
	int offset_y;

	if (pos_y < 0)
// outside
		return -1;
	wave_height_with_indent = m_wave_height + m_indent_y * 2;
	wave_height_with_indent_and_space = wave_height_with_indent + m_space_y;
	ix = pos_y / wave_height_with_indent_and_space;
	pos_y_of_wave = wave_height_with_indent_and_space * ix;
	offset_y = pos_y - pos_y_of_wave;
	if (offset_y >= wave_height_with_indent)
// without of wave rectangle; perhaps is at space
		return -1;

	return ix;
}

//int Display::GetFullWidth ()
fix64 Display::GetTimeLengthMax ()
{
//	int width;
	fix64 time_length;
	fix64 time_length_max;
	uint i;

	if (! m_proj->m_waves_num)
		return 0;
	time_length_max = 0;
	for (i = 0; i < m_proj->m_waves_num; i++)
	{
		Wave* wave = m_proj->m_waves[i];
		time_length = wave->GetTime ();
		time_length += wave->m_time_offset;
		if (time_length_max < time_length)
			time_length_max = time_length;
	}
// messed some atm
//	width = m_width;

//	return width;
	return time_length_max;
}

int Display::GetFullHeight ()
{
	int height;

	Assert(m_indent_y >= 1);
	Assert(m_space_y >= 0);
	if (! m_proj->m_waves_num)
		return 0;
	height = GetWaveOffsetY (m_proj->m_waves_num);
// subtract last space
	height -= m_space_y;

	return height;
}

// autosizing when necessary, or by request
// make resolution (zoom) by width and time length
// I imagine when this needs to be requested :  necessary after the first load of file(s), and then don't do that automatically anymore; I think don't do even if all files will be closed;
// the first is already done, but not the second.
void Display::Autosize ()
{
	int client_area_width;
	int client_area_height;
	fix64 time_length_max;

	GetClientSize (&client_area_width, &client_area_height);
	m_width = client_area_width;
	if (m_width < 100)
		m_width = 100;
// avoid crashes
	if (! m_proj->m_files_num)
		return;

	time_length_max = GetTimeLengthMax ();
//	m_time_per_pixel = F0_1;
	m_time_per_pixel = time_length_max / client_area_width;
}

// no resolution defining here, only size definition
// make width by time length and resolution (zoom; 'm_time_per_pixel')
void Display::DefineSize ()
{
//	int width;
//	int height;
	fix64 time_length_max;

	Assert(m_indent_x >= 1);

	if (! m_proj->m_waves_num)
	{
		SetVirtualSize (0, 0);
		return;
	}

	Assert(m_time_per_pixel);

//	m_width = GetFullWidth ();
	time_length_max = GetTimeLengthMax ();
	m_width = (uint)(time_length_max / m_time_per_pixel);
	m_height = GetFullHeight ();
	m_wave_width_max = m_width - m_indent_x * 2;
	Assert((signed)m_wave_width_max > 0);

#if 0
	wxSizer* sizer;
//	sizer = new wxSizer;
	sizer = new wxBoxSizer (wxVERTICAL);
	sizer->SetDimension (100, 100, 600, 500);
	SetSizer (sizer);
	SetAutoLayout (1);
#else
// no scrollbars, if not set that
	SetScrollRate (10, 10);
	SetVirtualSize (m_width, m_height);
//	SetVirtualSizeHints	(600, 500);
#endif
}

// calculate and limit
int Display::CalculateLevelPointOffset (wave_data_t level)
{
	int64 i64;
	int offset;

	i64 = level;
	i64 *= m_wave_height;
	i64 >>= WAVE_DATA_BITS;
	offset = (int)i64;
	if (offset >= m_wave_height_half)
		offset = m_wave_height_half - 1;
	if (offset < -m_wave_height_half)
		offset = -m_wave_height_half;

	return offset;
}

// draw with density more than 1 (up to 2) also here, because the other subroutine is for drawing straight vertical lines
//void Display::PaintWave_ConnectedDots (wxPaintDC* dc, wxRect& update_rect, Wave* wave, int pos_x, int pos_y)
void Display::PaintWave_ConnectedDots (wxPaintDC* dc, uint beg_x, uint end_x, Wave* wave, int pos_x, int pos_y)
{
	bool r;
	uint x;
	int y;
	uint64 sample_ix;
	fix64 sample_moment;
	int x_old;
	int y_old;
	bool ended = 0;
	int y_offset_middle;
	bool not_more_than_one_sample_per_column;

	Assert(end_x >= beg_x);
// something sane
	Assert((end_x - beg_x) < 1000000);

	if (! wave->m_samples_num)
// nothing to draw at all
		return;

	y_offset_middle = pos_y + m_wave_height_half - 1;
#if 0
// get first point offset
	sample_ix = 0;
#else
// get a sample, where to begin to draw
	x = beg_x - pos_x;
	if ((signed)x < 0)
		x = 0;
	sample_moment = x * m_time_per_pixel;
	r = wave->GetSampleIx (sample_ix, sample_moment);
//	sample_ix = 0;
// this sample may possibly appear right on the first point of the update region, and the connection lines with the previous point will not be drawn;
// correct it
	sample_ix--;
	if ((signed)sample_ix < 0)
		sample_ix = 0;
#endif

// returned 0 means that the wave has finished
// it may actually fail; the end value may appear a bit incorrect
	r = wave->GetSampleMoment (sample_ix, &sample_moment);
//	Assert(r);
	if (! r)
	{
// *compiler dependent
//		Console_ProgramError ("Display::PaintWave_ConnectedDots () :  Wave::GetSampleMoment () for sample index %I64u has failed, samples num is %I64u", sample_ix, wave->m_samples_num);
		return;
	}
	x = (int)(sample_moment / m_time_per_pixel);
	y = CalculateLevelPointOffset (wave->m_data[sample_ix]);
	x += pos_x;
//	y += pos_y;
	y = y_offset_middle - y;
//	not_more_than_one_sample_per_column = wave->m_segment_samples_num <= F1_0;
	not_more_than_one_sample_per_column = ! wave->IsDensityMoreThan1 (m_time_per_pixel);

// it is not stacked as I understood by Widget samples
// removing also isn't mandatory
	dc->SetClippingRegion (x, pos_y, m_wave_width_max, m_wave_height);
//// draw first background line, because it will not be drawn later
// draw first the background line, because it cannot be drawn later
//	dc->SetPen (m_colour_wave_background);
	dc->SetPen (m_colour_wave_background_detailed);
	dc->DrawLine (x, pos_y, x, pos_y + m_wave_height);

//// more than one pixel to draw; will be connection lines
loop:
// store previous point
	x_old = x;
	y_old = y;
// get next point offset
	sample_ix++;
	r = wave->GetSampleMoment (sample_ix, &sample_moment);
	if (r)
		y = CalculateLevelPointOffset (wave->m_data[sample_ix]);
	else
	{
		y = CalculateLevelPointOffset (0);
		ended = 1;
	}
	x = (int)(sample_moment / m_time_per_pixel);
// with the offsets in canvas
	x += pos_x;
//	y += pos_y;
	y = y_offset_middle - y;
// draw
// 1) draw background (only a currently necessary area)
// 2) draw the connection line
// 3) draw pixel in the beginning of the connection line
// last connection line will not have pixel on it's end
	dc->SetPen (m_colour_wave_background_detailed);
	dc->SetBrush (m_colour_wave_background_detailed);
// draw 1 pixel farther, to avoid of overdrawing upon already drawn connection lines
	dc->DrawRectangle (x_old + 1, pos_y, x - x_old + 1, m_wave_height);
	dc->SetPen (m_colour_wave_connection_lines);
	dc->DrawLine (x_old, y_old, x, y);
	if (not_more_than_one_sample_per_column)
	{
		dc->SetPen (m_colour_wave_dots_detailed);
		dc->DrawPoint (x_old, y_old);
	}
// maybe the canvas has been finished
	if (x >= m_wave_width_max)
		ended = 1;
// maybe the update region has been ended
	if (x >= end_x)
		ended = 1;
	if (! ended)
		goto loop;

	dc->DestroyClippingRegion ();
}

//void Display::PaintWave_Blended (wxPaintDC* dc, wxRect& update_rect, Wave* wave, int pos_x, int pos_y)
void Display::PaintWave_Blended (wxPaintDC* dc, uint beg_x, uint end_x, Wave* wave, int pos_x, int pos_y)
{
	bool r;
	uint x;
	wave_data_t level_min;
	wave_data_t level_avg;
	wave_data_t level_max;
	int y_min;
	int y_avg;
	int y_max;
//	int y_last_min;
//	int y_last_avg;
//	int y_last_max;
	uint i;
	fix64 moment;
//	int y_offset_low;
	int y_offset_middle;
//	bool not_less_than_two_samples_per_column;
//	uint beg;
//	uint end;

	Assert(end_x >= beg_x);
// something sane
	Assert((end_x - beg_x) < 1000000);

	moment = 0;
//	y_last_min = 0;
//	y_last_avg = 0;
//	y_last_max = 0;
// offset of lowest row in this wave
//	y_offset_low = y_offset + m_wave_height - 1;
// offset of middle row in this wave
	y_offset_middle = pos_y + m_wave_height_half - 1;
// test
//	dc->SetPen (*wxCYAN_PEN);
//	dc->SetBrush (*wxBLUE_BRUSH);
	dc->SetPen (m_colour_wave_background);
	dc->SetBrush (m_colour_wave_background);
// 'DrawRectangle ()' - the bottom line is displaced relatively to the top line on the shown difference between first height and the second one
// todo :  width  by necessary wave width
	dc->DrawRectangle (pos_x, pos_y, m_wave_width_max, m_wave_height);
//	dc->SetPen (*wxGREY_PEN);
//	dc->SetPen (m_colour_wave_main_pen);
	wave->SetSegmentDuration (m_time_per_pixel);
//	not_less_than_two_samples_per_column = wave->m_segment_samples_num >= F2_0;
// test
//	not_less_than_two_samples_per_column = 0;
	wave->m_view_blend_mode = m_view_wave_blend_mode;

// use x coordinate instead of drawing the whole virtual area
// note, that 'wxDC::DrawPoint ()' funcs, which call '::SetPixel ()', are using virtual/logical coordinates
//	for (i = 0; i < m_wave_width_max; i++)
//	for (x = 0; x < w; x++)
	for (x = beg_x; x < end_x; x++)
	{
// don't call 'DrawPoint ()' (which unconditionally calls 'SetPixel ()') all the time - that is very slow
//		if (i >= 300)
//			break;
//		if (i < beg)
//			continue;
//		if (i >= end)
//			break;
//		if (i < pos_x)
//			continue;
//		i = x + beg - pos_x;
//		i = x + beg;
		i = x - pos_x;
//		if (i < 0)
//			continue;
		if ((signed)i < 0)
			continue;
//		if (i < (unsigned)pos_x)
//			continue;
// actually isn't so necessary; 'wave->GetBlendForMoment ()' also checks
//		if ((unsigned)i >= m_wave_width_max)
//		if ((unsigned)i >= m_wave_width_max + pos_x)
		if (i >= m_wave_width_max)
			break;
//		if (x < pos_x)
//			continue;
		moment = m_time_per_pixel * i;
		r = wave->GetBlendForMoment (moment, level_min, level_avg, level_max);
		if (! r)
// end of wave
			break;
		y_min = CalculateLevelPointOffset (level_min);
		y_avg = CalculateLevelPointOffset (level_avg);
		y_max = CalculateLevelPointOffset (level_max);
//		x = i + pos_x;
//		x = i;
		dc->SetPen (m_colour_wave_main_pen);
// in case if less than two samples per segment - fills possible space
		dc->DrawLine (x, y_offset_middle - y_max, x, y_offset_middle - y_min + 1);
//		if (not_less_than_two_samples_per_column)
		{
// average level
// *grey; not good of course to take the value from there :)
			dc->SetPen (m_colour_wave_background_detailed);
			dc->DrawPoint (x, y_offset_middle - y_avg);
		}
// show middle line
		dc->SetPen (m_colour_wave_middle_line);
		dc->DrawPoint (x, y_offset_middle);
//		moment += m_time_per_pixel;
	}
}

//void Display::GetWaveBegin (wxDC* dc)
uint Display::GetWindowBegin (wxDC* dc)
{
#if 1
	wxCoord device_org_x;
//	dc->GetLogicalOrigin (&device_org_x, 0);
	dc->GetDeviceOrigin (&device_org_x, 0);
	device_org_x = -device_org_x;
#endif
#if 0
	uint w;
	dc->GetSize ((signed*)&w, 0);
#if 0
	wxCoord clipping_box_x;
	wxCoord clipping_box_w;
	dc->GetClippingBox (&clipping_box_x, 0, &clipping_box_w, 0);
#endif
#endif
#if 1
	uint beg;
// no, this is in scroll units, so is useless here
//	GetViewStart ((int*)&beg, 0);
// no, can't use it in 'PaintWave ()'; I think because it returns old position
// no, it isn't old; perhaps because 'GetDeviceOrigin ()' works relatively by 'm_xScrollPosition', and not directly by 'm_xScrollPosition'
	CalcUnscrolledPosition (0, 0, (int*)&beg, 0);
// lets find out how it differs
// it not differs
	Assert((unsigned)device_org_x == beg);
#endif
//	return -device_org_x;
	return device_org_x;
//	return beg;
}

// two absolutely different methods of draw; choose appropriate one of them
void Display::PaintWave (wxPaintDC* dc, wxRect& update_rect, uint wave_ix)
{
//	uint x;
	uint y;
	int y_offset;
	Wave* wave;
	bool needs_blend;
	wxColour colour;
	uint full_height;

// test
//	Console_Notify ("PaintWave (%u, %u, %u, %u)", update_rect.x, update_rect.y, update_rect.width, update_rect.height);
//	if (update_rect.width == 10)
//		__asm int 3

	Assert(wave_ix < m_proj->m_waves_num);
	Assert(m_wave_height >= 2);
	wave = m_proj->m_waves[wave_ix];
//	x = 0;
	y_offset = GetWaveOffsetY (wave_ix);
// indent
	if (m_proj->m_wave_is_selected[wave_ix])
		colour = m_colour_wave_indent_selected;
	else
		colour = m_colour_wave_indent;
// test
//	x = dc->GetLogicalFunction ();
// clear, if something like 'wxXOR' remains from a previous wave draw
	dc->SetLogicalFunction (wxCOPY);
	dc->SetPen (colour);
	dc->SetBrush (colour);
	full_height = m_wave_height + m_indent_y * 2;
// (maybe todo some with width)
//	dc->DrawRectangle (0, y_offset, m_wave_width_max + m_indent_x * 2, full_height);
	dc->DrawRectangle (0, y_offset, m_width, full_height);

// convert the update rectangle to logical coordinates
	uint beg;
	uint end;
	beg = GetWindowBegin (dc);
	beg += update_rect.x;
#if 0
//	end = beg + w;
// bad, with 'for (i =0;' loop, when scrolling left, lossing some (3, as 'pos_x' is) vertical lines; begin to draw earlier
//	beg;
//	moment = m_time_per_pixel * beg;
#endif
	end = beg + update_rect.width;
	Assert(end >= beg);
// something sane
	Assert((end - beg) < 1000000);

	y = y_offset + m_indent_y;
	needs_blend = wave->IsDensityMoreThan2 (m_time_per_pixel);
	if (needs_blend)
//		PaintWave_Blended (dc, update_rect, wave, m_indent_x, y);
		PaintWave_Blended (dc, beg, end, wave, m_indent_x, y);
	else
//		PaintWave_ConnectedDots (dc, update_rect, wave, m_indent_x, y);
		PaintWave_ConnectedDots (dc, beg, end, wave, m_indent_x, y);

	CursorDraw (dc, y_offset, full_height);
}

// *I think it will be better to not draw the cursor upon spaces, but only upon waves and their frames
//void Display::CursorDraw (wxPaintDC* dc, bool on)
void Display::CursorDraw (wxPaintDC* dc, uint pos_y, uint height)
{
	uint x;

	x = m_cursor_pos_x + m_indent_x;

//// think also about wxXOR
// hmmm, can't draw transparent, nor with pen, nor with brush
// *"drawing.cpp" uses some transparent wxColour, but I suspect it is not realized on Win
#if 1
	dc->SetBackgroundMode (wxTRANSPARENT);
// bad visible, if xor
//	dc->SetPen (m_colour_cursor);
	dc->SetPen (wxColour (0xff, 0x00, 0x00));
	dc->SetBackgroundMode (wxTRANSPARENT);
// not really what I want, but is acceptable; Win seems does not support transparent drawing of lines or rects
	dc->SetLogicalFunction (wxXOR);
	dc->DrawLine (x, pos_y, x, pos_y + height);
#else
	dc->SetPen (wxNullPen);
//	dc->SetPen (m_colour_cursor);
//	dc->SetBrush (m_colour_cursor);
	wxBrush brush;
//	brush = wxColour (0x80, 0xff, 0x80, 0x20);
	brush = wxColour (0x80, 0xff, 0x80, 0x80);
// *no fill at all
	brush.SetStyle (wxTRANSPARENT);
	dc->SetBrush (brush);
//	dc->DrawRectangle (x, y_offset, 5, full_height);
// same
	dc->DrawRoundedRectangle (x, y_offset, 10, full_height, 3);
#endif
}

// to remove cursor, it will be faster to redraw once, as a single region
// the simpliest way to move the cursor is :  mark for update the current rectangle with the cursor, and also mark for update the rectangle, where the cursor is on its new position
// this actually looks very hacky; duplicated calculation of the cursor pos; but I don't see any other way, since it is Windows have dual positions :  some functions are in virtual positions, and some are in the wnd positions
// the only one solution I see would be to move this calculation duplicate (of the 'PaintWave ()' drawn cursor) to such position, that it would be near at what it duplicates
// and the function name isn't appropriate a bit
//void Display::CursorClear (wxPaintDC* dc)
void Display::CursorClear (wxDC* dc)
{
	wxRect rect;
	uint x;
	int wnd_w;
	int wnd_h;
//	int virt_w;
//	int virt_h;
	uint beg;
	uint end;

	x = m_cursor_pos_x + m_indent_x;
	GetSize (&wnd_w, &wnd_h);
//	GetVirtualSize (&virt_w, &virt_h);
	beg = GetWindowBegin (dc);
	end = beg + wnd_w;
	if (x < beg)
// the cursor is beyond to the left
		return;
	if (x >= end)
// the cursor is beyond to the right
		return;
	x -= beg;
//	Console_Notify ("CursorClear () at %u", x);
	rect = wxRect (x, 0, 1, wnd_h);
// hmmm, takes only a client coordinates; I don't see any invalidation func in Win, which would take virtual coordinates
//	Refresh (1, &rect);
	RefreshRect (rect);
}

void Display::CursorMove_LeftRight (bool forward)
{
	wxClientDC dc (this);

// I am not sure that the device origin always originally is 0; lets assert it
	wxPoint pt = dc.GetDeviceOrigin();
	Assert(! pt.x);
	Assert(! pt.y);

	PrepareDC (dc);
// remove old cursor (at the next redraw)
	CursorClear (&dc);
	if (forward)
	{
		m_cursor_pos_x++;
//		if (m_cursor_pos_x > m_wave_width_max)
//			m_cursor_pos_x = m_wave_width_max;
	}
	else
	{
		m_cursor_pos_x--;
//		if ((signed)m_cursor_pos_x < 0)
//			m_cursor_pos_x = 0;
	}
	CursorBound ();
	UpdatePos (CURSOR_INDENT_X);
// draw new cursor (at the next redraw)
// *if the view has been scrolled, it gives wrong update pos, if not update the DC data
// reset the device scroll offset
//// no cursor at all
// 'PrepareDC ()' is relative; have to cleanup before calling it
	dc.SetDeviceOrigin (0, 0);
	PrepareDC (dc);
// not sure what is this :)
//	dc.ResetBoundingBox ();
	CursorClear (&dc);
// cusor is not redrawn, until the key is not released; redraw it after each move
	Update ();
}

void Display::CursorMove_HomeEnd (bool forward)
{
	wxClientDC dc (this);

	PrepareDC (dc);
// remove old cursor (at the next redraw)
	CursorClear (&dc);
	if (forward)
	{
		m_cursor_pos_x = m_wave_width_max;
	}
	else
	{
		m_cursor_pos_x = 0;
	}
	UpdatePos (CURSOR_INDENT_X);
// draw new cursor (at the next redraw)
	CursorClear (&dc);
}

// not much sense :)
//void Display::LogicalXToTime (fix64& time, uint x)
fix64 Display::LogicalXToTime (uint x)
{
	return x * m_time_per_pixel;
}

// hehe, no need
#if 0
uint Display::LogicalYToWave (uint y)
{
	return GetWaveIxByOffsetY (y);
}
#endif

// return 0, if the wave has ended
bool Display::TimeToSample (uint64& sample_ix, uint wave_ix, fix64 time)
{
	uint64 smpl;
	Wave* wave;

	sample_ix = 0;
#if 0
	if (wave_ix >= m_proj->m_waves_num)
		return 0;
#else
	Assert(wave_ix < m_proj->m_waves_num);
#endif
	wave = m_proj->m_waves[wave_ix];
	Assert(wave);
	smpl = time * wave->m_sample_rate;
	sample_ix = FIX64_TO_INT(smpl);
	if (sample_ix < wave->m_samples_num)
		return 1;
	return 0;
}

// *no need
#if 0
bool Display::XToSample (uint& sample_ix, uint x)
{
	fix64 time;

	sample_ix = 0;
	time = x * m_time_per_pixel;
	sample_ix = (uint)time;
	return 1;
}
#endif

void Display::GetMouseLogicalPosition (wxMouseEvent& event, uint* x, uint* y)
//void Display::GetMouseLogicalPosition (wxMouseEvent& event, uint* x, uint* y, uint* width, uint* height)
{
	wxPoint pos;

	wxClientDC dc (this);
// necessary, to get a proper scroll position
	PrepareDC (dc);
	pos = event.GetLogicalPosition (dc);
	if (x)
		*x = pos.x;
	if (y)
		*y = pos.y;

#if 0
	wxCoord w;
	wxCoord h;

	dc.GetSize (&w, &h);
	if (width)
		*width = w;
	if (height)
		*height = h;
#endif
}

void Display::ChangeTimeResolution (bool more_detailed, double coefficient)
{
	fix64 cursor_time_pos;

// the cursor is x position based, and not its time based
// to return the cursor back to its position, after the time resolution change, have to do few steps
// 1) convert x pos to time pos
// 2) change resolution
// 3) convert time pos to x pos
// *Audacity actually is doing selection by time

	cursor_time_pos = m_cursor_pos_x * m_time_per_pixel;

	if (more_detailed)
	{
#define TIME_RESOLUTION_MIN (F1_0 / (1000 * 1000 * 10))
		m_time_per_pixel = (fix64)(m_time_per_pixel / coefficient);
		if (m_time_per_pixel < TIME_RESOLUTION_MIN)
			m_time_per_pixel = TIME_RESOLUTION_MIN;
	}
	else
	{
#define TIME_RESOLUTION_MAX (F1_0 * (10 * 1000 * 1000))
		m_time_per_pixel = (fix64)(m_time_per_pixel * coefficient);
		if (m_time_per_pixel > TIME_RESOLUTION_MAX)
			m_time_per_pixel = TIME_RESOLUTION_MAX;
	}
	DefineSize ();

	CursorSetByTime (cursor_time_pos);

//	FictivlySetCursor ();
	UpdatePos (0.5);

	Refresh ();
}

void Display::CursorBound ()
{
	if ((signed)m_cursor_pos_x < 0)
		m_cursor_pos_x = 0;
	if (m_cursor_pos_x > m_wave_width_max)
		m_cursor_pos_x = m_wave_width_max;
}

void Display::CursorSetByTime (fix64 time_pos)
{
//	m_cursor_pos_x = time_pos / m_time_per_pixel;
	m_cursor_pos_x = (uint)(time_pos / m_time_per_pixel);
	CursorBound ();
}

//// move the scope so that the cursor would be in the center
// move the scope if the cursor steps upon the indent
// 'free_range' 0.0 - cursor only centered
// 'free_range' 0.5 - cursor can move between 1/4 and 3/4
// 'free_range' 1.0 - cursor should be on the screen
// 'indent' 0.0 - cursor should be on the screen
// 'indent' 0.25 - cursor can move between 1/4 and 3/4
// 'indent' 0.5 - cursor only centered
// *Audacity have no indent, and when the cursor goes out, it is moving the scope, so that the cursor appears in the center
//void Display::UpdatePos ()
//void Display::UpdatePos (double free_range)
void Display::UpdatePos (double indent)
{
//	fix64 full_time;
	uint x;
//	uint64 x;
	int w;
	uint x_min;
	uint x_max;
	uint x_indent;

//	Assert(free_range >= 0.0);
//	Assert(free_range <= 1.0);
	Assert(indent >= 0.0);
	Assert(indent <= 1.0);
	Assert((signed)m_cursor_pos_x >= 0);
	Assert(m_cursor_pos_x <= m_wave_width_max);
#if 0
	full_time = m_wave_width_max * m_time_per_pixel;
	if (m_cursor_time_pos < 0)
		m_cursor_time_pos = 0;
	if (m_cursor_time_pos > full_time)
		m_cursor_time_pos = full_time;
//	x = m_cursor_time_pos / m_time_per_pixel;
//	x = (uint)(m_cursor_time_pos / m_time_per_pixel);
//	Assert(x <= m_wave_width_max);
	GetSize (&w, 0);
	w /= 2;
	x -= w;
	if ((signed)x < 0)
		x = 0;
//// *'-' may lead to negative numbers, but not matters, since it is unsigned
	x_max = m_wave_width_max - w;
	if ((signed)x_max < 0)
		x_max = 0;
	if (x > x_max)
		x = x_max;
//	SetScrollPos (wxHORIZONTAL, x);
	Scroll (x, -1);
#endif
#if 0
	CursorBound ();
#endif
// crappy; have to create a dc, because 'CalcUnscrolledPosition ()' returns sometimes something wrong
	wxClientDC dc (this);

	PrepareDC (dc);
// get 'x_min' and 'x_max' - the allowed limits for the cursor
	GetSize (&w, 0);
	x_min = GetWindowBegin (&dc);
// nah, better correct the cursor x pos instead; comparisons will be simplier then
//	x_min -= m_indent_x;
//// *'-' may lead to negative numbers, but not matters, since it is unsigned
//	x_max = m_wave_width_max - w;
//	w /= 2;
//	free_range *= 0.5;
//	indent *= 0.5;
//	x = (int)(w * free_range);
	x_indent = (int)(w * indent);
	x_max = x_min + w;
	x_min += x_indent;
	x_max -= x_indent;
	Assert(x_min <= x_max);
#if 0
	x = m_cursor_pos_x - w;
	if ((signed)x < 0)
		x = 0;
	if ((signed)x_min < 0)
		x_min = 0;
	if (x > x_max)
		x = x_max;
#endif
	x = m_cursor_pos_x;
	x += m_indent_x;
	if (x_min > x)
// cursor is to the left
	{
		x -= x_indent;
		goto update_pos;
	}
	if (x_max <= x)
// cursor is to the right
	{
		x += x_indent;
		x -= w;
		goto update_pos;
	}
// the cursor is in range, no need to update
	return;

update_pos:
	if ((signed)x < 0)
		x = 0;
	if (x > m_wave_width_max)
		x = m_wave_width_max;
// scroll pos is not in pixels, so haven't to bump the left side at 'm_wave_width_max - w' (on the right end)

//	CursorBound ();

	int unit_x;
	GetScrollPixelsPerUnit (&unit_x, 0);
// even no need to correct
// and this correction not helps, if x = end - w; still remains a small chunk to go forward
//	x += unit_x / 2;
	x /= unit_x;
// quick hack
#if 0
	if (x >= m_wave_width_max / 2)
	{
		x /= unit_x;
		x++;
	}
	else
		x /= unit_x;
#endif
// both are in scroll units
//	SetScrollPos (wxHORIZONTAL, x);
	Scroll (x, -1);
}

// *temp
// meanwhile there is no real cursor, make a fake positioning; set the cursor to the center of the current view
//void Display::FictivlySetCursor (uint pos_x)
void Display::FictivlySetCursor ()
{
	int beg_x;
	int w;

//	GetViewStart (&beg_x, 0);
//	dc->GetDeviceOrigin (&device_org_x, 0);
	CalcUnscrolledPosition (0, 0, &beg_x, 0);
	GetSize (&w, 0);
	w /= 2;
	m_cursor_pos_x = beg_x + w;
	CursorBound ();
}

// +++++ events +++++

// *look 'wxPaintEvent' description in the WX manual, to see how to get an information about the necessary to update region
void Display::OnPaint (wxPaintEvent& event)
//void Display::OnPaint (wxPaintEvent& WXUNUSED(event))
{
	uint i;

// this string (or 'wxScrolledWindow::OnPaint ()') necessary mandatory,
// otherwise - strange effects (window close button not works; no message boxes are visible)
	wxPaintDC dc (this);
// if not this, then some areas are not redrawn after the scrolls
	PrepareDC (dc);
#if 0
    dc.DrawText (_T("Testing"), 50, 50);
	dc.SetPen (*wxRED_PEN);
	dc.SetBrush (*wxTRANSPARENT_BRUSH);
	dc.DrawRectangle (0, 0, 500, m_wave_height);
#endif
//	dc.SetBackground (*wxGREY_BRUSH);
//	dc.Clear ();
//	for (i = 0; i < m_proj->m_waves_num; i++)
//		PaintWave (&dc, i);

// iterating the update regions by WX sample
//	wxRegionIterator upd(GetUpdateRegion()); // get the update rect list
	wxRegionIterator upd;

// get the update rect list
	upd = GetUpdateRegion ();

	while (upd)
	{
		wxRect upd_rect;
		wxRect clip_rect;
		upd_rect = upd.GetRect ();
// and now have to clip the region, to not draw outside of it; this is necessary because draws may be a few per one event, and all will have access to each other areas
//		clip_rect = upd_rect;
		clip_rect.x = dc.DeviceToLogicalX (upd_rect.x);
		clip_rect.y = dc.DeviceToLogicalY (upd_rect.y);
		clip_rect.width = upd_rect.width;
		clip_rect.height = upd_rect.height;
		dc.SetClippingRegion (clip_rect);
		for (i = 0; i < m_proj->m_waves_num; i++)
			PaintWave (&dc, upd_rect, i);
		dc.DestroyClippingRegion ();
		upd++;
	}

//	wxScrolledWindow::OnPaint (event);
}

// hooks any area, including outside of app window, if right after first break at debugger breakpoint here that breakpoint will remain
// (because of sticky mouse button effect)
// first hook works always properly
void Display::OnLeftDown (wxMouseEvent& event)
{
	bool ctrl_down;
	uint wave_ix;
	uint i;
//	Wave* wave;
	bool outside_of_waves = 0;
	bool deselect_all = 0;
	bool set_select_wave = 0;
	bool toggle_select_wave = 0;
	uint x;
	uint y;

//	CaptureMouse ();
//	pos = event.GetPosition ();
	GetMouseLogicalPosition (event, &x, &y);

// set cursor
	m_cursor_pos_x = x - m_indent_x;
	CursorBound ();

	ctrl_down = event.m_controlDown;
	wave_ix = GetWaveIxByOffsetY (y);
//	if (wave_ix < 0)
//		outside_of_waves = 1;
// and the window easy may appear bigger than waves rendering needs
	if (wave_ix >= m_proj->m_waves_num)
		outside_of_waves = 1;

	if (ctrl_down)
// toggle
	{
// don't deselect
		deselect_all = 0;
		toggle_select_wave = 1;
	}
	else
// reset all, and set this one
	{
		deselect_all = 1;
		set_select_wave = 1;
	}

	if (deselect_all)
	{
		for (i = 0; i < m_proj->m_waves_num; i++)
			m_proj->m_wave_is_selected[i] = 0;
	}
//	if (wave_ix >= 0)
//		if (wave_ix < m_proj->m_waves_num)
	if (! outside_of_waves)
	{
		if (set_select_wave)
			m_proj->m_wave_is_selected[wave_ix] = 1;
		if (toggle_select_wave)
			m_proj->m_wave_is_selected[wave_ix] = ! m_proj->m_wave_is_selected[wave_ix];
	}

// initiate redraw
	Refresh ();

//finish:
// focus at the window
//	event.Skip ();
	return;
}

// == utilite ==

// fill with zeroes the left side
//bool PrintNumber_Decimal (char* buffer, uint64 value)
// separate thousands with commas
bool PrintNumber_Decimal_Punktuated (char* buffer, uint buffer_size, uint64 value, uint digits_num)
{
	uint i;
	uint j;
//	uint digits;
//	uint char_ix;
	uint chars_num;
	char c;

	Assert(digits_num);
// digits
	chars_num = digits_num;
// commas
	chars_num += (digits_num - 1) / 3;

	Assert(chars_num);
	i = chars_num;
loop_thousands:
	j = 0;
loop_tens:
	i--;
	c = (char)(value % 10);
	value /= 10;
	buffer[i] = c + '0';
	if (! i)
		goto finish;
	j++;
	if (j < 3)
		goto loop_tens;
	i--;
// this cannot be the first symbol
	Assert((signed)i > 0);
	buffer[i] = ',';
	goto loop_thousands;

finish:
	return 1;
}

#define PRINT_NUMBER_DECIMAL_PUNKTUATED(_buffer_, _value_, _digits_num_) PrintNumber_Decimal_Punktuated (_buffer_, sizeof(_buffer_), _value_, _digits_num_)

// *bad that it isn't generated on +/-, although it is, if select another wave, and then if the scrollbars appear or disappear
void Display::OnMotion (wxMouseEvent& event)
{
//	bool r;
	uint x;
	uint y;
	uint64 sample_ix;
	uint digits_num;
//	wxPoint pos;
	uint wave_ix;
	bool is_wave_ix_valid;
	bool is_x_upon_wave;
	fix64 time;
	int width;
	uint64 virtual_area_time;
	uint64 virtual_area_samples_num;
	uint t;
	uint milliseconds;
	uint seconds;
	uint minutes;
	uint hours;
	Wave* wave;

	if (! status_bar)
		return;

	is_x_upon_wave = 0;
	wave = 0;

	GetVirtualSize (&width, 0);
// how much time does represent the whole x size
	virtual_area_time = width * m_time_per_pixel;

//	x = event.GetX ();
//	XToSample (sample_ix, x);
	GetMouseLogicalPosition (event, &x, &y);
//	GetMouseLogicalPosition (event, &x, &y, &width);

//	XToSample (sample_ix, x);
	wave_ix = GetWaveIxByOffsetY (y);
	is_wave_ix_valid = wave_ix < m_proj->m_waves_num;
	time = LogicalXToTime (x);
	if (! is_wave_ix_valid)
	{
		status_bar->SetText (1, "----");
		goto show_time;
	}
	wave = m_proj->m_waves[wave_ix];

	is_x_upon_wave = TimeToSample (sample_ix, wave_ix, time);

// test
#if 0
	int x1;
	int y1;
	bool digits4_x;
	bool digits4_y;

	GetClientSize (&x1, &y1);
	digits4_x = x1 > 1000;
	digits4_y = y1 > 1000;
#if 0
	x = event.GetX ();
	y = event.GetY ();
#endif
	status_bar->SetText (0, "x: %0*u, y: %0*u", digits4_x + 3, x, digits4_y + 3, y);
//	digits_num = 8;
	status_bar->SetText (1, "%0*u", digits_num, sample_ix);
// work
#endif

// each wave may have own samplerate, so have to get it, and by it define the possible sample num
	Assert(wave);
//	virtual_area_samples_num = virtual_area_time * wave->m_sample_rate;
//	virtual_area_samples_num = fix64mul_int(virtual_area_time, wave->m_sample_rate);
	virtual_area_samples_num = (virtual_area_time >> 16) * wave->m_sample_rate >> 16;
// correct a bit, in case if the result will be with a small error
	digits_num = (uint)(log10((signed)virtual_area_samples_num) + 0.000001);
	digits_num++;
	if ((signed)digits_num < 1)
		digits_num = 1;

	static char buffer[64];
//	digits_num = 8;
	PRINT_NUMBER_DECIMAL_PUNKTUATED(buffer, sample_ix, digits_num);
// ergh, still misaligned, due to font
	status_bar->SetText (1, "%c%s", is_x_upon_wave ? ' ' : '*', buffer);

show_time:
	if (! m_proj->m_waves_num)
	{
		status_bar->SetText (0, "----");
		return;
	}
	t = FIX64_TO_INT(time);
	milliseconds = FIX64_TO_INT(FIX64_FRAC(time) * 1000);
	seconds = t % 60;
	t /= 60;
	minutes = t % 60;
	t /= 60;
	hours = t;
	status_bar->SetText (0, "%02u:%02u:%02u.%03u", hours, minutes, seconds, milliseconds);
}

void Display::OnLeaveWindow (wxMouseEvent& event)
{
	if (status_bar)
	{
		status_bar->SetText (0, "----");
		status_bar->SetText (1, "----");
	}
}

void Display::OnChar (wxKeyEvent& event)
{
// test
#if 0
	int i;
	i = event.GetRawKeyCode ();
#endif

	switch (event.GetKeyCode ())
	{
// test; works
//	case WXK_F2:
//		return;

// strange, some disgarmony is here :  not WXK_ADD, but '+' or '='
// as an alternative is offered only 'GetRawKeyCode ()'
	case WXK_ADD:
	case '=':
	case '+':
		ChangeTimeResolution (1, 2.0);
//		break;
		return;
	case WXK_SUBTRACT:
	case '-':
		ChangeTimeResolution (0, 2.0);
//		break;
		return;

#if 1
	case WXK_LEFT:
		CursorMove_LeftRight (0);
		return;
	case WXK_RIGHT:
		CursorMove_LeftRight (1);
		return;
	case WXK_HOME:
		CursorMove_HomeEnd (0);
		return;
	case WXK_END:
		CursorMove_HomeEnd (1);
		return;
#endif
	}

	event.Skip ();
}

// ----- events -----

