// by Alexander Ilyin
// errors handling (reporting, and not only)
// original belongs at "Drag Watcher" project

#include "error.h"
#ifdef WIN32
#include <windows.h>
#else
#include <stdarg.h>
#endif
//#include <stdlib.h>
#include <string.h>
#include <stdio.h>
//#include "config.h"
#include "types.h"


#define PROJECT_ERROR PROJECT_NAME" - error"

static char error_exit_message[1024] = {0};


#if ERROR_WITH_LOGFILE
#define ERROR_LOGFILE_NAME "error_log.txt"

void Error_Log_Print (const char* str)
{
#ifdef WIN32
	HANDLE h_log_file;
	uint string_length;
	BOOL B_result;
	DWORD bytes_written = 0;
	bool is_line_feed;
	bool is_carriage_return;
	bool was_carriage_return;
	char c;
	uint i;

	was_carriage_return = 0;
	string_length = strlen (str);
	h_log_file = CreateFile (ERROR_LOGFILE_NAME, GENERIC_WRITE, FILE_SHARE_READ, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
	if (h_log_file == INVALID_HANDLE_VALUE) 
		return;
	DWORD D_result = SetFilePointer (h_log_file, 0, NULL, FILE_END);
// :P
//	if (D_result == 0xffffffff)
//		Error ("Error_Log_Print () :  'SetFilePointer ()' fails");
#if 0
	B_result = WriteFile (h_log_file, str, string_length, &bytes_written, 0);
#else
// convert '\r' and '\n' endings to proper for ASCII "\r\n" 
// (yes, word "proper" here is very relative; but basic point is that ASCII doesn't have "line end" or "new line" or "next line";
// it have only a printer's "line feed" 0x0a and "carrage return" 0x0d (why have I write such crappy notice, every time when dealing with this stuff :P))
// some texts have "\r\r\n" and "\r\n\n" inplace of a single line end, and not of two, so they definitely are the bugs, and should not be treaten as a single line end,
// and will be treaten as a two line ends of an inconsistent line ended text.
	for (i = 0; i < string_length; i++)
	{
		c = str[i];
		is_line_feed = c == '\n';
		is_carriage_return = c == '\r';
// woah how complex :) ; not really complex, if remember that it let to avoid of three conditional 'WriteFile ()'
// so it is :  write "\r\n" if it is '\n' but was no '\r' or write "\r\n" if it is '\r'
		if (is_line_feed && was_carriage_return)
// already printed
			goto continue_loop;
		if (is_carriage_return || is_line_feed)
		{
			B_result = WriteFile (h_log_file, "\r", 1, &bytes_written, 0);
		}
		else
			B_result = WriteFile (h_log_file, &c, 1, &bytes_written, 0);
continue_loop:
		was_carriage_return = is_carriage_return;
	}
#endif

	const char* separator;
	separator = "-----\r\n";
	string_length = strlen (separator);
	B_result = WriteFile (h_log_file, separator, string_length, &bytes_written, 0);

	CloseHandle (h_log_file);
#endif
#ifdef __unix__
#endif
}
#endif

// no need to restrict it to 'static', and no need to declare it :)
void Error_VA (const char* format, va_list arglist)
{
//	strcpy (error_exit_message,"\nError: ");
//	va_start(arglist, format);
//	vsprintf (error_exit_message + strlen(error_exit_message), format, arglist);
	vsprintf (error_exit_message, format, arglist);
//	va_end(arglist);

	printf (error_exit_message);

#ifdef WIN32
	MessageBox (0, error_exit_message, PROJECT_ERROR, MB_OK);
#endif
#ifdef __unix__
	printf ("\a"PROJECT_ERROR" :\n%s\n", buffer_string);
#endif

// *will call all declared by 'atexit()'
	exit(1);
}

//terminates with error code 1, printing message
void Error (const char* format, ...)
{
	va_list arglist;

	va_start(arglist, format);
	Error_VA (format, arglist);
	va_end(arglist);

// *will call all declared by 'atexit()'
	exit(1);
}

// error, but not exit the process; continue
void ErrorReport (const char* format, ...)
{
	va_list arglist;

	va_start(arglist, format);
	Error_VA (format, arglist);
	va_end(arglist);
}

#if 0
void _Assert (int expr, const char* expr_text, const char* filename, int linenum)
{
	if (! (expr))
	{
		Error ("Assertion failed: %s, file %s, line %d", expr_text, filename, linenum);
	}
}
#endif

// *the return values is not used; just to satisfy a newer compilers, when doing '?:'
//void _AssertionFailed (const char* expr_text, const char* filename, int linenum)
bool _AssertionFailed (const char* expr_text, const char* filename, int linenum)
{
	Error ("Assertion failed: '%s', file \"%s\", line %d", expr_text, filename, linenum);
	return 0;
}

