#pragma once

// by Alexander Ilyin
// errors handling

#include "config.h"


#if ! defined (ERROR_WITH_LOGFILE)
#define ERROR_WITH_LOGFILE 0
#endif

#if ERROR_WITH_LOGFILE
void Error_Log_Print (const char* str);
#endif
void Error (const char* fmt, ...);
void ErrorReport (const char* format, ...);

// no !  a program should not call a function, just to check a single condition
//void _Assert (int expr, const char* expr_text, const char* filename, int linenum);
// MSVC 6.0 and 8.0 have no problem with that, but GCC 4.6.0 does a warning "... of type 'void' ..."
//void _AssertionFailed (const char* expr_text, const char* filename, int linenum);
bool _AssertionFailed (const char* expr_text, const char* filename, int linenum);

// ways to bring expression to boolean :  MSVC 6.0 SP 6 compiler compiles '!! expression' short and properly;
// 'expression || 0' compiles into much bigger code
//#define Assert(_expression_) _Assert(!! (_expression_), #_expression_, __FILE__, __LINE__)
// 'do, while' adds 'xor, test, jnz' overhead for MSVC 6.0, in the speed optimizing and nonoptimizing modes
//#define Assert(_expression_) do {if (! (_expression_)) _AssertionFailed (#_expression_, __FILE__, __LINE__);} while (0)
#define Assert(_expression_) ((_expression_) ? 0 : _AssertionFailed (#_expression_, __FILE__, __LINE__))

// *MSVC 6.0 does not compile ellipsis, nor knows '__FUNCTION__'
#if _MSC_VER >= 1300
#define Brk() Error ("Debug break: function \"%s\", file \"%s\", line %d", __FUNCTION__, __FILE__, __LINE__)
// eeeh, undefine it in "wingdi.h" - it is nonsense to use such important define for some garbage
//#undef ERROR
#define ERROR(_format_, ...) {Error (__FUNCTION__" () :  "_format_, __VA_ARGS__);}
#define REPORT_ERROR(_format_, ...) {ReportError (__FUNCTION__" () :  "_format_, __VA_ARGS__);}
#else
#define Brk() Error ("Debug break: file \"%s\", line %d", __FILE__, __LINE__)
#endif

