// by Alexander Ilyin

#include "file.h"
//#include "wx/wx.h"
#include "wx/file.h"
#include "wx/utils.h"
#include "console.h"
#include "error.h"


// read function must not reffer to size of this structure, and use hard defines instead
typedef struct
{
// Format category
	ushort format_tag;
// Number of channels
	ushort channels_num;
// Sampling rate
	ulong samples_per_sec;
// For buffer estimation
	ulong av_bytes_per_sec;
// veery strange, looks like the bps must be defined by that
// Data block size
	ushort block_align;
// ==== optional / outside of v1.0 scope ====
	ushort bits_per_sample;
// may be (and sometimes is in use)
//	ushort dummy;
} riff_fmt_t;
#define RIFF_V1_0_WAVE_FMT_SIZE (2 + 2 + 4 + 4 + 2)
//#define RIFF_V1_0_WAVE_FMT_SIZE__WITH_BPS_FIELD (sizeof(riff_fmt_t))
#define RIFF_V1_0_WAVE_FMT_SIZE__WITH_BPS_FIELD (RIFF_V1_0_WAVE_FMT_SIZE + 2)

// make write operation faster
byte File::m_file_buf[32768];


File::File ()
{
	memset (&m_info, 0, sizeof(m_info));
	m_info.format = FILE_FORMAT__NONE;
//	file_path[0] = 0;
// clears self by nonargumented constructor
//	file_path;
	m_save_name_chosen = 0;
	m_is_selected = 0;
//	m_file_tree_ctrl = 0;
//	m_containing_waves_ixs[0] = -1;
	m_containing_waves_num = 0;
	m_have_changes = 0;
	m_riff_chunk_infos_num = 0;
	m_riff_chunk_fmt = -1;
	m_riff_chunk_data = -1;

	m_bytes_per_sample = 0;
	m_sample_size = 0;

	m_waves_opened = 0;

	m_path = 0;
	m_file_buf_file = 0;
	m_file_buf_fill = 0;
}

bool File::Wave_Add (int offset, Wave* wave)
{
	int i;

	if (offset > m_containing_waves_num)
	{
		Console_Warning ("File::Wave_Add () :  offset %d is invalid, %d is limit", offset, m_containing_waves_num);
		return 0;
	}

	for (i = m_containing_waves_num - 1; i >= offset; i--)
	{
		m_containing_waves[i + 1] = m_containing_waves[i];
	}
	m_containing_waves_num++;
	m_containing_waves[offset] = wave;
	m_info.channels_num = m_containing_waves_num;

	return 1;
}

// index in file, and not the own index of the wave
bool File::Wave_Remove (int ix)
{
	int i;

	if (ix >= m_containing_waves_num)
	{
		Console_Warning ("File::Wave_Remove () :  index %d is invalid, %d is limit", ix, m_containing_waves_num);
		return 0;
	}

	for (i = ix; i < m_containing_waves_num - 1; i++)
	{
		m_containing_waves[i] = m_containing_waves[i + 1];
	}
	m_containing_waves_num--;
	m_info.channels_num = m_containing_waves_num;

	return 1;
}

void File::Waves_Selected_Delete ()
{
// remove must be done in reverse order, because the waves are stacked
	set_int::reverse_iterator i;

	for (i = m_selections.rbegin(); i != m_selections.rend(); i++)
		Wave_Remove (*i);
}

bool File::Waves_Selected_GetCommonSamplesNum (uint64& samples_num)
{
// remove must be done in reverse order, because the waves are stacked
	set_int::iterator i;
	Wave* wave;
	uint64 num;
	bool have_waves = 0;

	for (i = m_selections.begin(); i != m_selections.end(); i++)
	{
		wave = m_containing_waves[*i];
		if (have_waves)
		{
			if (num != wave->m_samples_num)
				return 0;
		}
		else
			num = wave->m_samples_num;
		have_waves = 1;
	}
	if (! have_waves)
		return 0;
	samples_num = num;
	return 1;
}

// wxFile doesn't hold file path
bool File::DefineTypeByFourCC (wxFile& file)
{
//	ulong ul[2];
	byte fourcc[4];
	int bytes_read;

	bytes_read = file.Read (fourcc, 4);
	if (bytes_read < 4)
	{
		Console_Error ("can't read fourcc from file :  \"%s\"", m_path);
		return 0;
	}
	if (fourcc[0] == 'R' && fourcc[1] == 'I' && fourcc[2] == 'F' && fourcc[3] == 'F')
		m_info.format = FILE_FORMAT__RIFF;
	else
		Console_Warning ("file fourcc has not been recognized :  \"%s\"", m_path);
	return 1;
}

// is tied at little endian processor (too much work to use wxINTXX_SWAP_ON_BE defines everywhere)
bool File::RIFF_Parse (wxFile& file)
{
	ulong riff_length;
//	ulong wave_length;
//	ulong fmt_length;
//	ulong data_length;
	int bytes_read;
	int64 file_size;
	byte form_type_id[4];
//	byte chunk_id[4];

	bytes_read = file.Read (&riff_length, 4);
	if (bytes_read < 4)
	{
		Console_Error ("can't read RIFF length from file :  \"%s\"", m_path);
		return 0;
	}
	file_size = file.Length ();
	file_size -= 8;
	if ((int64)riff_length < file_size)
		Console_Warning ("RIFF size %d is smaller than file length %I64d :  \"%s\"", riff_length, file_size, m_path);
	if ((int64)riff_length > file_size)
		Console_Warning ("RIFF size %d is greater than file length %I64d :  \"%s\"", riff_length, file_size, m_path);

// "By convention, the form-type code for registered form types contains only digits and uppercase letters."
	bytes_read = file.Read (form_type_id, 4);
	if (bytes_read < 4)
	{
		Console_Error ("can't read RIFF contained type from file :  \"%s\"", m_path);
		return 0;
	}
	if (memcmp (form_type_id, "WAVE", 4))
	{
		Console_Error ("RIFF contained type \"%.4s\" isn't \"WAVE\" in file :  \"%s\"", form_type_id, m_path);
		return 0;
	}
//	file_size -= 4;

	RIFF_DefineChunks (file, riff_length - 4, file_size - 4);

	if (m_riff_chunk_fmt < 0 || m_riff_chunk_data < 0)
	{
		if (m_riff_chunk_fmt < 0)
			Console_Error ("RIFF have no \"fmt \" chunk in file :  \"%s\"", m_path);
		if (m_riff_chunk_data < 0)
			Console_Error ("RIFF have no \"data\" chunk in file :  \"%s\"", m_path);
		return 0;
	}

	file_size -= riff_length;

	if (file_size)
//		Console_Warning ("file have more %I64d bytes of data after first RIFF chunk (not supported, if it contains some) :  \"%s\"", file_size, m_path);
	{
		if (file_size < 0)
			Console_Warning ("file have on %08I64x (%I64d) less bytes of data in the first RIFF chunk :  \"%s\"", -file_size, -file_size, m_path);
		else
			Console_Warning ("file have more %08I64x (%I64d) bytes of data after first RIFF chunk (if it contains something more there, it should be in some other chunk, instead of \"data\") :  \"%s\"", file_size, file_size, m_path);
	}

	if (! RIFF_ParseFmt (file))
		return 0;

	return 1;
}

// these are subchunks of main RIFF chunk actually
// "Programs must expect (and ignore) any unknown chunks encountered, as with all RIFF forms. However, <fmt-ck> must always occur before <wave-data>, and both of these chunks are mandatory in a WAVE file."
//bool File::RIFF_DefineChunks (wxFile& file)
void File::RIFF_DefineChunks (wxFile& file, ulong riff_chunk_size, int64 remaining_file_size)
{
	int bytes_read;
	riff_chunk_info_t chunk_info;
	ulong chunk_size;
	bool end = 0;
	int chunk_ix = 0;

loop:
	if (remaining_file_size < 8)
	{
		Console_Warning ("broken subchunk header (eof) in RIFF file :  \"%s\"", m_path);
		return;
	}
	if (riff_chunk_size < 8)
	{
		Console_Warning ("broken subchunk header (end of RIFF chunk) in RIFF file :  \"%s\"", m_path);
		return;
	}

	bytes_read = file.Read (chunk_info.id, 4);
	if (bytes_read < 4)
	{
		Console_Warning ("can't read RIFF chunk type from file :  \"%s\"", m_path);
		return;
	}
// must rtm about - what symbols the chunk name allowed to contain

	bytes_read = file.Read (&chunk_size, 4);
	if (bytes_read < 4)
	{
		Console_Warning ("can't read RIFF chunk length from file :  \"%s\"", m_path);
		return;
	}

	remaining_file_size -= 8;
	riff_chunk_size -= 8;

	if ((int64)chunk_size > remaining_file_size)
	{
		Console_Warning ("truncated subchunk \"%.4s\" data (eof) in RIFF file :  \"%s\"", chunk_info.id, m_path);
		chunk_size = remaining_file_size;
		end = 1;
	}
	if (chunk_size > riff_chunk_size)
	{
		Console_Warning ("truncated subchunk \"%.4s\" data (end of RIFF chunk) in RIFF file :  \"%s\"", chunk_info.id, m_path);
		chunk_size = riff_chunk_size;
		end = 1;
	}

	chunk_info.offset = file.Tell ();
	chunk_info.size = chunk_size;

	file.Seek (chunk_size, wxFromCurrent);

	remaining_file_size -= chunk_size;
	riff_chunk_size -= chunk_size;

	if (! riff_chunk_size)
		end = 1;

	if (! memcmp (chunk_info.id, "fmt ", 4))
		m_riff_chunk_fmt = chunk_ix;
	else if (! memcmp (chunk_info.id, "data", 4))
		m_riff_chunk_data = chunk_ix;

	m_riff_chunk_infos[chunk_ix] = chunk_info;
	chunk_ix++;

	if (! end)
		goto loop;
}

bool File::RIFF_ParseFmt (wxFile& file)
{
	int bytes_read;
	riff_chunk_info_t* chunk_fmt;
	riff_chunk_info_t* chunk_data;
//	ulong read_size;
	int read_size;
//	bool end = 0;
//	int chunk_ix = 0;
	riff_fmt_t fmt;
	bool version_above_1_0;
	int bits_per_sample;

	wxASSERT(m_riff_chunk_fmt >= 0);
	wxASSERT(m_riff_chunk_data >= 0);
	chunk_fmt = &m_riff_chunk_infos[m_riff_chunk_fmt];
	chunk_data = &m_riff_chunk_infos[m_riff_chunk_data];

// if by "Multimedia Programming Interface and Data Specification v1.0", then it have no bps field
	if (chunk_fmt->size < RIFF_V1_0_WAVE_FMT_SIZE)
	{
		Console_Error ("RIFF \"fmt \" chunk have size %d, smaller than is necessary by specification; in file :  \"%s\"", chunk_fmt->size, m_path);
		return 0;
	}
	version_above_1_0 = chunk_fmt->size > RIFF_V1_0_WAVE_FMT_SIZE;
	if (version_above_1_0)
		if (chunk_fmt->size < RIFF_V1_0_WAVE_FMT_SIZE__WITH_BPS_FIELD)
		{
			Console_Error ("RIFF \"fmt \" chunk have size %d which is incompatible with any specification; in file :  \"%s\"", chunk_fmt->size, m_path);
			return 0;
		}
	file.Seek (chunk_fmt->offset);

	read_size = version_above_1_0 ? RIFF_V1_0_WAVE_FMT_SIZE__WITH_BPS_FIELD : RIFF_V1_0_WAVE_FMT_SIZE;
	bytes_read = file.Read (&fmt, read_size);
	if (bytes_read < read_size)
	{
		Console_Error ("can't read RIFF \"fmt \" chunk from file :  \"%s\"", m_path);
		return 0;
	}

	m_info.subformat = fmt.format_tag;
	if (fmt.format_tag != 1)
	{
//		Console_Error ("RIFF format tag in \"fmt \" chunk is showing unsupported format %d in file :  \"%s\"", fmt.format_tag, m_path);
		Console_Warning ("RIFF format tag in \"fmt \" chunk is showing unsupported format %d in file :  \"%s\"", fmt.format_tag, m_path);
		return 0;
	}

	m_info.channels_num = fmt.channels_num;
	if (version_above_1_0)
//		m_info.bits_per_sample = fmt.bits_per_sample;
		bits_per_sample = fmt.bits_per_sample;
	else
// have to calculate bps by other value
//		m_info.bits_per_sample = (int)fmt.block_align * 8 / fmt.channels_num;
		bits_per_sample = (int)fmt.block_align * 8 / fmt.channels_num;
	SetBitsPerSample (bits_per_sample, 0);
	if (m_info.bits_per_sample & 7)
	{
		Console_Error ("RIFF format is showing noninteger (%d bits) number of bytes per sample in file :  \"%s\"", m_info.bits_per_sample, m_path);
		return 0;
	}
	m_info.samples_per_sec = fmt.samples_per_sec;
	m_info.samples_num = chunk_data->size / fmt.block_align;
	m_sample_size = fmt.block_align;

	return 1;
}

const char* File::GetName ()
{
	const char* path;
	m_path_str = m_file_path.GetFullPath ();
	path = m_path_str.c_str ();
	return path;
}

// the things are not so clean, as I would wish :)
//void File::UpdateSamplesNum ()
void File::UpdateParameters ()
{
	int i;
	uint64 samples_max;
	ulong samples_rate;

	samples_max = 0;
	samples_rate = 0;
	for (i = 0; i < m_containing_waves_num; i++)
	{
		Wave* wave;
		wave = m_containing_waves[i];
		if (samples_max < wave->m_samples_num)
			samples_max = wave->m_samples_num;
// any method, not matters, just let the result be equal to waves' rates, if possible
		if (samples_rate != wave->m_sample_rate)
			samples_rate = wave->m_sample_rate;
	}
	m_info.samples_num = samples_max;
	m_info.samples_per_sec = samples_rate;
}

bool File::IsWaveMy (int wave_ix)
{
	int i;

	for (i = 0; i < m_containing_waves_num; i++)
	{
		Wave* wave;
		wave = m_containing_waves[i];
		if (wave->m_ix == wave_ix)
			return 1;
	}
	return 0;
}

// parse contains
bool File::Parse ()
{
	wxFile file;
	bool parsed_ok = 1;

	m_path = GetName ();
	file.Open (m_path, wxFile::read);
	if (! file.IsOpened ())
	{
		Console_Error ("can't open file :  \"%s\"", m_path);
		return 0;
	}
	if (! DefineTypeByFourCC (file))
		return 0;
	switch (m_info.format)
	{
	case FILE_FORMAT__RIFF:
// RIFF have no header, it have chunked structure
// RIFF length handling must be here also, but now - is simplificated
		if (! RIFF_Parse (file))
			parsed_ok = 0;
		break;
	}
	m_path = 0;
	return parsed_ok;
}

#define CUT_LAST_SAMPLE_IF_INCOMPLETE 1

// do not parse contains
void File::DefineAsRaw (int bits_per_sample, int channels_num, int samples_per_sec)
{
	uint file_length;
	wxFile file;

	SetBitsPerSample (bits_per_sample, 0);
	m_path = GetName ();
	file.Open (m_path, wxFile::read);
	file_length = file.Length ();
#if CUT_LAST_SAMPLE_IF_INCOMPLETE
	m_info.samples_num = file_length / m_bytes_per_sample;
#else
	m_info.samples_num = ((file_length - 1) / m_bytes_per_sample + 1);
#endif
	m_info.format = FILE_FORMAT__RAW;
	m_info.subformat = 0;
// + default +
	m_info.channels_num = channels_num;
	m_info.samples_per_sec = samples_per_sec;
// - default -
	m_sample_size = m_bytes_per_sample * channels_num;
}

// dual buffer method
// buffer is aligned to offsets in file
#define LOAD_BUFFER_BITS 16
#define LOAD_BUFFER_SIZE (1 << LOAD_BUFFER_BITS)
#define LOAD_BUFFER_MASK (LOAD_BUFFER_SIZE - 1)
#define LOAD_BUFFER_DOUBLE_SIZE (1 << (LOAD_BUFFER_BITS + 1))
#define LOAD_BUFFER_DOUBLE_MASK (LOAD_BUFFER_DOUBLE_SIZE - 1)

bool File::Waves_Load (Wave** waves, int waves_num)
{
	wxFile file;
	riff_chunk_info_t* chunk_data;
	ulong chunk_size;
	byte* buffer;
// + what have +
	uint offset_of_buffer__begin[2];
	uint offset_of_buffer__end[2];
// - what have -
// + what necessary +
	uint offset_begin;
	uint offset_end;
	uint offset_last_byte;
	uint block_begin;
	uint block_end;
// - what necessary -
	wxFileOffset file_offset;
	uint offset_in_buffer;
// is 0 or 1 only, but it isn't a boolean, it is index
	int buffer_ix;
	uint sample_ix;
	int i;
	size_t  bytes_read;

	wxASSERT(waves_num == m_info.channels_num);
// the subroutine below can read only by pieces which are not greater than buffer size
	wxASSERT(m_sample_size);
	wxASSERT(m_sample_size <= LOAD_BUFFER_SIZE);

//open_file:
	m_path = GetName ();
	file.Open (m_path, wxFile::read);
	if (! file.IsOpened ())
	{
		Console_Error ("File::Waves_Load () :  can't open file :  \"%s\"", m_path);
		return 0;
	}

	switch (m_info.format)
	{
	case FILE_FORMAT__RAW:
//		goto open_file;
		chunk_size = file.Length ();
		break;
	case FILE_FORMAT__RIFF:
		wxASSERT(m_riff_chunk_data >= 0);
		chunk_data = &m_riff_chunk_infos[m_riff_chunk_data];
		chunk_size = chunk_data->size;
		break;
	default:
		Error ("File::Waves_Load () :  unknown format %d", m_info.format);
	}

	buffer = (byte*)malloc (LOAD_BUFFER_DOUBLE_SIZE);
	if (! buffer)
	{
		Console_Error ("File::Waves_Load () :  malloc () failed");
		return 0;
	}

	switch (m_info.format)
	{
	case FILE_FORMAT__RAW:
		offset_begin = 0;
		break;
	case FILE_FORMAT__RIFF:
		offset_begin = chunk_data->offset;
		break;
	default:
		Error ("File::Waves_Load () :  unknown format %d", m_info.format);
	}

#if 0
	file_offset = file.Seek (offset_begin);
	if (file_offset == wxInvalidOffset)
	{
		Console_Error ("File::Waves_Load () :  wxFile::Seek () failed in file :  \"%s\"", m_path);
		goto finish;
	}
#endif

// init
	offset_of_buffer__begin[0] = 0;
	offset_of_buffer__end[0] = 0;
	offset_of_buffer__begin[1] = 0;
	offset_of_buffer__end[1] = 0;
	sample_ix = 0;

#if 0
// data begins misaligned, so for that is necessary additional code
	offset_in_buffer = offset_begin & LOAD_BUFFER_DOUBLE_MASK;
	buffer_ix = offset_in_buffer >> LOAD_BUFFER_BITS;
#endif

// seek at the necessary block now
// omitting seekings before every 'file.Read ()' below, expecting, that the buffer filling will be taken from the file continuously
	block_begin = offset_begin & ~LOAD_BUFFER_MASK;
	file_offset = file.Seek (block_begin);
	if (file_offset == wxInvalidOffset)
	{
		Console_Error ("File::Waves_Load () :  wxFile::Seek () failed in file :  \"%s\"", m_path);
		goto finish;
	}

loop:
	offset_end = offset_begin + m_sample_size;
	offset_in_buffer = offset_begin & LOAD_BUFFER_DOUBLE_MASK;

// + begin +
	buffer_ix = offset_in_buffer >> LOAD_BUFFER_BITS;

	if (offset_begin < offset_of_buffer__begin[buffer_ix] || offset_begin >= offset_of_buffer__end[buffer_ix])
// this buffer doesn't has necessary data loaded
	{
		block_begin = offset_begin & ~LOAD_BUFFER_MASK;
		block_end = block_begin + LOAD_BUFFER_SIZE;
		offset_of_buffer__begin[buffer_ix] = block_begin;
		offset_of_buffer__end[buffer_ix] = block_end;
		bytes_read = file.Read (&buffer[buffer_ix << LOAD_BUFFER_BITS], LOAD_BUFFER_SIZE);
	}
// - begin -

// + end +
	offset_last_byte = offset_end - 1;
	buffer_ix = (offset_last_byte >> LOAD_BUFFER_BITS) & 1;

	if (offset_last_byte < offset_of_buffer__begin[buffer_ix] || offset_last_byte >= offset_of_buffer__end[buffer_ix])
// this buffer doesn't has necessary data loaded
	{
		block_begin = offset_begin & ~LOAD_BUFFER_MASK;
		block_end = block_begin + LOAD_BUFFER_SIZE;
		offset_of_buffer__begin[buffer_ix] = block_begin;
		offset_of_buffer__end[buffer_ix] = block_end;
		bytes_read = file.Read (&buffer[buffer_ix << LOAD_BUFFER_BITS], LOAD_BUFFER_SIZE);
	}
// - end -

	for (i = 0; i < waves_num; i++)
	{
		Wave* wave;
		wave = waves[i];
		switch (m_bytes_per_sample)
		{
		case 1:
// oops by Microsoft :  unsigned value
			wave->m_data[sample_ix] = (*(byte*)&buffer[offset_in_buffer] - 0x80) << (WAVE_DATA_BITS - 8);
			break;
		case 2:
			wave->m_data[sample_ix] = (*(short*)&buffer[offset_in_buffer]) << (WAVE_DATA_BITS - 16);
			break;
		case 4:
			wave->m_data[sample_ix] = *(int*)&buffer[offset_in_buffer];
			break;
		default:
			Console_Error ("File::Waves_Load () :  internal error - 'm_bytes_per_sample' is %u, for file :  \"%s\"", m_bytes_per_sample, m_path);
			goto finish;
		}
		offset_in_buffer += m_bytes_per_sample;
	}

	offset_begin = offset_end;
	chunk_size -= m_sample_size;
	sample_ix++;

	if (chunk_size)
	{
		if (chunk_size >= m_sample_size)
			goto loop;
		Console_Warning ("File::Waves_Load () :  %u bytes remain in data chunk in file :  \"%s\"", chunk_size, m_path);
	}

	for (i = 0; i < waves_num; i++)
	{
		Wave* wave;
		uint wave_size;
		wave = waves[i];
		wave->Reference ();
// check wave, is it filled fully
//		wave_size = wave->GetSize ();
		wave_size = wave->m_samples_num;
		if (wave_size != sample_ix)
			Console_Warning ("File::Waves_Load () :  wave size %u is not equal to samples number %u  in file :  \"%s\"", wave_size, sample_ix, m_path);
// set info for the wave
		wave->m_sample_rate = m_info.samples_per_sec;
		wave->m_time_offset = 0;
// write waves' pointers
// 'Wave_Add ()' :)
		m_containing_waves[i] = wave;
	}
	m_containing_waves_num = waves_num;

finish:
	free (buffer);

	return 1;
}

// hehe, this is not very necessary :)
void File::FileBuf_SetFile (wxFile* file)
{
//	Assert(file);
	m_file_buf_file = file;
}

//void File::FileBuffer_Write (wxFile& file)
void File::FileBuffer_Write ()
{
	m_file_buf_file->Write (m_file_buf, m_file_buf_fill);
	m_file_buf_fill = 0;
}

// currently unnecessary; and it's internal working princip is malformed
#if 0
void File::FileBuffer_Put (byte* data, uint size)
{
	uint total_size;

//	Assert(size);
	total_size = m_file_buf_fill + size;
// maybe will not fit into buffer remainder
	if (total_size > sizeof(m_file_buf))
	{
		FileBuffer_Write ();
		return;
	}
// maybe does not cost to put in the buffer at all (or does not fit)
	if (size > sizeof(m_file_buf) / 2)
	{
		if (! m_file_buf_fill)
			FileBuffer_Write ();
		m_file_buf_file->Write (data, size);
		return;
	}
// == add to buffer ==
#if 0
// fast copy
	switch (size)
	{
	case 1:
		m_file_buf[m_file_buf_fill] = *data;
		goto finish;
	case 2:
		*(ushort*)m_file_buf[m_file_buf_fill] = *(ushort*)data;
		goto finish;
	case 4:
		*(ulong*)m_file_buf[m_file_buf_fill] = *(ulong*)data;
		goto finish;
	case 8:
		*(uint64*)m_file_buf[m_file_buf_fill] = *(uint64*)data;
		goto finish;
	}
#endif
	memcpy (&m_file_buf[m_file_buf_fill], data, size);
//finish:
	m_file_buf_fill = total_size;
}
#endif

// for requirements to write a data of any size at any moment - need to write before putting a value, and not right after
void File::FileBuffer_Put_Byte (byte data)
{
	uint total_size;

	total_size = m_file_buf_fill + 1;
// maybe will not fit into buffer remainder
	if (total_size > sizeof(m_file_buf))
	{
		FileBuffer_Write ();
//		return;
	}
// add to buffer
	m_file_buf[m_file_buf_fill] = data;
//	m_file_buf_fill = total_size;
	m_file_buf_fill += 1;
}

void File::FileBuffer_Put_Word (ushort data)
{
	uint total_size;

	total_size = m_file_buf_fill + 2;
// maybe will not fit into buffer remainder
	if (total_size > sizeof(m_file_buf))
	{
		FileBuffer_Write ();
//		return;
	}
// add to buffer
	*(ushort*)&m_file_buf[m_file_buf_fill] = data;
//	m_file_buf_fill = total_size;
	m_file_buf_fill += 2;
}

void File::FileBuffer_Put_Dword (ulong data)
{
	uint total_size;

	total_size = m_file_buf_fill + 4;
// maybe will not fit into buffer remainder
	if (total_size > sizeof(m_file_buf))
	{
		FileBuffer_Write ();
//		return;
	}
// add to buffer
	*(ulong*)&m_file_buf[m_file_buf_fill] = data;
//	m_file_buf_fill = total_size;
	m_file_buf_fill += 4;
}

bool File::Save ()
{
	wxFile file;
	riff_fmt_t riff_fmt;
	uint offset_riff;
	uint offset_fmt;
	uint offset_data;
	uint sample_ix;
	int i;
//	ulong ul;
//	static byte buffer[1024];
	ulong chunk_size_riff;
	ulong chunk_size_fmt;
	ulong chunk_size_data;

	wxASSERT(m_save_name_chosen);
	wxASSERT(m_containing_waves_num == m_info.channels_num);
//	wxASSERT(m_sample_size <= sizeof(buffer));

	m_path = GetName ();
	file.Create (m_path, 1);
	if (! file.IsOpened ())
	{
		Console_Error ("File::Save () :  can't open file :  \"%s\"", m_path);
		return 0;
	}

	switch (m_info.format)
	{
	case FILE_FORMAT__RAW:
		break;

	case FILE_FORMAT__RIFF:
// just for the case, if will be more RIFF chunks (not this case)
		offset_riff = 0;
// RIFF, length, WAVE
		file.Write ("RIFF    WAVE", 12);

		m_sample_size = m_info.bits_per_sample * m_info.channels_num / 8;

		offset_fmt = file.Tell ();
// PCM
		riff_fmt.format_tag = 1;
		riff_fmt.channels_num = m_info.channels_num;
		riff_fmt.samples_per_sec = m_info.samples_per_sec;
		riff_fmt.av_bytes_per_sec = m_sample_size * m_info.samples_per_sec;
		riff_fmt.block_align = m_sample_size;
		riff_fmt.bits_per_sample = m_info.bits_per_sample;

//		file.Write ("fmt     ", 8);
		file.Write ("fmt ", 4);
		chunk_size_fmt = sizeof(riff_fmt);
		file.Write (&chunk_size_fmt, 4);
		file.Write (&riff_fmt, sizeof(riff_fmt));

		offset_data = file.Tell ();

		file.Write ("data    ", 8);

		break;

	default:
		Error ("File::Waves_Save () :  unknown format %d", m_info.format);
	}

	for (i = 0; i < m_containing_waves_num; i++)
	{
		Wave* wave;
		wave = m_containing_waves[i];
		if (wave->m_samples_num > m_info.samples_num)
			Console_Warning ("File::Save () :  wave have %I64u samples while necessary are %I64u, in file :  \"%s\"", wave->m_samples_num, m_info.samples_num, m_path);
	}

//write_data:
	sample_ix = 0;
	FileBuf_SetFile (&file);
//	m_file_buf_file = &file;
// *uses 100% cpu, but I wouldn't say that it is too slow now
loop:
	for (i = 0; i < m_containing_waves_num; i++)
	{
		Wave* wave;
		wave = m_containing_waves[i];
		if (sample_ix >= wave->m_samples_num)
		{
// do not put 64bit values, otherwise crash
//			Console_Warning ("File::Save () :  wave have only %u samples while necessary are %u, in file :  \"%s\"", (uint)wave->m_samples_num, (uint)m_info.samples_num, m_path);
			Console_Warning ("File::Save () :  wave have only %I64u samples while necessary are %I64u, in file :  \"%s\"", wave->m_samples_num, m_info.samples_num, m_path);
			goto finish_file;
		}
		switch (m_bytes_per_sample)
		{
		case 1:
// oops by Microsoft :  unsigned value; it may actually be ok, but only if would be flag, which shows that it is unsigned
//			buffer[i] = (wave->m_data[sample_ix] >> (WAVE_DATA_BITS - 8)) + 0x80;
			FileBuffer_Put_Byte ((wave->m_data[sample_ix] >> (WAVE_DATA_BITS - 8)) + 0x80);
			break;
		case 2:
//			((short*)buffer)[i] = wave->m_data[sample_ix] >> (WAVE_DATA_BITS - 16);
			FileBuffer_Put_Word (wave->m_data[sample_ix] >> (WAVE_DATA_BITS - 16));
			break;
		case 4:
//			((int*)buffer)[i] = wave->m_data[sample_ix];
			FileBuffer_Put_Dword (wave->m_data[sample_ix]);
			break;
		default:
			Console_Error ("File::Save () :  internal error - 'm_bytes_per_sample' is %u, for file :  \"%s\"", m_bytes_per_sample, m_path);
			file.Close ();
			wxRemoveFile (m_path);
			goto finish;
		}
	}
// put in the buffer, instead of direct requesting Widgets function, because very slow
//	file.Write (buffer, m_sample_size);
//	FileBuffer_Put (file, buffer, m_sample_size);
	sample_ix++;
// the 'm_info.samples_num' actually must be only for synchronous files, as MS PCMs are; but otherwise, it must use waves' lengths
	if (sample_ix < m_info.samples_num)
		goto loop;
// flush
	FileBuffer_Write ();
//	m_file_buf_file = 0;
	FileBuf_SetFile (0);

finish_file:
	switch (m_info.format)
	{
	case FILE_FORMAT__RAW:
		break;

	case FILE_FORMAT__RIFF:
// "data"
		file.Seek (offset_data + 4);
		chunk_size_data = sample_ix * m_sample_size;
		file.Write (&chunk_size_data, 4);

// "RIFF"
		file.Seek (offset_riff + 4);
		chunk_size_riff = 4 + (4 + 4 + chunk_size_fmt) + (4 + 4 + chunk_size_data);
		file.Write (&chunk_size_riff, 4);

		break;

	default:
		Error ("File::Waves_Save () :  unknown format %d", m_info.format);
	}

	Console_Notify ("file \"%s\" saved", m_path);

finish:
	return 1;
}

bool File::HaveSubformat (int format_ix)
{
	bool result;
	result = format_ix == FILE_FORMAT__RIFF;
	return result;
}

bool File::HaveSubformat ()
{
	bool result;
	result = HaveSubformat (m_info.format);
	return result;
}

bool File::HaveRIFFSamplerate (int subformat_ix)
{
	bool result = 0;

	switch (subformat_ix)
	{
	case FILE_RIFF_FORMAT_TAG__PCM1:
	case FILE_RIFF_FORMAT_TAG__PCM2:
		result = 1;
		break;
	default:
		result = 0;
	}
	return result;
}

bool File::HaveSamplerate (bool* can_be_modified, int format_ix, int subformat_ix)
{
	bool result = 0;

	switch (format_ix)
	{
	case FILE_FORMAT__RIFF:
		result = HaveRIFFSamplerate (subformat_ix);
		break;
	default:
		result = 0;
	}
	if (can_be_modified)
		*can_be_modified = result;
	return result;
}

bool File::IsSamplerateValid (int format_ix, int subformat_ix, int samplerate)
{
	if (samplerate <= 0)
		return 0;
	return 1;
}

bool File::GetDefaultSamplerate (int format_ix, int subformat_ix, uint* samplerate)
{
	*samplerate = 48000;
	return 1;
}

bool File::GetFormatName (int format_ix, const char** string)
{
	wxASSERT(string);
	if (format_ix < 1)
		return 0;
	if (format_ix >= FILE_FORMATS_NUM)
		return 0;
	*string = file_format_names[format_ix];
	return 1;
}

bool File::GetSubformatInfo (int format_ix, int subformat_ix, int* tag, const char** string)
{
	bool result = 0;

	wxASSERT(tag);
	wxASSERT(string);
	switch (format_ix)
	{
	case FILE_FORMAT__RIFF:
		if (subformat_ix < 1)
			break;
		if (subformat_ix >= FILE_FORMAT_RIFF_TAGS_NUM)
			break;
		*tag = riff_tag_strixes[subformat_ix].ix;
		*string = riff_tag_strixes[subformat_ix].str;
		result = 1;
		break;
	default:
		result = 0;
	}
	return result;
}

bool File::GetDefaultSubformat (int format_ix, int* subformat_ix)
{
	bool result = 0;

	wxASSERT(subformat_ix);
	switch (format_ix)
	{
	case FILE_FORMAT__RIFF:
		*subformat_ix = FILE_RIFF_FORMAT_TAG__PCM1;
		result = 1;
		break;
	default:
		result = 0;
	}
	return result;
}

// everything what 'Waves_Load ()' supports
bool File::GetIndexedBitsPerSampleInfo (int format_ix, int subformat_ix, int ix, int* bps)
{
	wxASSERT(bps);
	switch (format_ix)
	{
#if 0
	case FILE_FORMAT__RAW8:
		*bps = 8;
		break;
	case FILE_FORMAT__RAW16:
		*bps = 16;
		break;
	case FILE_FORMAT__RAW32:
		*bps = 32;
		break;
#endif
	case FILE_FORMAT__RAW:
	case FILE_FORMAT__RIFF:
		goto riff;
	default:
		return 0;
	}
	return 1;

riff:
	switch (ix)
	{
	case 0:
		*bps = 8;
		break;
	case 1:
		*bps = 16;
		break;
	case 2:
		*bps = 32;
		break;
	default:
		return 0;
	}
	return 1;
}

