#pragma once

#include "config.h"
//#include "wx/wx.h"
#include "wx/filename.h"
#include "types.h"
#include "wave.h"
// MSVC 6.0 SP 6 <set> have a bug, but perhaps nothing serious, and only warnings will produce
#if _MSC_VER == 1200
#pragma warning(disable:4284)
#pragma warning(disable:4786)
#endif
#include <set>
typedef std::set<int> set_int;


//typedef wxLongLong_t int64;
// :)
typedef struct
{
	int ix;
	char* str;
} strix_t;

// ! linked to strings below
enum
{
	FILE_FORMAT__NONE,
//	FILE_FORMAT__RAW8,
//	FILE_FORMAT__RAW16,
//	FILE_FORMAT__RAW32,
	FILE_FORMAT__RAW,
	FILE_FORMAT__RIFF,
	FILE_FORMATS_NUM
};
//const char* file_format_names[] =
char* const file_format_names[] =
{
	0,
//	"RAW8",
//	"RAW16",
//	"RAW32",
	"RAW",
	"RIFF",
	0
};

#define FILE_RIFF_CHUNK_INFOS_MAX 1024

typedef struct
{
	byte id[4];
	int64 offset;
	ulong size;
} riff_chunk_info_t;

enum
{
	FILE_RIFF_FORMAT_TAG__NONE,
	FILE_RIFF_FORMAT_TAG__PCM1,
	FILE_RIFF_FORMAT_TAG__PCM2,
	FILE_FORMAT_RIFF_TAGS_NUM
};
const strix_t riff_tag_strixes[] =
{
	{0, 0},
	{1, "PCM 1"},
	{2, "PCM 2"},
	{0, 0}
};

// *sorted in importance order; although all of them are important
// *the structure will be often used only partially
typedef struct
{
// container file format
	int format;
// for RIFF :  01 - PCM, 02 - PCM
	int subformat;
// "sample" here means single data element from one channel
// do put value direct; use 'SetBitsPerSample ()' instead (will need byte calculation for later use)
	ushort bits_per_sample;
//// (candidate to be "private:")
// *must not be used for processing; only for load/save
	ushort channels_num;
	ulong samples_per_sec;
	uint64 samples_num;
} file_base_information_t;


class File
{
public:
	File ();
//	~File ();
	bool Wave_Add (int offset, Wave* wave);
	bool Wave_Remove (int ix);
	void Waves_Selected_Delete ();
	bool Waves_Selected_GetCommonSamplesNum (uint64& samples_num);
	bool DefineTypeByFourCC (wxFile& file);
	bool RIFF_Parse (wxFile& file);
//	bool RIFF_DefineChunks (wxFile& file, ulong riff_chunk_size, int64 remaining_file_size);
	void RIFF_DefineChunks (wxFile& file, ulong riff_chunk_size, int64 remaining_file_size);
	bool RIFF_ParseFmt (wxFile& file);
	bool Parse ();
	void DefineAsRaw (int bits_per_sample, int channels_num, int samples_per_sec);
	bool Waves_Load (Wave** waves, int waves_num);
	bool Save ();
	const char* GetName ();
	void GotChanges ()
	{
		m_have_changes = 1;
	}
	void SetBitsPerSample (int bits_per_sample, bool report_changes = 1)
	{
		m_info.bits_per_sample = bits_per_sample;
		m_bytes_per_sample = m_info.bits_per_sample >> 3;
		if (report_changes)
			GotChanges ();
	}
	void SetSampeRate (int samplerate)
	{
		m_info.samples_per_sec = samplerate;
		GotChanges ();
	}
//	void UpdateSamplesNum ();
	void UpdateParameters ();
	bool IsWaveMy (int wave_ix);
	static bool HaveSubformat (int format_ix);
	bool HaveSubformat ();
	static bool HaveSamplerate (bool* can_be_modified, int format_ix, int subformat_ix);
	static bool IsSamplerateValid (int format_ix, int subformat_ix, int samplerate);
	static bool GetDefaultSamplerate (int format_ix, int subformat_ix, uint* samplerate);
	static bool GetFormatName (int format_ix, const char** string);
	static bool GetSubformatInfo (int format_ix, int subformat_ix, int* tag, const char** string);
	static bool GetDefaultSubformat (int format_ix, int* subformat_ix);
	static bool GetIndexedBitsPerSampleInfo (int format_ix, int subformat_ix, int ix, int* bps);

	file_base_information_t m_info;
// as for other GUIs :  same name for load and save; save name replaces load name
// but don't simply write into load-from file by Ctrl+S, if user doesn't chose that file as save-file, to avoid occasionallities
//	char file_path[FILE_PATH_BUF_SIZE];
	wxFileName m_file_path;
// + external +
	bool m_save_name_chosen;
	bool m_is_selected;
//	void* m_file_tree_ctrl;
	bool m_have_changes;
// - external -
// perhaps pointers will be better, not indexes;
// yes, pointers are better, but indexes will be necessary when displaying file contains;
// but indexes may change, and the best place for them :  right in their (in wave) classes
// use projects waves num limit also here
//	int m_containing_waves_ixs[PROJECT_WAVES_MAX];
	Wave* m_containing_waves[PROJECT_WAVES_MAX];
	int m_containing_waves_num;
// + RIFF +
	riff_chunk_info_t m_riff_chunk_infos[FILE_RIFF_CHUNK_INFOS_MAX];
	uint m_riff_chunk_infos_num;
	int m_riff_chunk_fmt;
	int m_riff_chunk_data;
// - RIFF -
// initially was in 'FileTreeCtrl', but seems much proper place for it is here
	set_int m_selections;
// this is only if the file format requires, that all waves would be the same type
// + waves info +
	ushort m_bytes_per_sample;
	ushort m_sample_size;
// - waves info -
// + just storage +
// the better place for this is here (same as for wave ix)
	int m_ix;
	int m_waves_opened;
// - just storage -

private:
	static bool HaveRIFFSamplerate (int subformat_ix);
	void FileBuf_SetFile (wxFile* file);
//	void FileBuffer_Write (wxFile& file);
//	void FileBuffer_Put (wxFile& file, byte* data, uint size);
//	void FileBuffer_Put_Byte (wxFile& file, byte data);
//	void FileBuffer_Put_Word (wxFile& file, ushort data);
//	void FileBuffer_Put_Dword (wxFile& file, ulong data);
	void FileBuffer_Write ();
	void FileBuffer_Put (byte* data, uint size);
	void FileBuffer_Put_Byte (byte data);
	void FileBuffer_Put_Word (ushort data);
	void FileBuffer_Put_Dword (ulong data);

// is necessary for 'm_path'
	wxString m_path_str;
	const char* m_path;
	wxFile* m_file_buf_file;
	static byte m_file_buf[];
	uint m_file_buf_fill;
};

