// by Alexander Ilyin

#include "frame.h"
#include "wx/config.h"
#include "wx/wx.h"
//#include "wx/generic/filedlgg.h"
//#include "wx/richtext/richtextctrl.h"
#include "wx/display.h"
#include "ids.h"
#include "console.h"
#include "resource.h"
#include "error.h"
#include "events.h"
#include "statusbar.h"
//#include "event_propagator.h"


// todo
// wxTextCtrl, wxLogTextCtrl

DEFINE_EVENT_TYPE(wxEVT__FRAME__UPDATE_DISPLAY)
DEFINE_EVENT_TYPE(wxEVT__FRAME__DISPATCH_SIMPLE_EVENT_TO__WINDOW_FILES)
// test
DEFINE_EVENT_TYPE(wxEVT__DUMMY0)

#if 1
BEGIN_EVENT_TABLE(Frame, wxFrame)
#else
// debug !
//#define BEGIN_EVENT_TABLE(theClass, baseClass)
const wxEventTable Frame::sm_eventTable =
{
	&wxFrame::sm_eventTable, &Frame::sm_eventTableEntries[0]
};
const wxEventTable *Frame::GetEventTable() const
{
	return &Frame::sm_eventTable;
}
wxEventHashTable Frame::sm_eventHashTable (Frame::sm_eventTable);
wxEventHashTable &Frame::GetEventHashTable () const
{
	return Frame::sm_eventHashTable;
}
const wxEventTableEntry Frame::sm_eventTableEntries[] =
{
#endif
	EVT_SIZE(Frame::OnSize)
	EVT_MOVE(Frame::OnMove)
// *also does not occur, since the frame is accounted to have no focus
	EVT_CHAR(Frame::OnChar)
//	EVT_NOTIFY(FRAME__UPDATE_DISPLAY, wxID_ANY, Frame::OnUpdateDisplay)
	EVT_NOTIFY(wxEVT__FRAME__UPDATE_DISPLAY, wxID_ANY, Frame::OnUpdateDisplay)
//	EVT_NOTIFY(FRAME__DISPATCH_SIMPLE_EVENT_TO__WINDOW_FILES, wxID_ANY, Frame::OnDispatchSimpleEventTo_WindowFiles)
	EVT_NOTIFY(wxEVT__FRAME__DISPATCH_SIMPLE_EVENT_TO__WINDOW_FILES, wxID_ANY, Frame::OnDispatchSimpleEventTo_WindowFiles)
// ==== (functions, which) are here ====
	EVT_MENU(FRAME__QUIT, Frame::OnQuit)
	EVT_MENU(FRAME__DELETE, Frame::OnDelete)
// ==== mapped to external functions ====
// == file menu ==
	EVT_MENU(FRAME__OPEN_ADDITIONALLY, Frame::OnOpenAdditionally)
	EVT_MENU(FRAME__SAVE_SELECTED_FILES, Frame::OnSaveSelectedFiles)
	EVT_MENU(FRAME__SAVE_ALL_FILES, Frame::OnSaveAllFiles)
	EVT_MENU(FRAME__CLOSE_CURRENT, Frame::OnCloseCurrent)
// == edit menu ==
//// == view menu ==
// == analyze menu ==
	EVT_MENU(FRAME__ANALYZE_LEVELS, Frame::OnAnalyzeLevels)
// == process menu ==
//	EVT_MENU(FRAME__PROCESS_SAMPLERATE_HALF, Frame::OnProcess_SampleRate_Half)
//	EVT_MENU(FRAME__PROCESS_SAMPLERATE_DOUBLE, Frame::OnProcess_SampleRate_Double)
//	EVT_MENU(FRAME__PROCESS_RESAMPLE, Frame::OnProcess_Resample)
//	EVT_MENU(FRAME__PROCESS_AMPLIFY_DOUBLE, Frame::OnProcess_Amplify_Double)
// == options menu ==
//// == help menu ==
END_EVENT_TABLE()


Frame::Frame ()
	:
//	wxFrame ((wxFrame*)0, wxID_ANY, _T("NanoWave")),
	wxFrame ((wxFrame*)0, wxID_ANY, _T("NanoWave"), wxPoint (50, 50), wxSize (450, 600)),
// this doesn't change anything; this window does not receive any char events, because it cannot be in the focus
//	wxFrame ((wxFrame*)0, wxID_ANY, _T("NanoWave"), wxPoint (50, 50), wxSize (450, 600), wxDEFAULT_FRAME_STYLE | wxWANTS_CHARS),
	m_menu_file (this, &m_project),
	m_menu_edit (this, &m_project),
	m_menu_view (this, &m_project),
	m_menu_analyze (this, &m_project),
	m_menu_process (this, &m_project),
	m_menu_options (this, &m_project),
	m_menu_help (this, &m_project)
//	m_display (&m_project)
//	m_display (this, &m_project)
{
//	bool result;

	m_once_after_add = 0;
//	m_had_waves = 0;

// by id not works, need a name (the name seems automatically created by resource compiler, if the identifier is not defined before)
#if 1
	wxIcon icon;
//	icon = wxIcon (IDI_ICON1);
//	icon = wxIcon (MAKEINTRESOURCE(IDI_ICON1));
//	icon = wxICON (IDI_ICON1);
// *had to rename from IDI_ICON1 to "application_icon"
	icon = wxICON (application_icon);
	SetIcon (icon);
#endif
// seems Widgets does not load by id; lets use ordinary Windows code
#if defined(__WXMSW__)
	;
#endif

// Alex :  hmmm, how such important thing appear conditional :)
#if wxUSE_STATUSBAR
//	CreateStatusBar ();
// with three subwindows
//	CreateStatusBar (3);
	status_bar = new StatusBar (this);
//	status_bar->Init (this);
	SetStatusBar (status_bar);
	if (status_bar)
		status_bar->UpdateValues ();
#endif // wxUSE_STATUSBAR

// + menu +
	m_menu_file.Create ();
	m_menu_edit.Create ();
	m_menu_view.Create ();
	m_menu_analyze.Create ();
	m_menu_process.Create ();
	m_menu_options.Create ();
	m_menu_help.Create ();
	m_menu_bar = new wxMenuBar;
// "Do not deallocate this menu after calling Append."
	m_menu_bar->Append (m_menu_file.m_menu, _T("&File"));
	m_menu_bar->Append (m_menu_edit.m_menu, _T("&Edit"));
//	m_menu_bar->Append (m_menu_view.m_menu, _T("&View"));
	m_menu_bar->Append (&m_menu_view, _T("&View"));
	m_menu_bar->Append (m_menu_analyze.m_menu, _T("&Analyze"));
	m_menu_bar->Append (m_menu_process.m_menu, _T("&Process"));
	m_menu_bar->Append (m_menu_options.m_menu, _T("&Options"));
//	m_menu_bar->Append (m_menu_help.m_menu, _T("&Help"));
	m_menu_bar->Append (&m_menu_help, _T("&Help"));
	SetMenuBar (m_menu_bar);
// - menu -

// + splitters +
	m_splitter1 = new WindowSplitter (this, "1");
	m_splitter2 = new WindowSplitter (m_splitter1, "2");
	m_window_files = new WindowFiles (m_splitter2, &m_project);
//	m_console = new wxTextCtrl;
//	m_console = new wxTextCtrl (m_splitter2, wxID_ANY);
//	text_ctrl = new wxRichTextCtrl (m_splitter2);
//	result = console_rich_text_ctrl->Create (m_splitter2);
//	result = console_rich_text_ctrl->Reparent (m_splitter2);
	Console_Create (m_splitter2);
	m_display = new Display (m_splitter1, &m_project);
//	m_display->SetBackgroundColour (*wxRED);
//	m_display->SetScrollbars (20, 20, 5, 5);
//	m_splitter1->SplitHorizontally (m_splitter2, m_display, -100);
//	m_splitter2->SplitVertically (m_window_files, console_rich_text_ctrl, -100);
	m_splitter1->Windows_Set (m_splitter2, m_display, SPLIT_HORIZONTALLY, 1, 0, 116, 0);
	m_splitter2->Windows_Set (m_window_files, console_rich_text_ctrl, SPLIT_VERTICALLY, 1, 1, 200, 0);
// - splitters -

	Console_Init ();
//	Console_Test ();

// is not in the corresponding constructor, and is not intended to be
//	m_display->DefineSize ();
	m_window_files->DefineSize ();

// restore the control's values from the config
	wxConfigBase* config = wxConfigBase::Get ();
//	if (! config)
//		goto default_config;
// we could write Read("/Controls/Text") as well, it's just to show SetPath()
	config->SetPath (_T("/Controls"));

//	m_text->SetValue (config->Read (_T("Text"), _T("")));
//	m_check->SetValue (config->Read (_T("Check"), 1l) != 0);

// SetPath() understands ".."
//	config->SetPath (_T("../Window"));
	config->SetPath (_T("/Window"));

// restore frame position and size
// members of this class - x, y, w, h and maximize - will be set, caused by sizing subroutines request
	bool maximized;
	int x, y, w, h;
	maximized = !! config->Read (_T("Maximized"), (long)0);
	x = config->Read (_T("x"), 100);
// must be screen middle
	y = config->Read (_T("y"), 100);
	w = config->Read (_T("w"), 800);
	h = config->Read (_T("h"), 570);
	WndCheckPos (x, y, w, h);
// size catching subroutines will be requested by next two calls; this hack isn't mandatory actually
	Maximize (0);
	Move (x, y);
//	SetClientSize (w, h);
	SetSize (w, h);
// 'm_maximized' will be set inside of it
// move size size
	Maximize (maximized);

	m_splitter1->Config_Read (config);
	m_splitter2->Config_Read (config);
	m_splitter1->Values_Apply ();
	m_splitter2->Values_Apply ();

//	config->SetPath (_T("../Directories"));
	config->SetPath (_T("/Directories"));

	m_menu_file.dir_open = config->Read (_T("dir_open"), "");
	m_menu_file.dir_save = config->Read (_T("dir_save"), "");

// *must be before menu "view" init
	m_display->Config_Read (config);
	m_menu_view.Config_Read (config);
	m_menu_process.Config_Read (config);
	m_menu_view.Init (m_display);

//	EventPropagator* event_propagator = 0;
//	event_propagator = new EventPropagator;

// some events need to propagate down from frame
// let menus proceed their events themself
	PushEventHandler (m_menu_help.GetEventHandler ());
	PushEventHandler (m_menu_process.GetEventHandler ());
	PushEventHandler (m_menu_view.GetEventHandler ());
// wow, this is really wrong way :) (it handles-overrides important window events)
// atm this problem is solved by a custom 'App::FilterEvent ()', which may propagate a necessary events up
//	PushEventHandler (m_window_files->GetEventHandler ());
//	PushEventHandler (event_propagator);

	return;
}

Frame::~Frame ()
{
	wxConfigBase* config;

// remove 'event_propagator'
//	PopEventHandler ();
// remove menu "view" event handler
	PopEventHandler ();
// remove menu "process" event handler
	PopEventHandler ();
// remove menu "help" event handler
	PopEventHandler ();

	config = wxConfigBase::Get ();
	if (! config)
		return;

	// save the control's values to the config
//	config->Write (_T("/Controls/Text"), m_text->GetValue ());
//	config->Write (_T("/Controls/Check"), m_check->GetValue ());

	config->SetPath (_T("/Window"));
//	m_maximized = IsMaximized ();
// save the frame position
	config->Write (_T("Maximized"), (long)m_maximized);
	config->Write (_T("x"), (long)m_x);
	config->Write (_T("y"), (long)m_y);
	config->Write (_T("w"), (long)m_w);
	config->Write (_T("h"), (long)m_h);

	m_splitter1->Config_Write (config);
	m_splitter2->Config_Write (config);

//	config->Write (_T("/TestValue"), wxT("A test value"));

	config->Write (_T("/Directories/dir_open"), m_menu_file.dir_open);
	config->Write (_T("/Directories/dir_save"), m_menu_file.dir_save);

	m_menu_process.Config_Write (config);
	m_menu_view.Config_Write (config);
	m_display->Config_Write (config);

// have to remove from the menu bar, so it wouldn't be freed at the bar deletion moment
//	m_menu_bar->Remove (m_menu_view);
//	m_menu_bar->Remove (m_menu_view.GetId ());
	int i;
	wxMenu* menu;
// baaad
	i = m_menu_bar->FindMenu ("View");
	Assert(i != wxNOT_FOUND);
	menu = m_menu_bar->Remove (i);
	i = m_menu_bar->FindMenu ("Help");
	Assert(i != wxNOT_FOUND);
	menu = m_menu_bar->Remove (i);

	if (status_bar)
		delete (status_bar);
}

// Windows have no ability to restore a window position (yes, sounds paradoxal)
void Frame::WndCheckPos (int& x, int& y, int& w, int& h)
{
	int display_ix;
	wxRect rect;
//	bool restored;
	int limit;
	int new_value;

//	restored = 0;
	display_ix = wxDisplay::GetFromWindow (this);
	wxDisplay display (display_ix);
//	rect = display.GetGeometry ();
	rect = display.GetClientArea ();

#if 0
	if (x < 0)
	{
		Console_Warning ("window pos x %d is negative; resetting to 0", x);
		x = 0;
	}
	if (y < 0)
	{
		Console_Warning ("window pos y %d is negative; resetting to 0", y);
		y = 0;
	}
#else
#define WND_POS_OUT_OF_SANE 1000000
	if ((unsigned)x >= WND_POS_OUT_OF_SANE)
	{
		Console_Warning ("window pos x %d is out of sane; resetting to 0", x);
		x = 0;
	}
	if ((unsigned)y >= WND_POS_OUT_OF_SANE)
	{
		Console_Warning ("window pos y %d is out of sane; resetting to 0", y);
		y = 0;
	}
#endif
	limit = rect.x + rect.width;
// give some ability to a user, to drag that window where is necessary
#define WND_CORRECT_BACK 100
	if (x >= limit)
	{
		new_value = limit - WND_CORRECT_BACK;
		Console_Notify ("window pos x %d is outside of screen %d; resetting to %u", x, limit, new_value);
		x = new_value;
//		restored = 1;
	}
	limit = rect.y + rect.height;
	if (y >= limit)
	{
		new_value = limit - WND_CORRECT_BACK;
		Console_Notify ("window pos y %d is outside of screen %d; resetting to %u", y, limit, new_value);
		y = new_value;
	}
}

void Frame::OnceAfterAdd ()
{
	if (m_once_after_add)
		return;
// after an app start, the focus isn't set to anywhere; although Tab still may activate one of a windows
//// *has also to be done only once, and not after each open
	m_display->SetFocus ();
	m_once_after_add = 1;
}

// test
// bad, character events aren't coming till here
#if 0
int Frame::FilterEvent (wxEvent& event)
{
//	if (event.GetEventType () == wxEVT_KEY_DOWN && ((wxKeyEvent&)event).GetKeyCode () == WXK_F1)
	if (event.GetEventType () == wxEVT_CHAR)
	{
//		frame->OnHelpF1 ((wxKeyEvent&)event);
		return 1;
	}

	return -1;
}
#endif

// sizing, maximizing, but not a movement
void Frame::OnSize (wxSizeEvent& event)
{
	wxSize size;

	m_maximized = IsMaximized ();
	if (m_maximized)
// skip for maximized
		goto finish;
#if 1
	size = event.GetSize ();
// strange, in 'wxSize' is a warning to not take these members directly,
// and in 'wxPoint' - they can be used only directly; no even functions to take them
	m_w = size.GetWidth ();
	m_h = size.GetHeight ();
#else
// whole window, and not just client area
//	GetClientSize (&w, &h);
	GetSize (&m_w, &m_h);
#endif

finish:
	event.Skip ();
}

// sizing to the left upper corner relative, maximizing, movements
void Frame::OnMove (wxMoveEvent& event)
{
	m_maximized = IsMaximized ();
	if (m_maximized)
// skip for maximized
		goto finish;
#if 0
// wrong; it perhaps is client area screen position
	wxPoint point;
	point = event.GetPosition ();
	m_x = point.x;
	m_y = point.y;
#else
	GetPosition (&m_x, &m_y);
#endif

finish:
	event.Skip ();
}

void Frame::OnChar (wxKeyEvent& event)
{
//	wxEvtHandler* handler;

	if (! event.GetSkipped ())
	{
#if 1
		m_display->OnChar (event);
// bad, it will propagate up, and come at here again :)
#else
		handler = m_display->GetEventHandler ();
		handler->ProcessEvent (event);
#endif
	}
// stop here (although not matters)
	event.Skip ();
}

void Frame::OnUpdateDisplay (wxNotifyEvent& event)
{
// *it does not changes resolution
	m_display->DefineSize ();
	Refresh ();
}

void Frame::OnQuit (wxCommandEvent& event)
{
	Close (1);
}

#if 0
void Frame::OnAbout (wxCommandEvent& event)
{
// test
	m_window_files->DefineSize ();
	wxMessageBox (_T("NanoWave\n2008-04-28 - 2009\nAlexander Ilyin"), _T("About"), wxICON_INFORMATION | wxOK);
}
#endif

void Frame::OnDelete (wxCommandEvent& event)
{
	wxConfigBase* pConfig = wxConfigBase::Get ();
	if (pConfig == NULL)
	{
		wxLogError (_T("No config to delete!"));
		return;
	}

	if (pConfig->DeleteAll ())
	{
		wxLogMessage (_T("Config file/registry key successfully deleted."));

		delete wxConfigBase::Set (NULL);
		wxConfigBase::DontCreateOnDemand ();
	}
	else
	{
		wxLogError (_T("Deleting config file/registry key failed."));
	}
}


// +++++ map +++++

void Frame::OnOpenAdditionally (wxCommandEvent& event)
{
	bool first_load;

	first_load = ! m_project.m_waves_num;
	m_menu_file.OnOpenAdditionally ();
	if (first_load)
		m_display->Autosize ();
	m_display->DefineSize ();
//	if (! m_had_waves && m_project.m_waves_num)
//		m_display->DefineSize ();
//	m_had_waves = !! m_project.m_waves_num;
	m_window_files->Reinit (m_project.m_files, m_project.m_files_num);
// hmmm, I suspect this should be done each time after there was no windows, and not just once per app start
// no, good, just once per app start
	OnceAfterAdd ();
// initiate redraw
	Refresh ();
//	Update ();
}

void Frame::OnSaveSelectedFiles (wxCommandEvent& event)
{
	m_menu_file.OnSaveSelectedFiles ();
}

void Frame::OnSaveAllFiles (wxCommandEvent& event)
{
	m_menu_file.OnSaveAllFiles ();
}

void Frame::OnCloseCurrent (wxCommandEvent& event)
{
	m_menu_file.OnCloseFile ();
	m_window_files->Reinit (m_project.m_files, m_project.m_files_num);
// initiate redraw
//	Refresh ();
}

void Frame::OnAnalyzeLevels (wxCommandEvent& event)
{
	m_menu_analyze.OnAnalyzeLevels ();
}

#if 0
void Frame::OnProcess_SampleRate_Half (wxCommandEvent& event)
{
	m_menu_process.OnProcess_SampleRate_Half ();
	m_display->DefineSize ();
// initiate redraw
	Refresh ();
}

void Frame::OnProcess_SampleRate_Double (wxCommandEvent& event)
{
	m_menu_process.OnProcess_SampleRate_Double ();
	m_display->DefineSize ();
	Refresh ();
}

void Frame::OnProcess_Resample (wxCommandEvent& event)
{
	m_menu_process.OnProcess_Resample ();
	m_display->DefineSize ();
	Refresh ();
}

void Frame::OnProcess_Amplify_Double (wxCommandEvent& event)
{
	m_menu_process.OnProcess_Amplify_Double ();
//	m_display->DefineSize ();
	Refresh ();
}
#endif

// ----- map -----

void Frame::OnDispatchSimpleEventTo_WindowFiles (wxNotifyEvent& event)
{
	wxEvtHandler* evt_handler;
//	wxNotifyEvent evt;
	int id;

	evt_handler = m_window_files->GetEventHandler ();
//	id = (int)event.GetClientData ();
	id = event.GetInt ();
// uhoh, I must leave the declaration and initialization to be together, since hard to predict what changes may appear internally around the 'wxEvent' code
//	evt (id, GetId());
	wxNotifyEvent evt (id, GetId());
	evt_handler->ProcessEvent (evt);
}

