#pragma once

// belongs at Discoloured project

#include "types.h"


// Binary constant generator macro
// By Tom Torfs - donated to the public domain

/* All macro's evaluate to compile-time constants */

/* *** helper macros *** */

/* turn a numeric literal into a hex constant
(avoids problems with leading zeroes)
8-bit constants max value 0x11111111, always fits in unsigned long
*/
#define HEX__(n) 0x##n##LU

/* 8-bit conversion function */
#define B8__(x) ((x&0x0000000FLU)?1:0) \
+((x&0x000000F0LU)?2:0) \
+((x&0x00000F00LU)?4:0) \
+((x&0x0000F000LU)?8:0) \
+((x&0x000F0000LU)?16:0) \
+((x&0x00F00000LU)?32:0) \
+((x&0x0F000000LU)?64:0) \
+((x&0xF0000000LU)?128:0)

/* *** user macros *** */

/* for upto 8-bit binary constants */
#define B8(d) ((unsigned char)B8__(HEX__(d)))

/* for upto 16-bit binary constants, MSB first */
#define B16(dmsb,dlsb) (((unsigned short)B8(dmsb)<<\
+ B8(dlsb))

/* for upto 32-bit binary constants, MSB first */
#define B32(dmsb,db2,db3,dlsb) (((unsigned long)B8(dmsb)<<24) \
+ ((unsigned long)B8(db2)<<16) \
+ ((unsigned long)B8(db3)<<\
+ B8(dlsb))

/* Sample usage:
B8(01010101) = 85
B16(10101010,01010101) = 43605
B32(10000000,11111111,10101010,01010101) = 2164238933
*/


// This self-explanatory example was sent by Bill Finke
#define Ob(x)  ((unsigned)Ob_(0 ## x ## uL))
#define Ob_(x) (x & 1 | x >> 2 & 2 | x >> 4 & 4 | x >> 6 & 8 |		\
	x >> 8 & 16 | x >> 10 & 32 | x >> 12 & 64 | x >> 14 & 128)

#define ISDIGIT(_char_) ((_char_) >= '0' && (_char_) <= '9')

//#ifndef __cplusplus
#ifndef max
#define max(a,b)    (((a) > (b)) ? (a) : (b))
#endif
#ifndef min
#define min(a,b)    (((a) < (b)) ? (a) : (b))
#endif
//#endif

// this macro is for a defined value for the comparison;
// for the case, where the comparison will need a calculation, and will be necessary a new var, use the second macros,
// which may appear slower, if will be used there, where is necessary this macro (where no calculation is necessary)
#define MINIMIZE(_var_, _value_) do {if ((_var_) > (_value_)) _var_ = _value_;} while (0)
#define MAXIMIZE(_var_, _value_) do {if ((_var_) < (_value_)) _var_ = _value_;} while (0)
// **can't do it now proper, with C++ 2003; only with the next standard it will be possible (due to a C type definition limits)
// minimize with a calculated argument
#define MINIMIZE_VAR(_var_, _value_) do {decltype(_value_) result_val; result_val = _value_; if ((_var_) > result_val) _var_ = result_val;} while (0)
#define MAXIMIZE_VAR(_var_, _value_) do {decltype(_value_) result_val; result_val = _value_; if ((_var_) < result_val) _var_ = result_val;} while (0)

// C must reserve 'numof' keyword for that
#define NUMOF(_obj_) (sizeof(_obj_) / sizeof(*(_obj_)))
//#define NUMOF(_obj_) (sizeof(_obj_) / sizeof((_obj_)[0]))

// *on MSVC 6.0 can be used only once per a body, because the variable is declared outside of the 'for' scope
//#define FORI(_begin_, _end_) for (i = (_begin_); i < (_end_); i++)
#define _FORx(_x_, _begin_, _end_) for (uint _x_ = (_begin_); _x_ < (unsigned)(_end_); _x_++)
#define FORI(_begin_, _end_) _FORx(i, _begin_, _end_)
#define FORJ(_begin_, _end_) _FORx(j, _begin_, _end_)
#define FORX(_begin_, _end_) _FORx(x, _begin_, _end_)
#define FORY(_begin_, _end_) _FORx(y, _begin_, _end_)
#define FORI0(_end_) FORI(0, _end_)
#define FORJ0(_end_) FORJ(0, _end_)
#define FORX0(_end_) FORX(0, _end_)
#define FORY0(_end_) FORY(0, _end_)
#define FORI_R(_begin_, _end_) for (uint i = (_end_) - 1, _befor_begin = (_begin_) - 1; i != (_befor_begin); i--)
#define FORI0_R(_end_) FORI_R(0, _end_)
// hmmm, hmmmmmm
//#define FORI_R(_begin_, _end_) for (uint i = (_begin_) - 1, _i_end = (_end_) - 1; i != (_i_end); i--)
//#define FORI0_R(_begin_) FORI_R(_begin_, 0)

// remove this garbage if real 'static_assert()' is in support
// it gives "error C2466: cannot allocate an array of constant size 0" in case of a failed assert
// trick from Microsoft's code
#define _STATIC_ASSERT(expr) typedef char __static_assert_t[!! (expr)]
#define static_assert(_expr_, _cstr_) _STATIC_ASSERT(_expr_)

