// by Alexander Ilyin

#include "menu_file.h"
#include "ids.h"
#include "wx/generic/filedlgg.h"
#include "console.h"
#include "error.h"


// *equal to file format and 'file_format_names[]' lists in "file.h"
#define MENU_FILE_WILDCARD_STR_RAW "Raw files (*.raw8; *.raw16; *.raw32)"
#define MENU_FILE_WILDCARD_EXT_RAW "*.raw8;*.raw16;*.raw32"
#define MENU_FILE_WILDCARD_STR_RIFF "Microsoft PCM files (*.wav)"
#define MENU_FILE_WILDCARD_EXT_RIFF "*.wav"
#define MENU_FILE_WILDCARD_STR_ALL "All files (*.*)"
#define MENU_FILE_WILDCARD_EXT_ALL "*.*"


Menu_File::Menu_File (wxFrame* frame, Project* project)
{
	m_menu = new wxMenu;
	m_proj = project;
	m_frame = frame;

	m_wildcards.Append (_T(MENU_FILE_WILDCARD_STR_RIFF"; "MENU_FILE_WILDCARD_STR_RAW));
	m_wildcards.Append (_T("|"));
	m_wildcards.Append (_T(MENU_FILE_WILDCARD_EXT_RIFF";"MENU_FILE_WILDCARD_EXT_RAW));
	m_wildcards.Append (_T("|"));

	m_wildcards.Append (_T(MENU_FILE_WILDCARD_STR_RIFF));
	m_wildcards.Append (_T("|"));
	m_wildcards.Append (_T(MENU_FILE_WILDCARD_EXT_RIFF));
	m_wildcards.Append (_T("|"));

	m_wildcards.Append (_T(MENU_FILE_WILDCARD_STR_RAW));
	m_wildcards.Append (_T("|"));
	m_wildcards.Append (_T(MENU_FILE_WILDCARD_EXT_RAW));
	m_wildcards.Append (_T("|"));

	m_wildcards.Append (_T(MENU_FILE_WILDCARD_STR_ALL));
	m_wildcards.Append (_T("|"));
	m_wildcards.Append (_T(MENU_FILE_WILDCARD_EXT_ALL));
}

Menu_File::~Menu_File ()
{
	;
}

void Menu_File::Create ()
{
	m_menu->Append (FRAME__NEW_ADDITIONALLY, _T("&New wave"), _T("Additively create new wave"));
	m_menu->Append (FRAME__OPEN_ADDITIONALLY, _T("&Open or/and Add"), _T("Additively open file"));
	m_menu->Append (FRAME__SAVE_SELECTED_FILES, _T("&Save selected files"), _T("Save selected files of project"));
	m_menu->Append (FRAME__SAVE_ALL_FILES, _T("&Save all files"), _T("Save all files of project"));
	m_menu->Append (FRAME__CLOSE_CURRENT, _T("&Close"), _T("Close current wave"));
	m_menu->AppendSeparator ();
	m_menu->Append (FRAME__QUIT, _T("E&xit\tAlt-X"), _T("Exit the program"));
	;
}

//void Menu_File::OnOpenAdditionally (wxCommandEvent& event)
void Menu_File::OnOpenAdditionally ()
{
	wxGenericFileDialog dialog (m_frame, _T("Open and/or add files"), wxEmptyString, wxEmptyString, m_wildcards, wxFD_MULTIPLE);

	if (dir_open[0])
		dialog.SetDirectory (dir_open);
    if (dialog.ShowModal () != wxID_OK)
		return;
	wxArrayString paths;
	dir_open = dialog.GetDirectory ();
	dialog.GetPaths (paths);
	m_proj->FilesAdd (paths);
//	m_frame->;
}

// todo :  OnOpenAs ()
// were would be possible to choose file type, for how to load it

void Menu_File::OnSaveSelectedFiles ()
{
	SaveFiles (0, 0);
}

void Menu_File::OnSaveAllFiles ()
{
	SaveFiles (1, 0);
}

// todo : Save As ...

// don't save anywhere if the save name hasn't been selected manually
void Menu_File::SaveFile (int file_ix)
{
	File* file;
	wxString path;
	int i;
	wxString wildcards;
//// *'Create' must be requested anyway before deleting, so I will better create/delete it myself
	wxGenericFileDialog dialog;
//	wxGenericFileDialog* dialog;

	Assert(file_ix >= 0);
	file = m_proj->m_files[file_ix];
	Assert(file);
	if (file->m_save_name_chosen)
	{
// *'Create' must be requested anyway before deleting
		dialog.Create (m_frame, _T("___dummy___"));
		goto save;
	}
// have to choose save file name
// do proper file type
//	wxString wildcards = _T("Microsoft PCM files (*.wav)|*.wav|All files (*.*)|*.*");
	switch (file->m_info.format)
	{
	case FILE_FORMAT__RAW:
		wildcards.Append (_T(MENU_FILE_WILDCARD_STR_RAW));
		wildcards.Append (_T("|"));
//		wildcards.Append (_T(MENU_FILE_WILDCARD_EXT_RAW));
		switch (file->m_info.bits_per_sample)
		{
		case 8:
			wildcards.Append (_T("*.raw8"));
			break;
		case 16:
			wildcards.Append (_T("*.raw16"));
			break;
		case 32:
			wildcards.Append (_T("*.raw32"));
			break;
		default:
			Console_ProgramError ("Menu_File::SaveFile () :  unknown bps %u for the raw file", file->m_info.bits_per_sample);
			wildcards.Append (_T("*.raw"));
		}
		break;

	case FILE_FORMAT__RIFF:
		wildcards.Append (_T(MENU_FILE_WILDCARD_STR_RIFF));
		wildcards.Append (_T("|"));
		wildcards.Append (_T(MENU_FILE_WILDCARD_EXT_RIFF));
		break;

	default:
		Error ("Menu_File::SaveFile () :  unknown format %d", file->m_info.format);
	}
	wildcards.Append (_T("|"));
	wildcards.Append (_T(MENU_FILE_WILDCARD_STR_ALL));
	wildcards.Append (_T("|"));
	wildcards.Append (_T(MENU_FILE_WILDCARD_EXT_ALL));

	dialog.Create (m_frame, _T("Save file as ..."), wxEmptyString, wxEmptyString, wildcards, wxFD_SAVE);
	if (dir_save[0])
		dialog.SetDirectory (dir_save);
    i = dialog.ShowModal ();
	if (i == wxID_CANCEL)
		return;
	dir_save = dialog.GetDirectory ();
	path = dialog.GetPath ();
//	file->m_file_path.Assign ("prb.wav");
	file->m_file_path.Assign (path);
	file->m_save_name_chosen = 1;

save:
	m_proj->FileSave (file_ix);
}

void Menu_File::SaveFiles (bool all, bool upon_origin)
{
	File* file;
	uint i;
	wxMessageDialog* mbox;
	int files_to_save = 0;

	if (upon_origin)
// prompt for overwrite
	{
		mbox = new wxMessageDialog (m_frame, "You have been requested origin files overwrite.\nAre you sure to overwrite origin files ?", "Files overwrite certification", wxOK | wxCANCEL);
		i = mbox->ShowModal ();
		if (i == wxID_CANCEL)
			return;
	}

	for (i = 0; i < m_proj->m_files_num; i++)
	{
		file = m_proj->m_files[i];
		if (! all)                                              
//			if (! m_proj->m_file_is_selected[i])
			if (! file->m_is_selected)
				continue;
		files_to_save++;
		if (upon_origin)
// don't ask for file names
			file->m_save_name_chosen = 1;
		SaveFile (i);
	}

// don't report "0 files saved", report "no files to save" instead
	if (! files_to_save)
	{
		if (all)
//			Console_Warning ("Menu_File::SaveFiles () :  no files to save");
			Console_Warning ("no files to save");
		else
			Console_Warning ("saving :  no files selected");
	}
}

void Menu_File::OnCloseFile ()
{
//**
//	m_proj->FileClose (0);
//	m_proj->FileDelete (0);
	if (m_proj->m_files_num)
		m_proj->File_Remove (0);
}

