// by Alexander Ilyin

#include "menu_process.h"
#include "ids.h"
#include "events.h"


BEGIN_EVENT_TABLE(Menu_Process, wxMenu)
	EVT_MENU(MENU_PROCESS__PROCESS_SAMPLERATE_HALF, Menu_Process::OnProcess_SampleRate_Half)
	EVT_MENU(MENU_PROCESS__PROCESS_SAMPLERATE_DOUBLE, Menu_Process::OnProcess_SampleRate_Double)
	EVT_MENU(MENU_PROCESS__PROCESS_RESAMPLE, Menu_Process::OnProcess_Resample)
	EVT_MENU(MENU_PROCESS__PROCESS_AMPLIFY_DOUBLE, Menu_Process::OnProcess_Amplify_Double)
END_EVENT_TABLE()


Menu_Process::Menu_Process (wxFrame* frame, Project* project)
{
	m_menu = new wxMenu;
	m_proj = project;
	m_frame = frame;
	m_dialog_resample = new DialogResample (m_frame);
}

Menu_Process::~Menu_Process ()
{
//	m_dialog_resample->Destroy ();
	delete m_dialog_resample;
	m_dialog_resample = 0;
}

void Menu_Process::Create ()
{
	m_menu->Append (MENU_PROCESS__PROCESS_SAMPLERATE_HALF, _T("Sample rate &half"), _T("Make half sample rate"));
	m_menu->Append (MENU_PROCESS__PROCESS_SAMPLERATE_DOUBLE, _T("Sample rate double"), _T("Make double sample rate"));
	m_menu->Append (MENU_PROCESS__PROCESS_RESAMPLE, _T("Resample"), _T("Remake to another sample rate"));
	m_menu->Append (MENU_PROCESS__PROCESS_AMPLIFY_DOUBLE, _T("Amplify double"), _T("Make two times higher volume"));
//	m_menu->AppendSeparator ();
}

void Menu_Process::Config_Read (wxConfigBase* config)
{
	m_dialog_resample->Config_Read (config);
}

void Menu_Process::Config_Write (wxConfigBase* config)
{
	m_dialog_resample->Config_Write (config);
}

void Menu_Process::SendUpdateEvent ()
{
//	wxNotifyEvent event (wxEVT__FRAME__UPDATE_DISPLAY, GetId());
// both (false ids) works, but I am not sure that this will not have some hidden bugs
//	wxNotifyEvent event (wxEVT__FRAME__UPDATE_DISPLAY, 0);
	wxNotifyEvent event (wxEVT__FRAME__UPDATE_DISPLAY, wxID_ANY);
	event.SetEventObject (this);
	GetEventHandler ()->ProcessEvent (event);
}

// I am thinking now to give for these some data array, instead of ptr to 'Menu_Process' as userdata,
// and move them out, perhaps to "process.cpp".
// +++ callbacks +++

void Menu_Process::Callback_SampleRate_Half (void* void_ptr, Wave* wave)
{
	Menu_Process* menu_process = (Menu_Process*)void_ptr;
	wave->Process_Samples_Half (1, 1);
}

void Menu_Process::Callback_SampleRate_Double (void* void_ptr, Wave* wave)
{
	Menu_Process* menu_process = (Menu_Process*)void_ptr;
//	wave->Process_Samples_Double (WAVE_RESAMPLE_TYPE__AVERAGE, 1, 0, 1);
//	wave->Process_Samples_Double (WAVE_RESAMPLE_TYPE__BENDS_PLUS_ONE, 1, 0, 1);
	wave->Process_Samples_Double (WAVE_RESAMPLE_TYPE__BENDS_CONTINUOUS, 1, 0, 1);
//	wave->Process_Samples_Double (WAVE_RESAMPLE_TYPE__POWER_SOME, 1, 0, 1);
}

void Menu_Process::Callback_Resample (void* void_ptr, Wave* wave)
{
	Menu_Process* menu_process = (Menu_Process*)void_ptr;
	wave->Process_Resample (menu_process->m_dialog_resample->m_val_cycle_samples_out, menu_process->m_dialog_resample->m_val_cycle_samples_in, menu_process->m_dialog_resample->m_val_change_samplerate);
}

// *simplified; really - should use rational fraction
void Menu_Process::Callback_Amplify_Double (void* void_ptr, Wave* wave)
{
	Menu_Process* menu_process = (Menu_Process*)void_ptr;
// meanwhile no rate dialog box is done, set amplification rate (here) manually :)
	wave->Process_Amplify (2, 1);
}

// --- callbacks ---

// using callbacks makes the thing more complex, but otherwise - will be much those 'for' loops to iterate between the waves

// +++ events +++

void Menu_Process::OnProcess_SampleRate_Half (wxCommandEvent& event)
{
//	m_proj->Process_Samples_Half (1, 1);
	m_proj->SelectedWaves_IterateFunction (Callback_SampleRate_Half, this);
	SendUpdateEvent ();
}

void Menu_Process::OnProcess_SampleRate_Double (wxCommandEvent& event)
{
//	m_proj->Process_Samples_Double (1, 0, 0, 1);
	m_proj->SelectedWaves_IterateFunction (Callback_SampleRate_Double, this);
	SendUpdateEvent ();
}

void Menu_Process::OnProcess_Resample (wxCommandEvent& event)
{
	bool result;
	int i_result;

//	m_proj->Process_Resample (3, 4, 1);
//	m_proj->Process_Resample (4, 3, 1);
//	if (! dialog_resample->m_is_confirmed)
//		return;
	result = m_proj->SelectedWaves_GetSingleSamplerate (m_dialog_resample->m_start_samplerate);
	if (! result)
// no common samplerate
		m_dialog_resample->m_start_samplerate = 0;
	i_result = m_dialog_resample->ShowModal ();
	if (i_result != wxID_OK)
		return;
//	m_proj->Process_Resample (m_dialog_resample->m_val_cycle_samples_out, m_dialog_resample->m_val_cycle_samples_in, 1);
	m_proj->SelectedWaves_IterateFunction (Callback_Resample, this);
	SendUpdateEvent ();
}

void Menu_Process::OnProcess_Amplify_Double (wxCommandEvent& event)
{
	m_proj->SelectedWaves_IterateFunction (Callback_Amplify_Double, this);
	SendUpdateEvent ();
}

// --- events ---

