// by Alexander Ilyin

#include "menu_view.h"
#include "ids.h"
#include "error.h"


BEGIN_EVENT_TABLE(Menu_View, wxMenu)
	EVT_MENU(VIEW_WAVE_SIMPLE, Menu_View::OnCommand)
	EVT_MENU(VIEW_WAVE_MIPMAP, Menu_View::OnCommand)
	EVT_MENU(VIEW_WAVE_FULL, Menu_View::OnCommand)
END_EVENT_TABLE()


Menu_View::Menu_View (wxFrame* frame, Project* project)
{
	m_frame = frame;
	m_display = 0;
	m_proj = project;
//	m_menu = new wxMenu;
	m_menu_wave = new wxMenu;
//	m_dialog_resample = new DialogResample (m_frame);
}

Menu_View::~Menu_View ()
{
//	m_dialog_resample->Destroy ();
//	delete m_dialog_resample;
//	m_dialog_resample = 0;
}

void Menu_View::Create ()
{
//	Assert(display);
//	Assert(! m_display);
//	m_display = display;
	m_menuitem_view_wave_simple = m_menu_wave->Append (VIEW_WAVE_SIMPLE, _T("Simple (fast)"), _T("fast method"), wxITEM_CHECK);
	m_menuitem_view_wave_mipmap = m_menu_wave->Append (VIEW_WAVE_MIPMAP, _T("Mipmap (optimized)"), _T("optimal method"), wxITEM_CHECK);
//	m_menuitem_view_wave_mipmap->Enable (0);
	m_menuitem_view_wave_full = m_menu_wave->Append (VIEW_WAVE_FULL, _T("Full processing (slow)"), _T("slow method"), wxITEM_CHECK);
//	m_menu->AppendSubMenu (m_menu_wave, _T("Wave"), _T("View of a wave"));
	AppendSubMenu (m_menu_wave, _T("Wave"), _T("View of a wave"));
//	m_menu->AppendSeparator ();
}

void Menu_View::Config_Read (wxConfigBase* config)
{
//	m_dialog_resample->Config_Read (config);
}

void Menu_View::Config_Write (wxConfigBase* config)
{
//	m_dialog_resample->Config_Write (config);
}

void Menu_View::Init (Display* display)
{
//	Assert(m_display);
	Assert(display);
	Assert(! m_display);
	m_display = display;

	Menu_View::SetCheckboxes_WaveView ();
}

void Menu_View::SetCheckboxes_WaveView ()
{
	Assert(m_display);

	m_menuitem_view_wave_simple->Check (0);
	m_menuitem_view_wave_mipmap->Check (0);
	m_menuitem_view_wave_full->Check (0);
	switch (m_display->m_view_wave_blend_mode)
	{
	case WAVE_VIEW__SIMPLE:
		m_menuitem_view_wave_simple->Check (1);
		break;
	case WAVE_VIEW__MIPMAP:
		m_menuitem_view_wave_mipmap->Check (1);
		break;
	case WAVE_VIEW__FULL:
		m_menuitem_view_wave_full->Check (1);
		break;
	default:
		Error ("Menu_View::SetCheckboxes_WaveView () :  wave view mode %d is unknown", m_display->m_view_wave_blend_mode);
	}
}

void Menu_View::OnCommand (wxCommandEvent& event)
{
	int id;

	id = event.GetId ();
	switch (id)
	{
	case VIEW_WAVE_SIMPLE:
		m_display->m_view_wave_blend_mode = WAVE_VIEW__SIMPLE;
		break;
	case VIEW_WAVE_MIPMAP:
		m_display->m_view_wave_blend_mode = WAVE_VIEW__MIPMAP;
		break;
	case VIEW_WAVE_FULL:
		m_display->m_view_wave_blend_mode = WAVE_VIEW__FULL;
		break;
	default:
		event.Skip ();
		return;
	}
	Menu_View::SetCheckboxes_WaveView ();
	m_frame->Refresh ();
}

#if 0
// +++ callbacks +++

void Menu_View::Callback_SampleRate_Half (void* void_ptr, Wave* wave)
{
	Menu_View* Menu_View = (Menu_View*)void_ptr;
	wave->Process_Samples_Half (1, 1);
}

void Menu_View::Callback_SampleRate_Double (void* void_ptr, Wave* wave)
{
	Menu_View* menu_view = (Menu_View*)void_ptr;
//	wave->Process_Samples_Double (WAVE_RESAMPLE_TYPE__AVERAGE, 1, 0, 1);
	wave->Process_Samples_Double (WAVE_RESAMPLE_TYPE__PLUS_ONE, 1, 0, 1);
}

void Menu_View::Callback_Resample (void* void_ptr, Wave* wave)
{
	Menu_View* menu_view = (Menu_View*)void_ptr;
	wave->Process_Resample (menu_view->m_dialog_resample->m_val_cycle_samples_out, menu_process->m_dialog_resample->m_val_cycle_samples_in, menu_process->m_dialog_resample->m_val_change_samplerate);
}

// --- callbacks ---

// using callbacks makes the thing more complex, but otherwise - will be much those 'for' loops to iterate between the waves

// +++ events +++

void Menu_View::OnProcess_SampleRate_Half ()
{
//	m_proj->Process_Samples_Half (1, 1);
	m_proj->SelectedWaves_IterateFunction (Callback_SampleRate_Half, this);
}

void Menu_View::OnProcess_SampleRate_Double ()
{
//	m_proj->Process_Samples_Double (1, 0, 0, 1);
	m_proj->SelectedWaves_IterateFunction (Callback_SampleRate_Double, this);
}

void Menu_View::OnProcess_Resample ()
{
	bool result;
	int i_result;

//	m_proj->Process_Resample (3, 4, 1);
//	m_proj->Process_Resample (4, 3, 1);
//	if (! dialog_resample->m_is_confirmed)
//		return;
	result = m_proj->SelectedWaves_GetSingleSamplerate (m_dialog_resample->m_start_samplerate);
	if (! result)
// no common samplerate
		m_dialog_resample->m_start_samplerate = 0;
	i_result = m_dialog_resample->ShowModal ();
	if (i_result != wxID_OK)
		return;
//	m_proj->Process_Resample (m_dialog_resample->m_val_cycle_samples_out, m_dialog_resample->m_val_cycle_samples_in, 1);
	m_proj->SelectedWaves_IterateFunction (Callback_Resample, this);
}

// --- events ---
#endif

