// by Alexander Ilyin

#include "project.h"
#include "string.h"
#include "console.h"
#include "error.h"


Project::Project ()
{
	memset (m_files, 0, sizeof(m_files));
	memset (m_waves, 0, sizeof(m_waves));
	m_files_num = 0;
	m_waves_num = 0;
	m_have_changes = 0;

//	m_files_opened = 0;
// +++ settings +++
	m_update_files_after_waves_processing = 1;
// --- settings ---
}

Project::~Project ()
{
}

bool Project::File_Add (File* file, uint insert_at_ix)
{
	int i;

	Assert(insert_at_ix < PROJECT_FILES_MAX);
	Assert(insert_at_ix <= m_files_num);

	for (i = m_files_num - 1; i >= (signed)insert_at_ix; i--)
	{
		m_files[i + 1] = m_files[i];
		m_files[i + 1]->m_ix = i + 1;
	}

	file->m_ix = insert_at_ix;
	m_files[insert_at_ix] = file;
//	m_file_is_selected[insert_at_ix] = 0;
	m_files_num++;

	return 1;
}

bool Project::File_Remove (uint position_at_ix)
{
	uint i;
	File* file;

	Assert(position_at_ix < PROJECT_FILES_MAX);
	Assert(position_at_ix < m_files_num);
	Assert(m_files_num > 0);

	file = m_files[position_at_ix];
	delete file;
	m_files_num--;

	for (i = position_at_ix; i < m_files_num; i++)
	{
		m_files[i] = m_files[i + 1];
		m_files[i]->m_ix = i;
	}

	return 1;
}

bool Project::File_ParseAndAdd (wxString& file_name, File*& file)
{
	bool result;
//	int i;
	int insert_at_ix;
	File* file_new;
	wxFileName* wx_file_name;
	wxString file_ext;
	int bits_per_second;

	insert_at_ix = m_files_num;
//	ASSERT(insert_at_ix < PROJECT_FILES_MAX);
	if (insert_at_ix >= PROJECT_FILES_MAX)
// nowhere to add
		return 0;

	file_new = new File;
//	file_new->m_format = FILE_FORMAT__RIFF;
//	Strncpy (file_new->file_path, file_name, sizeof(file_new->file_path));
	wx_file_name = &file_new->m_file_path;
	wx_file_name->Assign (file_name);
	file_ext = wx_file_name->GetExt ();

// == define type by extension ==
// (necessary for raw files; their content must not be parsed anyhow)
	result = ! file_ext.CmpNoCase (_T("raw8"));
	if (result)
	{
		bits_per_second = 8;
		goto load_raw;
	}
	result = ! file_ext.CmpNoCase (_T("raw16"));
	if (result)
	{
		bits_per_second = 16;
		goto load_raw;
	}
	result = ! file_ext.CmpNoCase (_T("raw32"));
	if (result)
	{
		bits_per_second = 32;
		goto load_raw;
	}
//	goto parse;
// == define type by contain ==
//parse:
	result = file_new->Parse ();
	if (! result)
	{
		delete file_new;
		return 0;
	}
	goto file_add;

load_raw:
	file_new->DefineAsRaw (bits_per_second, 1, 48000);

file_add:
	File_Add (file_new, insert_at_ix);
//	m_files_opened++;

	if (file)
		file = file_new;

	return 1;
}

#define WAVES_IN_FILE_MAX 16

bool Project::File_CreateAndLoadWaves (File* file)
{
	int waves_sum;
	int i;
	int waves_num_new;
	const char* path;
	bool result;

	path = file->GetName ();
	waves_num_new = file->m_info.channels_num;
	if (waves_num_new > WAVES_IN_FILE_MAX)
	{
		Console_Error ("Project::File_CreateAndLoadWaves () :  %u waves, max is %u, in file :  \"%s\"", waves_num_new, WAVES_IN_FILE_MAX, path);
		return 0;
	}
	waves_sum = m_waves_num + waves_num_new;
	if (waves_sum > PROJECT_WAVES_MAX)
	{
		Console_Error ("Project::File_CreateAndLoadWaves () :  waves limit reached, %u  is max, breaking on file :  \"%s\"", PROJECT_WAVES_MAX, path);
		return 0;
	}

	for (i = 0; i < waves_num_new; i++)
	{
		int wave_ix;
		Wave* wave;
		wave = new Wave;
		result = wave->SetSize (file->m_info.samples_num);
		wave_ix = m_waves_num + i;
		wave->m_ix = wave_ix;
		m_waves[wave_ix] = wave;
// init
		m_wave_is_selected[wave_ix] = 0;
	}

// the method to give ptr to array maybe isn't pretty;
// the array is intended to be continuous at a load time;
// continuousity must be supported by wave(s) removing function
	result = file->Waves_Load (&m_waves[m_waves_num], waves_num_new);

	m_waves_num = waves_sum;

	return 1;
}

void Project::File_ParseAddAndLoad (wxString& file_name)
{
	bool result;
	File* file_new;

	result = File_ParseAndAdd (file_name, file_new);
	if (! result)
		return;

	result = File_CreateAndLoadWaves (file_new);
}

void Project::FilesAdd (wxArrayString& paths)
{
//	const char* str;
	wxString str;
	size_t count = paths.GetCount ();
	for (size_t n = 0; n < count; n++)
	{
//		str = paths[n].c_str ();
		str = paths[n];
		File_ParseAddAndLoad (str);
	}
}

void Project::FileSave (uint file_ix)
{
	File* file;

	if (! m_files_num)
	{
		Console_Error ("Project::FileSave () :  no files while requested to save file %d", file_ix);
		return;
	}
	if (file_ix >= m_files_num)
	{
		Console_Error ("Project::FileSave () :  requested to save file %d, %d currently exist", file_ix, m_files_num);
		return;
	}

	file = m_files[file_ix];
	if (! file->m_save_name_chosen)
	{
//		Console_Error ("Project::FileSave () :  file %d save name has not been chosen :  \"%s\"", file_ix, file->m_file_path.c_str ());
		Console_Error ("Project::FileSave () :  file %d save name has not been chosen", file_ix);
		return;
	}
	file->Save ();
}

#if 0
void Project::FilesSave ()
{
	File* file;

	if (! m_files_num)
		return;

// test
	file = m_files[0];
	file->m_file_path.Assign ("prb.wav");
	file->m_save_name_chosen = 1;
	file->Save ();
}
#endif

// hmmm, no need; there is 'File_Remove ()'
#if 0
//void Project::FileClose (uint file_ix)
void Project::FileDelete (uint file_ix)
{
}
#endif

// return 1 if some wave is selected
bool Project::SelectedWaves_CheckAndReport ()
{
	uint i;

	for (i = 0; i < m_waves_num; i++)
		if (m_wave_is_selected[i])
			return 1;
	Console_Warning ("no waves selected");
	return 0;
}


// +++ analyze +++

void Project::Analyze_Levels ()
{
	uint i;
//	wave_data_t level_totally_max = 0;
//	wave_data_t level_totally_min = 0;
	bool result;

	result = SelectedWaves_CheckAndReport ();
	if (! result)
		return;
	m_level_totally_max = 0;
	m_level_totally_min = 0;
	for (i = 0; i < m_waves_num; i++)
	{
		Wave* wave;
		wave = m_waves[i];
		if (! m_wave_is_selected[i])
			continue;
		wave->Analyze_Level ();
//		if (level_totally_max < wave->m_level_max)
//			level_totally_max = wave->m_level_max;
//		if (level_totally_min > wave->m_level_min)
//			level_totally_min = wave->m_level_min;
		if (m_level_totally_max < wave->m_level_max)
			m_level_totally_max = wave->m_level_max;
		if (m_level_totally_min > wave->m_level_min)
			m_level_totally_min = wave->m_level_min;
	}
}

// --- analyze ---

// +++ process +++

void Project::UpdateFiles (uint wave_ix)
{
	uint i;

	for (i = 0; i < m_files_num; i++)
	{
		bool result;
		File* file;
		file = m_files[i];
		result = file->IsWaveMy (wave_ix);
		if (result)
//			file->UpdateSamplesNum ();
			file->UpdateParameters ();
	}
}

void Project::SelectedWaves_IterateFunction (void(* func) (void* void_ptr, Wave* wave), void* void_ptr)
{
	uint i;
	bool result;

	result = SelectedWaves_CheckAndReport ();
	if (! result)
		return;
	for (i = 0; i < m_waves_num; i++)
	{
		Wave* wave;
		wave = m_waves[i];
		if (! m_wave_is_selected[i])
			continue;
		func (void_ptr, wave);
		if (m_update_files_after_waves_processing)
			UpdateFiles (wave->m_ix);
	}
}

// find out common sample rate; return 0, if rates are different
bool Project::SelectedWaves_GetSingleSamplerate (uint& samplerate)
{
	bool result;
	uint rate;
	uint i;
	bool got_rate = 0;

	result = SelectedWaves_CheckAndReport ();
	if (! result)
		return 0;
	for (i = 0; i < m_waves_num; i++)
	{
		Wave* wave;
		wave = m_waves[i];
		if (! m_wave_is_selected[i])
			continue;
		if (! got_rate)
		{
			rate = wave->m_sample_rate;
			got_rate = 1;
			continue;
		}
		if (rate != wave->m_sample_rate)
			return 0;
	}
	samplerate = rate;
	return 1;
}

#if 0
void Project::Process_Samples_Half (bool average_level, bool downrate)
{
	int i;
	bool result;

	result = SelectedWaves_CheckAndReport ();
	if (! result)
		return;
	for (i = 0; i < m_waves_num; i++)
	{
		Wave* wave;
		wave = m_waves[i];
		if (! m_wave_is_selected[i])
			continue;
		wave->Process_Samples_Half (average_level, downrate);
		if (m_update_files_after_waves_processing)
			UpdateFiles (wave->m_ix);
	}
//	if (downrate)
//		;
}

void Project::Process_Samples_Double (bool uprate, bool duplicate, bool extend_begin, bool extend_end)
{
	int i;
	bool result;

	result = SelectedWaves_CheckAndReport ();
	if (! result)
		return;
	for (i = 0; i < m_waves_num; i++)
	{
		Wave* wave;
		wave = m_waves[i];
		if (! m_wave_is_selected[i])
			continue;
		wave->Process_Samples_Double (uprate, duplicate, extend_begin, extend_end);
		if (m_update_files_after_waves_processing)
			UpdateFiles (wave->m_ix);
	}
}

void Project::Process_Resample (uint cycle_size_out, uint cycle_size_in, bool change_rate)
{
	int i;
	bool result;

	result = SelectedWaves_CheckAndReport ();
	if (! result)
		return;
	for (i = 0; i < m_waves_num; i++)
	{
		Wave* wave;
		wave = m_waves[i];
		if (! m_wave_is_selected[i])
			continue;
		wave->Process_Resample (cycle_size_out, cycle_size_in, change_rate);
		if (m_update_files_after_waves_processing)
			UpdateFiles (wave->m_ix);
	}
}
#endif

// --- process ---

