// by Alexander Ilyin
// *original belongs at the Dragging Watcher project


#include "string.h"

#include <string.h>
#include <stdarg.h>

#include "error.h"


// 'strncpy ()' copy at most n characters of string -> remaked -> 'Strncpy ()' copy into buffer of size n
// return 1, if string was copied with it's '\0'
bool Strncpy (char* dest, const char* source, size_t buffer_size)
{
// *leave one char for '\0'
	if (buffer_size < sizeof(*dest))
// no buffer
		return 0;

	while (1)
	{
// copy string character
		if (! (*dest = *source))
// string successfully finished
			return 1;

		if (buffer_size == sizeof(*dest))
// no more place in buffer
		{
// finish string
			*dest = L'\0';
			return 0;
		}

		dest++;
		source++;
		buffer_size--;
	}

//	return 0;
}

// 'wcsncpy ()'
// return 1, if string was copied with it's L'\0'
bool WStrncpy (wchar_t* dest, const wchar_t* source, size_t buffer_size)
{
// lets simplificate calculations below
	buffer_size &= 0xfffffffe;

// *leave one char for '\0'
	if (buffer_size < sizeof(*dest))
// no buffer
		return 0;

	while (1)
	{
// copy string character
		if (! (*dest = *source))
// string successfully finished
			return 1;

		if (buffer_size == sizeof(*dest))
// no more place in buffer
		{
// finish string
			*dest = L'\0';
			return 0;
		}

		dest++;
		source++;
		buffer_size--;
	}

//	return 0;
}

// copy, and return 0, if it is not alphanum or does not fit into buffer
bool Str_IsAlphaNum (const char* str)
{
//	bool result;
	char c;

//	result = Strncpy (dest, source, buffer_size);
//	if (! result)
//		return 0;
	while (c = *str)
	{
		if (c >= 'a' && c <= 'z')
			goto continue_loop;
		if (c >= 'A' && c <= 'Z')
			goto continue_loop;
		if (c >= '0' && c <= '9')
			goto continue_loop;
// this one allow also
		if (c == '_')
			goto continue_loop;
// not alphanum character
		return 0;
continue_loop:
		str++;
	}
// string successfully finished
	return 1;
}

// insert one char in a string in sized buffer
bool StringN_InsertChar (char* string, uint string_buffer_size, char char_insert, uint position)
{
uint string_full_lenght = strlen (string) + 1;
uint i;

	if (string_full_lenght >= string_buffer_size)
// no place to insert
		return 0;

	i = string_full_lenght;

// copy the string one symbol forward
	while (i-- > position)
//		string[i] = string[--i];
		string[i + 1] = string[i];

// place the symbol
	string[position] = char_insert;

	return 1;
}

// delete one char, to the left from the position
bool String_BackspaceChar (char* string, uint position)
{
uint string_lenght = strlen (string);

	if (! position || position > string_lenght)
// no char, to the left from the position
		return 0;

// copy the string one symbol back
	while (string[position - 1] = string[position])
		position++;

	return 1;
}

// delete one char, on the position
bool String_DeleteChar (char* string, uint position)
{
uint string_lenght = strlen (string);

	if (position >= string_lenght)
// no char, on the position
		return 0;

// copy the string one symbol back
	while (string[position] = string[position + 1])
		position++;

	return 1;
}

bool String_UnicodeToANSI (char* ansi_string, int buffer_size, ushort* unicode_string, int string_size)
{
int i;
int chars_number = string_size / sizeof(*unicode_string);
bool result = 1;

	if (chars_number >= buffer_size)
	{
		chars_number = buffer_size - 1;
		result = 0;
	}

	for (i = 0; i < chars_number; i++)
		ansi_string[i] = (char)unicode_string[i];

	ansi_string[i] = '\0';

	return result;
}


#if 0
//	_snprintf (string, buffer_size - 1, format, ...);
bool Snprintf (char* string, uint buffer_size, const char* format, ...)
{
#if 0

format:
	while (buffer_size)
//	while (*format)
	{
		if (*format == '%')
		{
			goto argument;
		}
		else
		{
			if (! (*string++ = *format++))
// format end
				return 1;
			buffer_size--;
		}
	}

// buffer is full
	return 0;

argument:

	goto format;

#else

FILE str;
FILE *outfile = &str;
va_list arglist;
int retval;
uint count = buffer_size - 1;

	va_start(arglist, format);

	Assert(string != NULL);
	Assert(format != NULL);

	outfile->_flag = _IOWRT|_IOSTRG;
	outfile->_ptr = outfile->_base = string;
	outfile->_cnt = count;

	retval = _output(outfile,format,arglist);

// no-lock version
	_putc_lk('\0',outfile);

//	return(retval);
	return outfile->_cnt < count;

#endif
}
#endif
