// by Alexander Ilyin

#include "wave.h"
// for wxASSERT ()
#include "wx/defs.h"
// nah, don't enable it; here is too much such conversions, and I don't want to cast them all; and them is intended to be more and more
//#pragma warning(default:4244)
#include <stdlib.h>
#include <math.h>
#include "console.h"
#include "error.h"


Wave::Wave ()
{
	m_link_previous = -1;
	m_link_next = -1;
	m_references_count = 0;

	m_bits_per_sample = 0;
	m_sample_rate = 0;
	m_sample_time = 0;
	m_time_offset = 0;

	m_samples_num = 0;
	m_data = 0;

	m_data__in_process = 0;
	m_samples_num__in_process = 0;
	m_was_level_cuts = 0;
	m_segment_duration = 0;
//	Mipmap_InvalidateLevels ();
//	memset (m_is_mipmap_level_valid, 0, sizeof(m_is_mipmap_level_valid));
	memset (m_is_mipmap_level_allocated, 0, sizeof(m_is_mipmap_level_allocated));
	memset (m_is_mipmap_level_filled, 0, sizeof(m_is_mipmap_level_filled));
	memset (m_data_mipmap_max, 0, sizeof(m_data_mipmap_max));
	memset (m_data_mipmap_avg, 0, sizeof(m_data_mipmap_avg));
	memset (m_data_mipmap_min, 0, sizeof(m_data_mipmap_min));
	memset (m_mipmap_level_size, 0, sizeof(m_mipmap_level_size));
}

Wave::~Wave ()
{
	wxASSERT(! m_references_count);
}

void Wave::Reference ()
{
	wxASSERT(m_references_count >= 0);
	m_references_count++;
}

void Wave::Dereference ()
{
	wxASSERT(m_references_count > 0);
	m_references_count--;
	if (! m_references_count)
	{
		free (m_data);
		m_data = 0;
	}
}

#define MALLOC_SIZE_MAX (1 << 31)

bool Wave::SetSize (uint64 samples_num)
{
	uint64 size;

	wxASSERT (! m_data);

	size = sizeof(*m_data) * samples_num;
// "ISO/IEC 9899:1999" does not specify the max size
	if (size >= MALLOC_SIZE_MAX)
// atm I am not thinking how to handle such size
	{
		Console_Error ("Wave :  attempt to allocate %u bytes of memory has been blocked, %u max allowed", size, MALLOC_SIZE_MAX);
		return 0;
	}

	m_data = (wave_data_t*)malloc ((uint)size);
	if (! m_data)
	{
		Console_Error ("Wave :  malloc () fail, can't allocate %u bytes of memory", size);
		return 0;
	}
// this seems is logical; then externally looking it will be 'SetSize ()'/'Dereference ()' paired
	Reference ();
	m_samples_num = samples_num;

	return 1;
}

// some processes requires new memory, which also may differs in size, while old memory must also remain
// allocation is here, and freeing (of old or new memory) must be in the requestor function
bool Wave::ForProcessing_SetSize (uint64 samples_num)
{
	uint64 size;

	size = sizeof(*m_data__in_process) * samples_num;
	if (size >= MALLOC_SIZE_MAX)
	{
		Console_Error ("Wave::SetSize_ForProcessing () :  attempt to allocate %u bytes of memory has been blocked, %u max allowed", size, MALLOC_SIZE_MAX);
		return 0;
	}

	m_data__in_process = (wave_data_t*)malloc ((uint)size);
	if (! m_data__in_process)
	{
		Console_Error ("Wave::SetSize_ForProcessing () :  malloc () fail, can't allocate %u bytes of memory", size);
		return 0;
	}
	m_samples_num__in_process = samples_num;

	return 1;
}

void Wave::ForProcessing_ReplaceOld ()
{
	free (m_data);
	m_data = m_data__in_process;
	m_data__in_process = 0;
	m_samples_num = m_samples_num__in_process;
	m_samples_num__in_process = 0;
// reallocate mipmap, but do not fill yet
//	Mipmap_InvalidateLevels ();
	Mipmap_Free ();
// perhaps not necessary; will be requested when necessary anyway
//	Mipmap_Allocate ();
}


// +++ analyze +++

void Wave::Analyze_Level ()
{
	uint64 i;
	wave_data_t level_max = 0;
	wave_data_t level_min = 0;

	wxASSERT(m_data);
	for (i = 0; i < m_samples_num; i++)
	{
		wave_data_t level;
		level = m_data[i];
		if (level >= 0)
		{
			if (level_max < level)
				level_max = level;
		}
		else
		{
			if (level_min > level)
				level_min = level;
		}
	}
	m_level_max = level_max;
	m_level_min = level_min;
}

// --- analyze ---

// +++ process +++

// "flat" wave model
// that 'average_level' is near to be completely unnecessary thing, although isn't abs meaningless
void Wave::Process_Samples_Half (bool average_level, bool downrate)
{
	uint64 samples_num_new;
	uint64 i;
	uint64 j;
//	bool result;
	bool level_over = 0;

// last sample must remain
	samples_num_new = (m_samples_num + 1) >> 1;
	if (! ForProcessing_SetSize (samples_num_new))
		return;

	for (i = 0, j = 0; i < samples_num_new; i++)
	{
//		wave_data_u_t lvl;
		int64 lvl;

		wxASSERT(j < m_samples_num);
		lvl = m_data[j];
		j++;
// if not available (last unpaired value, if odd samples num), then think about it as about 0 level
		if (j < m_samples_num)
			lvl += m_data[j];
		j++;
// test
//		lvl *= 1.3;
		if (average_level)
			lvl >>= 1;
		else
// it may be over max/min level
		{
			if (lvl < WAVE_LEVEL_MIN)
			{
				lvl = WAVE_LEVEL_MIN;
				level_over = 1;
			}
			else if (lvl > WAVE_LEVEL_MAX)
			{
				lvl = WAVE_LEVEL_MAX;
				level_over = 1;
			}
		}
		m_data__in_process[i] = (wave_data_t)lvl;
	}
	ForProcessing_ReplaceOld ();
	if (downrate)
	{
		m_sample_rate >>= 1;
		if (! m_sample_rate)
		{
			Console_Warning ("Wave::Process_Samples_Half () :  sample rate is minimal");
			m_sample_rate = 1;
		}
		Console_Notify ("%d samplerate %u", m_ix, m_sample_rate);
	}
	if (level_over)
		Console_Warning ("Wave::Process_Samples_Half () :  the wave had samples over the max level");
}

void Wave::CalculateAverageWithCorrection_PlusOne (int64& lvl, wave_data_t& lvl0, wave_data_t& lvl1, wave_data_t& lvl2, wave_data_t& lvl3)
{
// pseudo magnitude
//	wave_data_t pmag;
//	wave_data_t d;
//	wave_data_t lvl_min;
//	wave_data_t lvl_max;
//	double mag0;
//	double mag2;
//	double mag1;
#if 0
	double slope0;
	double slope1;
	double slope2;
	double bend0;
	double bend1;
#else
	int64 slope0;
	int64 slope1;
	int64 slope2;
	int64 bend0;
	int64 bend1;
#endif
	double correction;

//	mag0 = lvl1 - lvl0;
//	mag1 = lvl2 - lvl1;
//	mag2 = lvl3 - lvl2;
	slope0 = (int64)lvl1 - lvl0;
	slope1 = (int64)lvl2 - lvl1;
	slope2 = (int64)lvl3 - lvl2;

// left to right; bottom to top; all three
//	relation = mag0 / ;
// positive bend will mean, that the positive correction will be needed
	bend0 = slope0 - slope1;
	bend1 = slope1 - slope2;

// exactly 1/16 for low frequency peaks
// near 1/10 (it is "(2^0.5 - 1) / 4") on 1/4 of samplerate frequency
// exactly 1/6 on 1/3 of samplerate freq
	correction = (double)(bend0 + bend1);
// 1/14 I don't feel; 1/13, acts very slightly; I would give 1db as like by volume scale :)
// 1/11 feels with some significance
// 1/20 relatively to 1/16 is like 1-1.5 db of added distortion
//// so I think to left 1/12
//	correction *= 2.0/12.0;
// after one test I can say, that the 1/12 only adds some nonlinear distortion and not gives what I expected
// (expected that it will decrease "computer like" low-samplerate sounding effect)
// the mostly "deaf" sound seems is around 1/14 of correction, it slightly subtracts some distortion comparing with 1/16
	correction *= 2.0/14.0;
	lvl = lvl1;
	lvl += lvl2;
	lvl += (int64)correction;
	lvl >>= 1;

	CheckLevel (lvl);

// I messed; no any angle can be here
#if 0
//	double relation;
	double angle0;
	double angle1;

//// x   -   2 PI full period   - what period ? there is not period
// x   -   1 - distance between samples
// y   -   -1.0 - 1.0 full range

#if 0
	lvl_min = lvl0;
	lvl_max = lvl0;
	if (lvl_min > lvl1)
		lvl_min = lvl1;
	if (lvl_max < lvl1)
		lvl_max = lvl1;
	if (lvl_min > lvl2)
		lvl_min = lvl2;
	if (lvl_max < lvl2)
		lvl_max = lvl2;
	if (lvl_min > lvl3)
		lvl_min = lvl3;
	if (lvl_max < lvl3)
		lvl_max = lvl3;
	pmag = lvl_max - lvl_min;
#endif

	angle0 = slope1 - slope0;
	angle1 = slope2 - slope1;
#endif
}

// introduction :  slope - getting from two samples, bend - getting from two slopes, it are three samples

// 2^32 precision will need around 3^20 cycles
#define RESAMPLE_FULL__BUFFER_BENDS_NUM 20
// must be 1/3, but actually 2-2.5 sounds a bit better
//#define RESAMPLE_FULL__PART_FROM_PREVIOUS_BENDS (1.0 / 3.0)
//#define RESAMPLE_FULL__PART_FROM_PREVIOUS_BENDS (1.0 / 2.0)
#define RESAMPLE_FULL__PART_FROM_PREVIOUS_BENDS (1.0 / 2.5)
// by low freq condition; it is 1/16 of each (left and right) bend; and correct it, because previous bends will also be applied
#define RESAMPLE_FULL__THIS_BEND_ACTION ((RESAMPLE_FULL__PART_FROM_PREVIOUS_BENDS + 1.0) / 16.0)
//// doubtly this is one is necessary, doubtly that is possible easy
// perhaps relates only at the end, and not at the beginning
// don't know don't know, need to use some tool to define what happens where;
// ordinary playback programs may give own bugs, and can't define by them by hearing
#define RESAMPLE_FULL__REMOVE_ENDS_CLICKS 0

void Wave::Process_Samples_Double_Continuous ()
{
	uint64 samples_num_new;
	uint64 ix_in;
	uint64 ix_out;
	int i;
//	int j;
//	wave_data_t lvl0;
//	wave_data_t lvl1;
	wave_data_t lvl_new;
	int64 lvl;
// will be used as the left side value
	int64 lvl_last;
#if 0
	double slope;
// have to remember one last slope, it will be necessary for last bend calculation
	double slope_last;
#else
	int64 slope;
// have to remember one last slope, it will be necessary for last bend calculation
	int64 slope_last;
#endif
// no need to remember left side bends, only right side
// and these ones actually aren't bend; it are bends processing results
	double bend_proc_left;
	double bend_proc_right;
// remember left side bends' processing result
//	double bend_old;
// bends will go parallel with samples (by ix)
	double bends_right[RESAMPLE_FULL__BUFFER_BENDS_NUM];
// have to remember and rotate the right side samples, to avoid accessing input twice for the same data
	wave_data_t samples_right[RESAMPLE_FULL__BUFFER_BENDS_NUM];
	double test;

// test what actions the compiler would do about these values, did it get them as single number
	test = RESAMPLE_FULL__PART_FROM_PREVIOUS_BENDS;
	test = RESAMPLE_FULL__THIS_BEND_ACTION;

	Assert(m_samples_num);
	samples_num_new = m_samples_num << 1;
	if (! ForProcessing_SetSize (samples_num_new))
		return;

	bend_proc_left = 0.0;
// *float point values are 0, when they are 0
	memset (bends_right, 0, sizeof(bends_right));
// cleanup, so the first two bends would be calculated properly
	memset (samples_right, 0, sizeof(samples_right));
//	lvl0 = 0;
//	lvl1 = 0;
	ix_in = 0;
//	ix_in = 0 - RESAMPLE_FULL__BUFFER_BENDS_NUM;
	ix_out = 0;
//	ix_out = 0 - RESAMPLE_FULL__BUFFER_BENDS_NUM * 2;
//	samples_num_minus_one = m_samples_num - 1;
//// avoid extra cycle; init now
//	samples_right[RESAMPLE_FULL__BUFFER_BENDS_NUM - 1] = m_data[0];
//	lvl_new = m_data[0];
	slope_last = 0;
#if 0
//#if RESAMPLE_FULL__REMOVE_ENDS_CLICKS
	samples_right[RESAMPLE_FULL__BUFFER_BENDS_NUM - 1] = m_data[0] / 2;
	samples_right[RESAMPLE_FULL__BUFFER_BENDS_NUM - 1] = -m_data[0];
	samples_right[RESAMPLE_FULL__BUFFER_BENDS_NUM - 1] = -m_data[0] / 2;
#endif

	while (ix_out < samples_num_new)
	{
// for the new bend, which is for put on the stack, will be necessary :  last slope, two samples (one from stack, one new) -> new slope
//		lvl_new = m_data[0];
// get new sample
		if (ix_in < m_samples_num)
		{
// using of two variables for the in ix would be slower
//			samples_right[RESAMPLE_FULL__BUFFER_BENDS_NUM - 1] = m_data[ix_in - RESAMPLE_FULL__BUFFER_BENDS_NUM];
			lvl_new = m_data[ix_in];
//			ix_in++;
		}
		else
#if RESAMPLE_FULL__REMOVE_ENDS_CLICKS
		{
			lvl_new = 0;
			goto subtract_slopes;
		}
#else
//			samples_right[RESAMPLE_FULL__BUFFER_BENDS_NUM - 1] = 0;
			lvl_new = 0;
#endif
// slope goes next value up; positive means rising
		slope = (int64)lvl_new - samples_right[RESAMPLE_FULL__BUFFER_BENDS_NUM - 1];
//subtract_slopes:
// bend value goes by middle sample rise to the up side
		bends_right[RESAMPLE_FULL__BUFFER_BENDS_NUM - 1] = (double)(slope_last - slope);
		slope_last = slope;

//		if (ix_out < 0)
// skip until 'lvl_last' will get first value
//		if (ix_in < RESAMPLE_FULL__BUFFER_BENDS_NUM)
		if (ix_in < RESAMPLE_FULL__BUFFER_BENDS_NUM + 1)
// buffer is not full yet
			goto pump_stack;
//			continue;
//			goto next_ix_in;

// for the interpolated sample will be necessary :  two samples, the correction from two results of bendings continuous processing
//		bend_proc_right = 0.0;
		bend_proc_right = bends_right[RESAMPLE_FULL__BUFFER_BENDS_NUM - 1];
		for (i = RESAMPLE_FULL__BUFFER_BENDS_NUM - 2; i >= 0; i--)
		{
			bend_proc_right *= -RESAMPLE_FULL__PART_FROM_PREVIOUS_BENDS;
			bend_proc_right += bends_right[i];
		}

		m_data__in_process[ix_out] = lvl_last;
		ix_out++;
// *currently always 1
		if (ix_out < samples_num_new)
		{
// get average ariphmetical
			lvl = lvl_last + samples_right[0];
			lvl >>= 1;
// and correct it
			lvl += (bend_proc_left + bend_proc_right) * RESAMPLE_FULL__THIS_BEND_ACTION;
			CheckLevel (lvl);
			m_data__in_process[ix_out] = lvl;
			ix_out++;
		}

// can't rotate stack right after new sample value pick up, because then the last value in the bends row will not be defined yet;
// and a new bend always require one more new sample
pump_stack:
// update (rotate) input values row
//		lvl0 = lvl1;
//		lvl1 = samples_right[0];
		lvl_last = samples_right[0];
		bend_proc_left *= -RESAMPLE_FULL__PART_FROM_PREVIOUS_BENDS;
		bend_proc_left += bends_right[0];
		for (i = 0; i < RESAMPLE_FULL__BUFFER_BENDS_NUM - 1; i++)
		{
			samples_right[i] = samples_right[i + 1];
			bends_right[i] = bends_right[i + 1];
		}
#if 0
// get new sample
		if (ix_in < m_samples_num)
		{
// using of two variables for the in ix would be slower
//			samples_right[RESAMPLE_FULL__BUFFER_BENDS_NUM - 1] = m_data[ix_in - RESAMPLE_FULL__BUFFER_BENDS_NUM];
			lvl_new = m_data[ix_in];
			ix_in++;
		}
		else
//			samples_right[RESAMPLE_FULL__BUFFER_BENDS_NUM - 1] = 0;
			lvl_new = 0;
#endif
		samples_right[RESAMPLE_FULL__BUFFER_BENDS_NUM - 1] = lvl_new;

//next_ix_in:
		ix_in++;
	}
//	Assert(ix_in - RESAMPLE_FULL__BUFFER_BENDS_NUM == m_samples_num);
	Assert(ix_in - (RESAMPLE_FULL__BUFFER_BENDS_NUM + 1) == m_samples_num);
}

// +++ from external sources +++

// is taken from Miranda; and there is from FreeImage

/*
This code was taken and adapted from the following reference : 

[1] Philippe Th?venaz, Spline interpolation, a C source code 
implementation. http://bigwww.epfl.ch/thevenaz/
*/

// Given an array of spline coefficients,
// return the value of the underlying continuous spline model, sampled at the location pos.
// pos must be in -0.5 - +0.5 range near middle of given values row; it is 0.5 - 1.5 for 2nd degree
// The model degree can be 2 (quadratic), 3 (cubic), 4 (quartic), or 5 (quintic).
//static double InterpolatedValue (double* coeff, uint length, double pos, int spline_degree)
static double InterpolatedValue (wave_data_t* coeff, uint length, double pos, uint spline_degree)
{
//	double* p;
	double weights[6];
	double interpolated;
	double w, w2, w4, t, t0, t1;
	uint i;
	int central_sample;

	Assert(length >= spline_degree + 1);
#if 1
// used must be this range; make sure in a caller
// so the beginning sample ix will be 0
	double test;
//	test = pos - (double)length / 2;
//	test = pos - central_sample;
	test = pos - spline_degree / 2.0;
	Assert(test >= -0.5);
	Assert(test <= 0.5);
#else
//	Assert(pos >= spline_degree / 2.0);
//	Assert(pos <= length - spline_degree / 2.0);
#endif

#if 0
	// compute the interpolation indexes
	if (spline_degree & 1L)
	{
		i = (long)floor(x) - spline_degree / 2L;
		j = (long)floor(y) - spline_degree / 2L;
		for(k = 0; k <= spline_degree; k++) {
			indices[k] = i++;
			yIndex[k] = j++;
		}
	}
	else
	{
		i = (long)floor(x + 0.5) - spline_degree / 2L;
		j = (long)floor(y + 0.5) - spline_degree / 2L;
		for (k = 0; k <= spline_degree; k++) {
			indices[k] = i++;
			yIndex[k] = j++;
		}
	}
#endif
//	central_sample = (int)(length + pos) / 2;
	central_sample = (int)(pos + 0.5);

// compute the interpolation weights
	w = pos - central_sample;
// test
//	w = -w;
// recheck that the value is ok
// negative w seems ok; and positive seems can be up to 1.0
	Assert(w >= -0.5);
//	Assert(w <= 1.0);
//	Assert(w <= 0.5);
	Assert(w < 0.5);
	switch (spline_degree)
	{
		case 2:
			Assert(central_sample == 1);
			weights[1] = 3.0 / 4.0 - w * w;
			weights[2] = (1.0 / 2.0) * (w - weights[1] + 1.0);
			weights[0] = 1.0 - weights[1] - weights[2];
			break;
		case 3:
			Assert(central_sample == 1);
			weights[3] = (1.0 / 6.0) * w * w * w;
			weights[0] = (1.0 / 6.0) + (1.0 / 2.0) * w * (w - 1.0) - weights[3];
			weights[2] = w + weights[0] - 2.0 * weights[3];
			weights[1] = 1.0 - weights[0] - weights[2] - weights[3];
			break;
		case 4:
			Assert(central_sample == 2);
			w2 = w * w;
			t = (1.0 / 6.0) * w2;
			weights[0] = 1.0 / 2.0 - w;
			weights[0] *= weights[0];
			weights[0] *= (1.0 / 24.0) * weights[0];
			t0 = w * (t - 11.0 / 24.0);
			t1 = 19.0 / 96.0 + w2 * (1.0 / 4.0 - t);
			weights[1] = t1 + t0;
			weights[3] = t1 - t0;
			weights[4] = weights[0] + t0 + (1.0 / 2.0) * w;
			weights[2] = 1.0 - weights[0] - weights[1] - weights[3] - weights[4];
			break;
		case 5:
			Assert(central_sample == 2);
			w2 = w * w;
			weights[5] = (1.0 / 120.0) * w * w2 * w2;
			w2 -= w;
			w4 = w2 * w2;
			w -= 1.0 / 2.0;
			t = w2 * (w2 - 3.0);
			weights[0] = (1.0 / 24.0) * (1.0 / 5.0 + w2 + w4) - weights[5];
			t0 = (1.0 / 24.0) * (w2 * (w2 - 5.0) + 46.0 / 5.0);
			t1 = (-1.0 / 12.0) * w * (t + 4.0);
			weights[2] = t0 + t1;
			weights[3] = t0 - t1;
			t0 = (1.0 / 16.0) * (9.0 / 5.0 - t);
			t1 = (1.0 / 24.0) * w * (w4 - w2 - 5.0);
			weights[1] = t0 + t1;
			weights[4] = t0 - t1;
			break;
		default:
//			Console_ProgramError ("InterpolatedValue () :  invalid spline degree");
			Error ("InterpolatedValue () :  invalid spline degree");
			return 0;
	}

// perform interpolation
	interpolated = 0.0;
	for (i = 0; i <= spline_degree; i++)
	{
		interpolated += weights[i] * coeff[i];
	}

	return interpolated;
}

// from Audacity

float CubicInterpolate (float y0, float y1, float y2, float y3, float x)
{
   float a, b, c, d;

   a = y0 / -6.0 + y1 / 2.0 - y2 / 2.0 + y3 / 6.0;
   b = y0 - 5.0 * y1 / 2.0 + 2.0 * y2 - y3 / 2.0;
   c = -11.0 * y0 / 6.0 + 3.0 * y1 - 3.0 * y2 / 2.0 + y3 / 3.0;
   d = y0;

   float xx = x * x;
   float xxx = xx * x;

   return (a * xxx + b * xx + c * x + d);
}

// --- from external sources ---

#define RESAMPLE_POWER_MAX 5

void Wave::Process_Samples_Double_Power_Some (int spline_degree)
{
	uint64 samples_num_new;
	uint64 ix_in;
	uint64 ix_out;
	int i;
	wave_data_t lvl_new;
	int64 lvl;
	wave_data_t samples_row[RESAMPLE_POWER_MAX + 1];
	double d_removal;
	int i_removal0;
	int i_removal1;
	double val;

	Assert(spline_degree >= 2);
	Assert(spline_degree <= RESAMPLE_POWER_MAX);
	Assert(m_samples_num);
	samples_num_new = m_samples_num << 1;
	if (! ForProcessing_SetSize (samples_num_new))
		return;

// satisfy conditions in the calculating subroutine
//	d_removal = spline_degree / 2.0;
	i_removal0 = spline_degree / 2;
//	d_removal = i_removal0 + 0.5;
	d_removal = i_removal0 + 0.49999999;
	i_removal1 = spline_degree + 1 - i_removal0;

	memset (samples_row, 0, sizeof(samples_row));
	ix_in = 0;
//	ix_in = 0 - RESAMPLE_FULL__BUFFER_BENDS_NUM;
	ix_out = 0;
//	ix_out = 0 - RESAMPLE_FULL__BUFFER_BENDS_NUM * 2;

	while (ix_out < samples_num_new)
	{
// get new sample
		if (ix_in < m_samples_num)
		{
// using of two variables for the in ix would be slower
			lvl_new = m_data[ix_in];
		}
		else
			lvl_new = 0;

// skip until the first value will appear at 'i_removal' ix
		if (ix_in < i_removal1)
// buffer is not full yet
			goto pump_stack;

		m_data__in_process[ix_out] = samples_row[i_removal0];
		ix_out++;
// *currently always 1
		if (ix_out < samples_num_new)
		{
#if 0
// wrong to put this condition here of course, at least for these necessaries
//			if (spline_degree != 3)
			{
				val = InterpolatedValue (samples_row, sizeof(samples_row), d_removal, spline_degree);
				lvl = val;
			}
//			else
#else
			{
//				val = CubicInterpolate (samples_row[i_removal0 - 1], samples_row[i_removal0], samples_row[i_removal0 + 1], samples_row[i_removal0 + 2], 0.5);
				val = CubicInterpolate (samples_row[i_removal0 - 1], samples_row[i_removal0], samples_row[i_removal0 + 1], samples_row[i_removal0 + 2], 1.5);
//				val = CubicInterpolate (samples_row[i_removal0], samples_row[i_removal0 + 1], samples_row[i_removal0 + 2], samples_row[i_removal0 + 3], 0.5);
				lvl = val;
			}
#endif
			CheckLevel (lvl);
			m_data__in_process[ix_out] = (wave_data_t)lvl;
			ix_out++;
		}

pump_stack:
// update (rotate) input values row
		for (i = 0; i < spline_degree; i++)
		{
			samples_row[i] = samples_row[i + 1];
		}
		samples_row[spline_degree] = lvl_new;

		ix_in++;
	}
	Assert(ix_in - i_removal1 == m_samples_num);
}

// now I think, that those 'extend_xxxx' was enough bad idea :) ; must be extended from the end side, and that's whole story :)
void Wave::Process_Samples_Double (int resample_type, bool uprate, bool extend_begin, bool extend_end)
{
	uint64 samples_num_new;
	uint64 i;
	uint64 j;
	int64 lvl_cur;
	int64 lvl_next;
	int64 lvl;
	uint64 samples_num_minus_one;
	bool result;

	m_was_level_cuts = 0;

	switch (resample_type)
	{
	case WAVE_RESAMPLE_TYPE__DUPLICATE:
		goto process_duplicate;
	case WAVE_RESAMPLE_TYPE__AVERAGE:
		break;
	case WAVE_RESAMPLE_TYPE__BENDS_PLUS_ONE:
		goto process_plus_one;
	case WAVE_RESAMPLE_TYPE__BENDS_CONTINUOUS:
		Process_Samples_Double_Continuous ();
		goto finish;
	case WAVE_RESAMPLE_TYPE__POWER_SOME:
//		Process_Samples_Double_Power_Some (2);
		Process_Samples_Double_Power_Some (5);
		goto finish;
	default:
		Console_ProgramError ("Wave::Process_Samples_Double () :  'resample_type' %d is unknown");
		return;
	}

// "point" wave model
	Assert(m_samples_num);
	samples_num_new = (m_samples_num << 1) + extend_begin + extend_end - 1;
	if (! ForProcessing_SetSize (samples_num_new))
		return;

//	for (i = 0, j = 0; i < m_samples_num; i++)
	i = 0;
	j = 0;
	samples_num_minus_one = m_samples_num - 1;

	if (extend_begin)
	{
// average with 0
		m_data__in_process[j] = m_data[i] >> 1;
//		m_data__in_process[j] = average_level ? m_data[i] >> 1 : m_data[i] >> 2;
		j++;
	}

	lvl_next = m_data[i];
	while (i < samples_num_minus_one)
	{
// *and not actually -1, but -2 or -3, because this loop does not works until end
		Assert(j < samples_num_new - 1);
		lvl_cur = lvl_next;
		i++;
		lvl_next = m_data[i];
		lvl = lvl_cur;
		m_data__in_process[j] = lvl;
		j++;
		lvl += lvl_next;
		lvl >>= 1;
		m_data__in_process[j] = lvl;
		j++;
	}

	Assert(j < samples_num_new);
	lvl = lvl_next;
	m_data__in_process[j] = lvl;
	j++;
	if (extend_end)
	{
		Assert(j < samples_num_new);
		lvl >>= 1;
		m_data__in_process[j] = lvl;
		j++;
	}

	goto finish;

// "flat" wave model
process_duplicate:
	Assert(m_samples_num);
	samples_num_new = m_samples_num << 1;
	if (! ForProcessing_SetSize (samples_num_new))
		return;

	i = 0;
	j = 0;

	while (i < m_samples_num)
	{
		Assert(j < samples_num_new);
		lvl = m_data[i];
		m_data__in_process[j] = lvl;
		j++;
		m_data__in_process[j] = lvl;
		j++;
		i++;
	}

	goto finish;

// fast, and intended to be sound much better than by the average ariphmetical method
// for processing is took ((1 + 1) * 2) samples instead of (1 * 2)
// those additional samples will help to make correction, so the output would be much more close to real resample
// practical results :  with 1/12 correction, no sensitive difference between "0.7, 0.7, -0.7, -0.7" and "0, 1, 0, -1" (4 samples per period);
// from 375 rate to 48000 (7 times) resampled sounds
process_plus_one:
	wave_data_t lvl0;
	wave_data_t lvl1;
	wave_data_t lvl2;
	wave_data_t lvl3;

	Assert(m_samples_num);
	samples_num_new = (m_samples_num << 1) + extend_begin + extend_end - 1;
	if (! ForProcessing_SetSize (samples_num_new))
		return;

	lvl0 = 0;
	lvl1 = 0;
	i = 0;
	j = 0;
	samples_num_minus_one = m_samples_num - 1;

	if (i < m_samples_num)
	{
		lvl2 = m_data[i];
		i++;
		if (i < m_samples_num)
		{
			lvl3 = m_data[i];
			i++;
		}
	}
// **not exact
	if (extend_begin)
	{
// average with 0
		m_data__in_process[j] = m_data[i] >> 1;
//		m_data__in_process[j] = average_level ? m_data[i] >> 1 : m_data[i] >> 2;
		j++;
	}
	while (j < samples_num_new)
	{
// update (rotate) input values row
		lvl0 = lvl1;
		lvl1 = lvl2;
		lvl2 = lvl3;
		if (i < m_samples_num)
		{
			lvl3 = m_data[i];
			i++;
		}
		else
			lvl3 = 0;

		m_data__in_process[j] = lvl1;
		j++;
		if (j < samples_num_new)
		{
			CalculateAverageWithCorrection_PlusOne (lvl, lvl0, lvl1, lvl2, lvl3);
			m_data__in_process[j] = lvl;
			j++;
		}
	}

	goto finish;

finish:
	ForProcessing_ReplaceOld ();
	if (m_was_level_cuts)
		Console_Warning ("Wave::Process_Samples_Double () :  the wave had level cuts");
	if (uprate)
	{
// get highest bit
		result = !! (m_sample_rate >> (sizeof(m_sample_rate) * 8 - 1));
		if (result)
			Console_Warning ("Wave::Process_Samples_Double () :  the wave samplerate value is %08x (%u), and is at the 32bit value limt, and can not be doubled", m_sample_rate, m_sample_rate);
		else
			m_sample_rate <<= 1;
		Console_Notify ("%d samplerate %u", m_ix, m_sample_rate);
	}

	Console_Notify ("wave samplerate doubled");

	return;
}

// "point" wave model
void Wave::Process_Resample_Up (uint cycle_size_out, uint cycle_size_in, bool change_rate)
{
//	bool result;
	uint64 samples_num_new;
// how much input samples to take, to produce one output sample
	fix32 quant_size;
//	fix32 quant_reminder;
	fix32 sample_frac;
// from here a cycle begins; a cycle base
//	uint base_input;
	uint base_output;
// offset relatively to base; now is necessary only for find out a cycle end
	fix32 offset_input;
	int i_offset_input;
//	fix32 offset_input_next;
	uint offset_output;
// also for input offset, but this one is integer, and with base added
	uint pos_input;
	uint pos_output;
	fix64 lvl0;
	fix64 lvl1;
//	fix64 lvl_in;
	fix64 lvl_out;

	Assert(cycle_size_in);
	Assert(cycle_size_out > cycle_size_in);
	Assert(m_samples_num);
	samples_num_new = m_samples_num * cycle_size_out;
	samples_num_new /= cycle_size_in;
	if (! ForProcessing_SetSize (samples_num_new))
		return;

	quant_size = fix32div(cycle_size_in, cycle_size_out);
//	base_input = 0;
	base_output = 0;
	pos_input = 0;
	pos_output = 0;

// cycle, until both offsets will become integer again
loop_cycle:
	offset_input = 0;
	offset_output = 0;
//	offset_input_next = 0;

// single input sample, to produce single cluster of output samples; word "quant" used to show that cluster size
loop_cluster:

	Assert(pos_input < m_samples_num);
	lvl0 = m_data[pos_input];
	pos_input++;
// "point" model, but last sample will also have it's time; so make next sample zero, when will be at the end
	if (pos_input < m_samples_num)
		lvl1 = m_data[pos_input];
	else
		lvl1 = 0;
	i_offset_input = FIX32_TO_INT(offset_input);

// inside of one input sample period
//	while (FIX32_TO_INT(offset_input) == i_offset_input)
	do
	{
		sample_frac = fix32frac(offset_input);
		lvl_out = fix64mul32(lvl0, F32_1_0 - sample_frac);
		lvl_out += fix64mul32(lvl1, sample_frac);
//		pos_output = base_output + offset_output;
		m_data__in_process[pos_output] = lvl_out;
		offset_output++;
		pos_output++;
// *it may make a bit smaller step than necessary
		offset_input += quant_size;
		if (pos_output >= samples_num_new)
			goto finish;
		if (offset_output >= cycle_size_out)
// next cycle
		{
			base_output += cycle_size_out;
			goto loop_cycle;
		}
	}
	while (FIX32_TO_INT(offset_input) == i_offset_input);

//	if (offset_input < cycle_size_in)
//	if (offset_output < cycle_size_out)
// next cluster
	goto loop_cluster;

//	Assert(offset_output == cycle_size_out);
//	base_output += cycle_size_out;
//	goto loop_cycle;

finish:
//	int64 lost_num;
// rough
//	lost_num = m_samples_num - pos_input - 1;
//	if (lost_num > 0)
//		Console_Warning ("%I64u input samples have been completely skipped at the end of the input wave", lost_num);
	Assert(pos_output == samples_num_new);
	ForProcessing_ReplaceOld ();
	if (change_rate)
	{
		uint64 sample_rate_new;

		sample_rate_new = m_sample_rate * cycle_size_out;
		sample_rate_new /= cycle_size_in;
		m_sample_rate = sample_rate_new;
//		if (m_sample_rate != sample_rate_new)
		if (sample_rate_new >= 0x100000000)
		{
			Console_Warning ("Wave::Process_Resample_Up () :  sample rate is maximal");
			m_sample_rate = 0xffffffff;
		}
	}
	Console_Notify ("wave resampled");
}

// slower but more exact
#define RESAMPLE_USES_MULTIPLY_FOR_OFFSET 1

// basing on using of "vulgar fraction" (or other name - "common fraction") mechanism, which minimally spoils a sound
// "flat" wave model
void Wave::Process_Resample_Down (uint cycle_size_out, uint cycle_size_in, bool change_rate)
{
	uint64 samples_num_new;
// how much input samples to take, to produce one output sample
	fix32 quant_size;
	fix32 quant_reminder;
// part from first sample in this cluster
//	fix32 part_begin;
// part from last sample in this cluster
//	fix32 part_end;
	fix32 sample_frac;
// number of full samples in this cluster
//	uint fulls_num;
// from here a cycle begins; a cycle base
	uint base_input;
	uint base_output;
// offset relatively to base
	fix32 offset_input;
	fix32 offset_input_next;
	uint offset_output;
// also for input offset, but this one is integer, and with base added
	uint pos_input;
	uint pos_output;
	fix64 lvl_in;
	fix64 lvl_out;

	Assert(cycle_size_out);
	Assert(cycle_size_out < cycle_size_in);
	Assert(m_samples_num);
	samples_num_new = m_samples_num * cycle_size_out;
	samples_num_new /= cycle_size_in;
	if (! ForProcessing_SetSize (samples_num_new))
		return;

	quant_size = fix32div(cycle_size_in, cycle_size_out);
	base_input = 0;
	base_output = 0;
	offset_input = 0;

// cycle, until both offsets will become integer again
loop_cycle:
	offset_output = 0;
	offset_input_next = 0;

// single cluster of input samples, to produce single output sample; word "quant" used to show that cluster size
loop_cluster:
	offset_input = offset_input_next;
#if RESAMPLE_USES_MULTIPLY_FOR_OFFSET
// *also would be possible to use add instead, but that would be less exact perhaps, and not much faster I think
//	offset_input = fix32mul(offset_output, quant_size);
//	offset_input_next = fix32mul(offset_output + 1, quant_size);
//	offset_input_next = (offset_output + 1) * quant_size;
	offset_input_next = (offset_output + 1) * cycle_size_in;
	offset_input_next = fix32div(offset_input_next, cycle_size_out);
	quant_size = offset_input_next - offset_input;
#else
// this way perhaps will be possible sensitive decline, when a cycle will have much output samples
	offset_input_next += quant_size;
#endif

	quant_reminder = quant_size;
	lvl_out = 0;
	pos_input = base_input + FIX32_TO_INT(offset_input);

// == part 1 :  add partial first sample ==
	Assert(pos_input < m_samples_num);
	sample_frac = fix32frac(offset_input);
	if (! sample_frac)
		goto part_2;
//	part_begin = F32_1_0 - sample_frac;
	sample_frac = F32_1_0 - sample_frac;
	lvl_in = m_data[pos_input];
	pos_input++;
	lvl_out = fix64mul32(lvl_in, sample_frac);
	quant_reminder -= sample_frac;

// == part 2 :  add complete samples ==
part_2:
	while (quant_reminder >= F32_1_0)
	{
		Assert(pos_input < m_samples_num);
		lvl_in = m_data[pos_input];
		pos_input++;
		lvl_out += lvl_in;
		quant_reminder -= F32_1_0;
	}

// == part 3 :  add partial last sample ==
	if (! quant_reminder)
		goto put_value;
	Assert(pos_input < m_samples_num);
//	sample_frac = fix32frac(quant_reminder);
//	part_end = sample_frac;
	lvl_in = m_data[pos_input];
	lvl_out += fix64mul32(lvl_in, quant_reminder);

put_value:
	pos_output = base_output + offset_output;
	lvl_out = fix64div32(lvl_out, quant_size);
	m_data__in_process[pos_output] = lvl_out;

#if ! RESAMPLE_USES_MULTIPLY_FOR_OFFSET
// this way perhaps will be possible sensitive decline, when a cycle will have much output samples
//	offset_input += quant_size;
//	offset_input_next += quant_size;
#endif
	offset_output++;
	pos_output = base_output + offset_output;
	if (pos_output >= samples_num_new)
		goto finish;
	if (offset_output < cycle_size_out)
		goto loop_cluster;

// here integer step is possible
	base_input += cycle_size_in;
	base_output += cycle_size_out;
	goto loop_cycle;

finish:
	int64 lost_num;
// rough
	lost_num = m_samples_num - pos_input - 1;
	if (lost_num > 0)
		Console_Warning ("%I64u input samples have been completely skipped at the end of the input wave", lost_num);
	ForProcessing_ReplaceOld ();
	if (change_rate)
	{
		uint sample_rate_new;

		sample_rate_new = m_sample_rate * cycle_size_out;
		sample_rate_new /= cycle_size_in;
		m_sample_rate = sample_rate_new;
		if (! sample_rate_new)
		{
			Console_Warning ("Wave::Process_Samples_Double () :  sample rate is minimal");
			m_sample_rate = 1;
		}
	}
	Console_Notify ("wave resampled");
}

void Wave::Process_Resample (uint cycle_size_out, uint cycle_size_in, bool change_rate)
{
	if (cycle_size_out == cycle_size_in)
	{
		Console_Error ("Wave::Process_Resample () :  output samplerate is equal to the input one; exiting");
		return;
	}
	if (! cycle_size_out)
	{
		Console_Error ("Wave::Process_Resample () :  output cycle size is zero; exiting");
		return;
	}
	if (! cycle_size_in)
	{
		Console_Error ("Wave::Process_Resample () :  input cycle size is zero; exiting");
		return;
	}

	if (cycle_size_out > cycle_size_in)
		Process_Resample_Up (cycle_size_out, cycle_size_in, change_rate);
	else
		Process_Resample_Down (cycle_size_out, cycle_size_in, change_rate);

	if (change_rate)
		Console_Notify ("%d samplerate %u", m_ix, m_sample_rate);
}

void Wave::Samples_Insert (uint64 offset, uint64 amount)
{
	uint64 samples_num_new;
	uint64 offset2;
	uint64 size0;
	uint64 size1;
	uint64 size2;

	if (offset == -1)
		offset = m_samples_num;
	if (offset > m_samples_num)
	{
		Console_Warning ("Samples_Insert () :  offset %llu is higher than samples number %llu", offset, m_samples_num);
		return;
	}
	samples_num_new = m_samples_num + amount;
	if (! ForProcessing_SetSize (samples_num_new))
		return;
	size0 = offset * WAVE_DATA_SIZEOF;
	size1 = amount * WAVE_DATA_SIZEOF;
	offset2 = offset + amount;
	size2 = (m_samples_num - offset) * WAVE_DATA_SIZEOF;
	memcpy (m_data__in_process, m_data, size0);
	memset (m_data__in_process + offset, 0, size1);
	memcpy (m_data__in_process + offset2, m_data + offset, size2);
	ForProcessing_ReplaceOld ();
}

void Wave::Samples_Remove (uint64 offset, uint64 amount)
{
	uint64 samples_num_new;
	uint64 offset2;
	uint64 size0;
//	uint64 size1;
	uint64 size2;

	if (offset > m_samples_num)
	{
		Console_Warning ("Samples_Remove () :  offset %llu is higher than samples number %llu", offset, m_samples_num);
		return;
	}
	if (amount == -1)
		amount = m_samples_num - offset;
	samples_num_new = m_samples_num - amount;
	if (! ForProcessing_SetSize (samples_num_new))
		return;
	size0 = offset * WAVE_DATA_SIZEOF;
//	size1 = amount * WAVE_DATA_SIZEOF;
	offset2 = offset + amount;
	size2 = (samples_num_new - offset) * WAVE_DATA_SIZEOF;
	memcpy (m_data__in_process, m_data, size0);
	memcpy (m_data__in_process + offset, m_data + offset2, size2);
	ForProcessing_ReplaceOld ();
}

// 'quant_out'/'quant_in' were meaningless, so renamed
void Wave::Process_Amplify (uint quant_multiply, uint quant_divide)
{
	uint64 offset;
	int64 val;
//	bool had_cuts = 0;

	if (quant_multiply == quant_divide)
	{
		Console_Error ("Wave::Process_Amplify () :  output quant is equal to the input one; exiting");
		return;
	}
	if (! quant_multiply)
	{
		Console_Error ("Wave::Process_Amplify () :  multiply quant is zero; exiting");
		return;
	}
	if (! quant_divide)
	{
		Console_Error ("Wave::Process_Amplify () :  divide quant is zero; exiting");
		return;
	}

	m_was_level_cuts = 0;
	for (offset = 0; offset < m_samples_num; offset++)
	{
		val = m_data[offset];
		val *= quant_multiply;
		val /= quant_divide;
		CheckLevel (val);
		m_data[offset] = (wave_data_t)val;
//		had_cuts = 1;
	}
	if (m_was_level_cuts)
		Console_Error ("Wave::Process_Amplify () :  wave had level cuts");
	Mipmap_Free ();
}

// --- process ---

// other processing subroutines, which do not go into the Process menu

// +++ process 2 +++

void Wave::FillRemainderWithZeroes (uint64 from)
{
	uint64 i;

	for (i = from; i < m_samples_num; i++)
		m_data[i] = 0;
}

#define D_ULONG_MAX (256.0 * 256.0 * 256.0 * 256.0)

// only integer frequency atm
void Wave::GenerateSine (double d_amplitude, double d_level_offset, double d_frequency, double d_phase_beginning)
{
	wave_data_t level;
	double d_level;
	double d_level_multiplier;
	uint64 sample_ix;
	fix64 phase;
	double d_phase;
	int frequency;

	CalculateSubParameters ();
	d_level_multiplier = ((int64)WAVE_LEVEL_MAX - WAVE_LEVEL_MIN) / 2;
//	d_level_multiplier *= d_amplitude;
// hmmm, again 64 bit are not enough :)
	frequency = d_frequency;

	sample_ix = 0;
loop:
	if (sample_ix >= m_samples_num)
		goto finish;
// moment
	phase = m_sample_time * sample_ix;
// phase
	phase *= frequency;
// remove integer part, only frac is necessary
//	phase = (ulong)phase;
	d_phase = (ulong)phase;
	d_phase /= D_ULONG_MAX;
	d_level = sin (d_phase);
	d_level *= d_amplitude;
	d_level += d_level_offset;
	d_level *= d_level_multiplier;
	level = d_level;
	m_data[sample_ix] = level;
	sample_ix++;
	goto loop;

finish:
	return;
}

// constance of 1.0 means the signal will remain as was
// double float arithmetics atm
// (this filter actually is of 2nd power !)
// k for the lower than samplerate frequences is :
// k = 2 * PI * freq_filter_k
// on higher frequences the k seems becomes nonlinear, but doubtly such filter could be useful there
void Wave::Filter (Wave* wave_low, Wave* wave_high, double d_constance)
{
	uint64 sample_ix;
//	wave_data_t level;
	double level;
// low frequence
//	wave_data_t level_low;
	double level_low;
// high frequence
//	wave_data_t level_high;
	double level_high;
	bool proceed_low;
	bool proceed_high;
	double previous_part;

// no assert on constance, let user do a crap
	wxASSERT(wave_low || wave_high);
	if (wave_low)
	{
		if (wave_low->m_samples_num < m_samples_num)
			Console_Notify ("Wave::Filter () :  the low frequence wave has less samples, %I64u < %I64u", wave_low->m_samples_num, m_samples_num);
		if (wave_low->m_samples_num > m_samples_num)
			Console_Notify ("Wave::Filter () :  the low frequence wave has more samples, %I64u > %I64u", wave_low->m_samples_num, m_samples_num);
	}
	if (wave_high)
	{
		if (wave_high->m_samples_num < m_samples_num)
			Console_Notify ("Wave::Filter () :  the high frequence wave has less samples, %I64u < %I64u", wave_high->m_samples_num, m_samples_num);
		if (wave_high->m_samples_num > m_samples_num)
			Console_Notify ("Wave::Filter () :  the high frequence wave has more samples, %I64u > %I64u", wave_high->m_samples_num, m_samples_num);
	}
	proceed_low = !! wave_low;
	proceed_high = !! wave_high;
	previous_part = 1.0 - d_constance;
	level_low = 0;

	sample_ix = 0;
loop:
	if (sample_ix >= m_samples_num)
		goto finish;
	if (! proceed_low && ! proceed_high)
		goto finish;
	level = m_data[sample_ix];
	level_low = level_low * previous_part;
	level_low += level * d_constance;
	if (proceed_low)
	{
		if (sample_ix >= wave_low->m_samples_num)
			proceed_low = 0;
		else
			wave_low->m_data[sample_ix] = level_low;
	}
	level_high = level - level_low;
	if (proceed_high)
	{
		if (sample_ix >= wave_high->m_samples_num)
			proceed_high = 0;
		else
			wave_high->m_data[sample_ix] = level_high;
	}
	sample_ix++;
	goto loop;

finish:
// if some samples in output waves remain, then fill it with zeroes
	if (wave_low)
		wave_low->FillRemainderWithZeroes (sample_ix);
	if (wave_high)
		wave_high->FillRemainderWithZeroes (sample_ix);
}

// --- process 2 ---

// will be no need to take values directly from 'm_data',
// because if calculate intermediate mipmap layers, then it will not be sensitively slower;
// but only positive effect will be, that all intermediate levels will be already calculated

// +++ mipmap +++

//void Wave::Mipmap_InvalidateLevels ()
//{
//	memset (m_is_mipmap_level_valid, 0, sizeof(m_is_mipmap_level_valid));
//}

void Wave::Mipmap_Level_Allocate (int lvl)
{
	uint64 level_size;
	int lvl_previous;

//	Assert(lvl > 0);
	if (! lvl)
	{
		Assert(m_samples_num);
		m_mipmap_level_size[lvl] = m_samples_num;
// always already allocated; although the data is not in mipmap array of data, but in 'm_data'
		m_is_mipmap_level_allocated[lvl] = 1;
		return;
	}
	lvl_previous = lvl - 1;

	Assert(! m_is_mipmap_level_allocated[lvl]);
// avoid this one also; so it should be 0 when no mem allocated
	Assert(! m_is_mipmap_level_filled[lvl]);
	Assert(! m_mipmap_level_size[lvl]);
	Assert(! m_data_mipmap_min[lvl]);
	Assert(! m_data_mipmap_avg[lvl]);
	Assert(! m_data_mipmap_max[lvl]);
	Assert(m_samples_num);

// check previous level
	if (! m_is_mipmap_level_allocated[lvl_previous])
		Mipmap_Level_Allocate (lvl_previous);

// size
	level_size = m_mipmap_level_size[lvl_previous];
// cannot be ! always will be at least 1, even if the level is too high
//	if (! level_size)
//		return;
	Assert(level_size);
	level_size -= 1;
	level_size >>= WAVE_MIPMAP_LEVEL_INCREMENT_BITS;
	level_size += 1;

// cannot be !
//	if (! level_size)
//		return;
	m_data_mipmap_min[lvl] = (wave_data_t*)malloc (level_size * WAVE_DATA_SIZEOF);
	m_data_mipmap_avg[lvl] = (wave_data_t*)malloc (level_size * WAVE_DATA_SIZEOF);
	m_data_mipmap_max[lvl] = (wave_data_t*)malloc (level_size * WAVE_DATA_SIZEOF);
	m_is_mipmap_level_allocated[lvl] = 1;
	m_mipmap_level_size[lvl] = level_size;
}

void Wave::Mipmap_Level_Free (int lvl)
{
	int lvl_previous;

//	Assert(lvl > 0);
	if (! lvl)
	{
		m_mipmap_level_size[lvl] = 0;
		m_is_mipmap_level_filled[lvl] = 0;
// no freeing, it is not our data in this case; just mark ad not allocated
		m_is_mipmap_level_allocated[lvl] = 0;
		return;
	}
	lvl_previous = lvl - 1;
	Mipmap_Level_Free (lvl_previous);
// this is not mandatory; it can be not allocated
//	Assert(m_is_mipmap_level_allocated[lvl]);
	if (! m_is_mipmap_level_allocated[lvl])
		return;
	Assert(m_data_mipmap_min[lvl]);
	Assert(m_data_mipmap_avg[lvl]);
	Assert(m_data_mipmap_max[lvl]);

	free (m_data_mipmap_min[lvl]);
	free (m_data_mipmap_avg[lvl]);
	free (m_data_mipmap_max[lvl]);
	m_data_mipmap_min[lvl] = 0;
	m_data_mipmap_avg[lvl] = 0;
	m_data_mipmap_max[lvl] = 0;
// reset now, so this value can be used without additional check of 'm_is_mipmap_level_allocated[]'
	m_is_mipmap_level_filled[lvl] = 0;
	m_is_mipmap_level_allocated[lvl] = 0;
	m_mipmap_level_size[lvl] = 0;
}

// must automatically request allocation when necessary
void Wave::Mipmap_Level_Fill (int lvl)
{
	bool result;
	uint64 level_size;
	uint64 i;
	int lvl_previous;
	wave_data_t* levels_min;
	wave_data_t* levels_avg;
	wave_data_t* levels_max;
	wave_data_t lmin;
	wave_data_t lavg;
	wave_data_t lmax;

//	Assert(lvl > 0);
	Assert(! m_is_mipmap_level_filled[lvl]);
	if (! m_is_mipmap_level_allocated[lvl])
		Mipmap_Level_Allocate (lvl);
	Assert(m_is_mipmap_level_allocated[lvl]);
	if (! lvl)
	{
// always already filled; the data is in 'm_data'
		m_is_mipmap_level_filled[lvl] = 1;
		return;
	}
	lvl_previous = lvl - 1;

	if (! m_is_mipmap_level_filled[lvl_previous])
		Mipmap_Level_Fill (lvl_previous);
	Assert(m_is_mipmap_level_filled[lvl_previous]);

	level_size = m_mipmap_level_size[lvl];
	levels_min = m_data_mipmap_min[lvl];
	levels_avg = m_data_mipmap_avg[lvl];
	levels_max = m_data_mipmap_max[lvl];
	Assert(levels_min);
	Assert(levels_avg);
	Assert(levels_max);
	for (i = 0; i < level_size; i++)
	{
//		lmax = 0;
//		lavg = 0;
//		lmin = 0;
		result = Mipmap_Level_GetSegmentAverage (lvl_previous, i << WAVE_MIPMAP_LEVEL_INCREMENT_BITS, WAVE_MIPMAP_LEVEL_INCREMENT, lmin, lavg, lmax);
		Assert(result);
		levels_min[i] = lmin;
		levels_avg[i] = lavg;
		levels_max[i] = lmax;
	}
	m_is_mipmap_level_filled[lvl] = 1;
}

// does not automatically fills
// return 0, if is out of range
bool Wave::Mipmap_Level_GetSegmentAverage (int mipmap_lvl, uint64 offset, uint64 segment_size, wave_data_t& level_min, wave_data_t& level_avg, wave_data_t& level_max)
{
	uint64 level_size;
	wave_data_t lmin;
	wave_data_t lmax;
	int64 lavg;
	wave_data_t* levels_min;
	wave_data_t* levels_avg;
	wave_data_t* levels_max;
	int64 lvl;
	uint64 segment_end;
// need without sign, otherwise will be unsigned division
	int64 samples_num;

	Assert(segment_size);
// automatically fill it if necessary
//	if (! m_is_mipmap_level_filled[mipmap_lvl])
//		Mipmap_Level_Fill (mipmap_lvl);
	Assert(m_is_mipmap_level_filled[mipmap_lvl]);
	level_size = m_mipmap_level_size[mipmap_lvl];
	if (offset >= level_size)
		return 0;
	segment_end = offset + segment_size;
	if (segment_end > level_size)
		segment_end = level_size;
	samples_num = segment_end - offset;
	if (mipmap_lvl)
	{
		levels_min = m_data_mipmap_min[mipmap_lvl];
		levels_avg = m_data_mipmap_avg[mipmap_lvl];
		levels_max = m_data_mipmap_max[mipmap_lvl];
	}
	else
	{
		levels_min = m_data;
		levels_avg = m_data;
		levels_max = m_data;
	}
	Assert(levels_min);
	Assert(levels_avg);
	Assert(levels_max);
	lmin = levels_min[offset];
	lavg = levels_avg[offset];
	lmax = levels_max[offset];
	offset++;

	while (offset < segment_end)
	{
		lvl = levels_min[offset];
		if (lmin > lvl)
			lmin = lvl;
		lvl = levels_max[offset];
		if (lmax < lvl)
			lmax = lvl;
		lavg += levels_avg[offset];
		offset++;
	}
	level_min = lmin;
	level_max = lmax;
	level_avg = lavg / samples_num;
	return 1;
}

// maybe even is not necessary
//void Wave::Mipmap_Levels_Allocate ()
void Wave::Mipmap_Allocate ()
{
//	int i;

//	for (i = 0; i < WAVE_MIPMAP_LEVELS_NUM; i++)
//	{
//		Mipmap_Level_Allocate (i);
//	}
	Mipmap_Level_Allocate (WAVE_MIPMAP_LEVELS_NUM - 1);
}

//void Wave::Mipmap_Levels_Free ()
void Wave::Mipmap_Free ()
{
//	int i;

//	for (i = 0; i < WAVE_MIPMAP_LEVELS_NUM; i++)
//	{
//		Mipmap_Level_Free (i);
//	}
	Mipmap_Level_Free (WAVE_MIPMAP_LEVELS_NUM - 1);
}

void Wave::Mipmap_GetLevel (int& lvl, uint64 segment_size)
{
	lvl = 0;
	while (segment_size >>= WAVE_MIPMAP_LEVEL_INCREMENT_BITS)
		lvl++;
}

// --- mipmap ---

fix64 Wave::GetTime ()
{
	fix64 time;
	time = INT_TO_FIX64(m_samples_num);
	Assert(m_sample_rate);
	time /= m_sample_rate;
	return time;
}

// similar to 'SetSegmentDuration ()'
void Wave::CalculateSubParameters ()
{
	wxASSERT(m_sample_rate);
//	if (! m_sample_rate)
//		return 0;
	m_sample_time = F1_0;
	m_sample_time /= m_sample_rate;
}

// +++ blend view +++

bool Wave::IsDensityMoreThan1 (fix64 pixel_time)
{
	bool need_blend;

	CalculateSubParameters ();
	need_blend = pixel_time > m_sample_time;

	return need_blend;
}

bool Wave::IsDensityMoreThan2 (fix64 pixel_time)
{
	bool need_blend;

	CalculateSubParameters ();
	need_blend = pixel_time > m_sample_time * 2;

	return need_blend;
}

// using this, when drawing more detailed; less than one sample per pixel
// must make 'sample_moment' value in any case
bool Wave::GetSampleMoment (uint64 sample_ix, fix64* sample_moment)
{
	fix64 moment;

	moment = sample_ix * m_sample_time;
	moment += m_time_offset;
	*sample_moment = moment;
	if (sample_ix >= m_samples_num)
		return 0;
	return 1;
}

// backward operation; is necessary when need to get a point where to start to draw an update region
bool Wave::GetSampleIx (uint64& sample_ix, fix64 sample_moment)
{
	fix64 moment;

	Assert(m_sample_time);
	sample_ix = 0;
	moment = sample_moment - m_time_offset;
	if (moment < 0)
		return 0;
	sample_ix = moment / m_sample_time;
	return 1;
}

// similar to 'CalculateSubParameters ()'
// atm is used for the blend subroutines
void Wave::SetSegmentDuration (fix64 segment_duration)
{
	m_segment_duration = segment_duration;
#if 0
	m_segment_samples_num = segment_duration -1;
	m_segment_samples_num *= m_sample_rate;
	m_segment_samples_num += 1;
#else
	m_segment_samples_num = segment_duration * m_sample_rate;
#endif
}

// very simplified; no any blend
bool Wave::GetBlendForMoment_Simple (fix64 moment, wave_data_t& lvl_min, wave_data_t& lvl_avg, wave_data_t& lvl_max)
{
	uint64 sample_ix;
	wave_data_t level;

	if (moment + m_segment_duration < m_time_offset)
// not began yet
		return 0;

// 64 bits are not enough :)
	moment *= m_sample_rate;
	sample_ix = FIX64_TO_INT(moment);
	if (sample_ix >= m_samples_num)
		return 0;
	level = m_data[sample_ix];
	if (level >= 0)
	{
		lvl_min = 0;
		lvl_max = level;
	}
	else
	{
		lvl_min = level;
		lvl_max = 0;
	}
	lvl_avg = 0;

	return 1;
}

bool Wave::GetBlendForMoment_Mipmap (fix64 moment, wave_data_t& lvl_min, wave_data_t& lvl_avg, wave_data_t& lvl_max)
{
	int lvl;
	int samples_num;
	bool result;
	int64 sample_ix;
	int64 sample_last;
	uint64 segment_size;

	samples_num = FIX64_TO_INT(m_segment_samples_num);
// get level, from which will be necessary to take at least 4 samples to draw a position
	Mipmap_GetLevel (lvl, samples_num / 4);
	moment *= m_sample_rate;
	sample_ix = FIX64_TO_INT(moment);
	sample_last = FIX64_TO_INT(moment + m_segment_samples_num);
	segment_size = sample_last - sample_ix;
	if (! m_is_mipmap_level_filled[lvl])
		Mipmap_Level_Fill (lvl);
	sample_ix >>= WAVE_MIPMAP_LEVEL_INCREMENT_BITS * lvl;
	segment_size >>= WAVE_MIPMAP_LEVEL_INCREMENT_BITS * lvl;
	segment_size++;
	result = Mipmap_Level_GetSegmentAverage (lvl, sample_ix, segment_size, lvl_min, lvl_avg, lvl_max);
	return result;
}

bool Wave::GetBlendForMoment_Full (fix64 moment, wave_data_t& lvl_min, wave_data_t& lvl_avg, wave_data_t& lvl_max)
{
//	int64 sample_first;
	int64 sample_ix;
	int64 sample_last;
	uint64 ix;
	uint64 ix_last;
	wave_data_t level;
	wave_data_t lmin;
//	wave_data_t lavg;
	int64 lavg;
	wave_data_t lmax;
	int samples_num;

	if (moment + m_segment_duration < m_time_offset)
// not began yet
		return 0;

// 64 bits are not enough :)
	moment *= m_sample_rate;
	sample_ix = FIX64_TO_INT(moment);
//	segment_duration *= m_sample_rate;
//	sample_last = FIX64_TO_INT(moment + segment_duration);
	sample_last = FIX64_TO_INT(moment + m_segment_samples_num);
// if nothing to show
	if ((uint64)sample_ix >= m_samples_num)
		return 0;
	if (sample_last < 0)
		return 0;
	Assert(sample_last >= sample_ix);
	samples_num = sample_last - sample_ix + 1;

	lavg = m_data[sample_ix];
	lmin = lavg;
	lmax = lavg;
	sample_ix++;
#if 0
// slow; contains multiplication
	while (sample_ix <= sample_last)
	{
		level = m_data[sample_ix];
		if (lmin > level)
			lmin = level;
		if (lmax < level)
			lmax = level;
		lavg += level;
		sample_ix++;
	}
#else
// faster
// I would say - optimizing is finished; now drags where it must drag - on noncachable data access
	ix = sample_ix * WAVE_DATA_SIZEOF;
	ix_last = sample_last * WAVE_DATA_SIZEOF;
	while (ix <= ix_last)
	{
		level = *(wave_data_t*)&(((byte*)m_data)[ix]);
		if (lmin > level)
			lmin = level;
		if (lmax < level)
			lmax = level;
		lavg += level;
		ix += WAVE_DATA_SIZEOF;
	}
#endif
	lvl_min = lmin;
	lvl_avg = lavg / samples_num;
	lvl_max = lmax;

	return 1;
}

// mipmap, but ends must not contain partially taken values,
// all included values must be fully taken, I am talking about first and last samples,
// which mostly probable will be included in nearest pixels also, and also - must be included there fully.
// that is necessary so a maximal level would be always visible.
// returns 0, if levels are nothing to define from
// average value is an ariphmetical average of all samples in the segment
bool Wave::GetBlendForMoment (fix64 moment, wave_data_t& lvl_min, wave_data_t& lvl_avg, wave_data_t& lvl_max)
{
	bool result;

	Assert(moment >= 0);
//	Assert(segment_duration >= 0);
	Assert(m_segment_duration >= 0);
//	Assert(lvl_min);
//	Assert(lvl_avg);
//	Assert(lvl_max);
	switch (m_view_blend_mode)
	{
	case WAVE_VIEW__SIMPLE:
		result = GetBlendForMoment_Simple (moment, lvl_min, lvl_avg, lvl_max);
		break;
	case WAVE_VIEW__MIPMAP:
		result = GetBlendForMoment_Mipmap (moment, lvl_min, lvl_avg, lvl_max);
		break;
	case WAVE_VIEW__FULL:
		result = GetBlendForMoment_Full (moment, lvl_min, lvl_avg, lvl_max);
		break;
	default:
		Error ("Wave::GetBlendForMoment () :  wave view mode %d is unknown", m_view_blend_mode);
	}

	return result;
}

// --- blend view ---

