#pragma once

//#include "types.h"
#include "fix.h"


// always signed (low level AC97 driver also work with signed)
// constantly 32bits (at least), signed
// careful, less than 32 bit may cause "signed comparison error" with MSVC compilers
// cannot be rised to 64 bits easy (if will become necessary); all temporary values in subroutines then should rise to more than 64 bit
typedef int wave_data_t;
//typedef unsigned wave_data_t wave_data_u_t;
typedef uint wave_data_u_t;
// if someone needs this
//typedef __int64 wave_data_t;
#define WAVE_DATA_SIZEOF (sizeof(wave_data_t))
#define WAVE_DATA_BITS (WAVE_DATA_SIZEOF << 3)
#define WAVE_HIGHEST_BIT_MASK (1 << (WAVE_DATA_BITS - 1))
#define WAVE_LEVEL_MIN ((wave_data_t)WAVE_HIGHEST_BIT_MASK)
#define WAVE_LEVEL_MAX ((wave_data_t)(WAVE_HIGHEST_BIT_MASK - 1))
// how much values will go in the next mipmap level
#define WAVE_MIPMAP_LEVEL_INCREMENT_BITS 8
#define WAVE_MIPMAP_LEVEL_INCREMENT (1 << WAVE_MIPMAP_LEVEL_INCREMENT_BITS)
// first level will not be used; it is the original
// data pointer is 64 bit; hope this will be ok in the nearest feature
#define WAVE_MIPMAP_LEVELS_NUM (64 / WAVE_MIPMAP_LEVEL_INCREMENT_BITS)

enum
{
	WAVE_RESAMPLE_TYPE__NONE,
	WAVE_RESAMPLE_TYPE__DUPLICATE,
	WAVE_RESAMPLE_TYPE__AVERAGE,
	WAVE_RESAMPLE_TYPE__BENDS_PLUS_ONE,
	WAVE_RESAMPLE_TYPE__BENDS_CONTINUOUS,
	WAVE_RESAMPLE_TYPE__POWER_SOME,
	WAVE_RESAMPLE_TYPES_NUM
};

enum
{
	WAVE_VIEW__NONE,
	WAVE_VIEW__SIMPLE,
	WAVE_VIEW__MIPMAP,
	WAVE_VIEW__FULL,
	WAVE_VIEWS_NUM
};


// individual wave
class Wave
{
public:
// processing linkage
	int m_link_previous;
	int m_link_next;
	int m_references_count;

//	void* data;
	wave_data_t* m_data;
// information about data
	uint64 m_samples_num;
// +++ analyze +++
	wave_data_t m_level_min;
	wave_data_t m_level_max;
// --- analyze ---

// + external +
//// + just storage +
// for the such intended waves creating/removing system, the best place for this - here
// index in the project; not in a file
	int m_ix;
	int m_bits_per_sample;
	uint m_sample_rate;
	fix64 m_sample_time;
// it may begin later, some time after the other waves begin
	fix64 m_time_offset;
//// - just storage -
	uint64 m_segment_samples_num;
// - external -
	int m_view_blend_mode;

	Wave ();
	~Wave ();
	void Reference ();
	void Dereference ();
	bool SetSize (uint64 samples_num);
private:
	bool ForProcessing_SetSize (uint64 samples_num);
	void ForProcessing_ReplaceOld ();
public:
// +++ analyze +++
	void Analyze_Level ();
// --- analyze ---
// +++ process +++
	void Process_Samples_Half (bool average_level, bool downrate);
private:
	void CalculateAverageWithCorrection_PlusOne (int64& lvl, wave_data_t& lvl0, wave_data_t& lvl1, wave_data_t& lvl2, wave_data_t& lvl3);
	void Process_Samples_Double_Continuous ();
	void Process_Samples_Double_Power_Some (int spline_degree);
public:
	void Process_Samples_Double (int resample_type, bool uprate, bool extend_begin, bool extend_end);
	void Process_Resample_Up (uint cycle_size_out, uint cycle_size_in, bool change_rate);
	void Process_Resample_Down (uint cycle_size_out, uint cycle_size_in, bool change_rate);
	void Process_Resample (uint cycle_size_out, uint cycle_size_in, bool change_rate);
	void Samples_Insert (uint64 offset, uint64 amount);
	void Samples_Remove (uint64 offset, uint64 amount);
	void Process_Amplify (uint quant_multiply, uint quant_divide);
// --- process ---
// +++ process 2 +++
	void FillRemainderWithZeroes (uint64 from);
	void GenerateSine (double d_amplitude, double d_level_offset, double d_frequency, double d_phase_beginning);
	void Filter (Wave* wave_low, Wave* wave_high, double d_constance);
// --- process 2 ---
// +++ mipmap +++
//	void Mipmap_InvalidateLevels ();
	void Mipmap_Level_Allocate (int lvl);
	void Mipmap_Level_Free (int lvl);
	void Mipmap_Level_Fill (int lvl);
	bool Mipmap_Level_GetSegmentAverage (int mipmap_lvl, uint64 offset, uint64 segment_size, wave_data_t& level_min, wave_data_t& level_avg, wave_data_t& level_max);
//	void Mipmap_Levels_Allocate ();
//	void Mipmap_Levels_Free ();
	void Mipmap_Allocate ();
	void Mipmap_Free ();
	void Mipmap_GetLevel (int& lvl, uint64 segment_size);
// --- mipmap ---
#if 0
	uint64 GetSize ()
	{
		return m_samples_num;
	}
#endif
	fix64 GetTime ();
	void CalculateSubParameters ();
	inline void CheckLevel (int64 lvl)
	{
		if (lvl < WAVE_LEVEL_MIN)
		{
			lvl = WAVE_LEVEL_MIN;
			m_was_level_cuts = 1;
		}
		else
		if (lvl > WAVE_LEVEL_MAX)
		{
			lvl = WAVE_LEVEL_MAX;
			m_was_level_cuts = 1;
		}
	}
// + blend view +
//	bool DoesNeedBlend (fix64 pixel_time);
	bool IsDensityMoreThan1 (fix64 pixel_time);
	bool IsDensityMoreThan2 (fix64 pixel_time);
	bool GetSampleMoment (uint64 sample_ix, fix64* moment);
	bool GetSampleIx (uint64& sample_ix, fix64 sample_moment);
// for getting blend or mipmap
	void SetSegmentDuration (fix64 segment_duration);
	bool GetBlendForMoment_Simple (fix64 moment, wave_data_t& lvl_min, wave_data_t& lvl_avg, wave_data_t& lvl_max);
	bool GetBlendForMoment_Mipmap (fix64 moment, wave_data_t& lvl_min, wave_data_t& lvl_avg, wave_data_t& lvl_max);
	bool GetBlendForMoment_Full (fix64 moment, wave_data_t& lvl_min, wave_data_t& lvl_avg, wave_data_t& lvl_max);
	bool GetBlendForMoment (fix64 moment, wave_data_t& lvl_min, wave_data_t& lvl_avg, wave_data_t& lvl_max);
// - blend view -

private:
// temp store
	wave_data_t* m_data__in_process;
	uint64 m_samples_num__in_process;
	bool m_was_level_cuts;
	fix64 m_segment_duration;
// + mipmap +
//	bool m_is_mipmap_level_valid[WAVE_MIPMAP_LEVELS_NUM];
	bool m_is_mipmap_level_allocated[WAVE_MIPMAP_LEVELS_NUM];
	bool m_is_mipmap_level_filled[WAVE_MIPMAP_LEVELS_NUM];
// ix 0 will not be used on these three
	wave_data_t* m_data_mipmap_min[WAVE_MIPMAP_LEVELS_NUM];
	wave_data_t* m_data_mipmap_avg[WAVE_MIPMAP_LEVELS_NUM];
	wave_data_t* m_data_mipmap_max[WAVE_MIPMAP_LEVELS_NUM];
	uint64 m_mipmap_level_size[WAVE_MIPMAP_LEVELS_NUM];
// - mipmap -
};

