// by Alexander Ilyin
// abstract splitter class; belongs at NanoWave project
// initialized in this way :  1) new WindowSplitter, 2) Windows_Set (), 3) Config_Read (), 4) Values_Apply ();
// and on exit need only 'Config_Write ()'

#include "window_splitter.h"
#include "string.h"
#include "error.h"


BEGIN_EVENT_TABLE(WindowSplitter, wxSplitterWindow)
//	EVT_SPLITTER_SASH_POS_CHANGED(wxID_ANY, MySplitterWindow::OnPositionChanged)
//	EVT_SPLITTER_SASH_POS_CHANGING(wxID_ANY, MySplitterWindow::OnPositionChanging)
//	EVT_SPLITTER_DCLICK(wxID_ANY, MySplitterWindow::OnDClick)
//	EVT_SPLITTER_UNSPLIT(wxID_ANY, MySplitterWindow::OnUnsplitEvent)
	EVT_SIZE(WindowSplitter::OnSize)
END_EVENT_TABLE()


//WindowSplitter::WindowSplitter (wxFrame* parent)
WindowSplitter::WindowSplitter (wxWindow* parent, char* name)
	:
	wxSplitterWindow (parent, wxID_ANY, wxDefaultPosition, wxDefaultSize, wxSP_3D | wxSP_LIVE_UPDATE | wxCLIP_CHILDREN)
{
//	m_frame = parent;
	m_window_parent = parent;
	Name_Set (name);

//	m_vertically = 0;
	m_split_direction = 0;
	m_is_split = 0;
	m_active_window = 0;
	m_from_end = 0;
	m_splitter_pos = 100;
}

void WindowSplitter::Name_Set (char* name)
{
	bool result;

//	strcpy (m_name, "splitter_");
//	result = Strncat (m_name, name, WINDOW_SPLITTER_NAME_SIZE);
//	if (! result)
//		Error ("WindowSplitter::Name_Set () :  string does not fit into buffer");
//	result = Str_IsAlphaNum (m_name);
	m_name = "splitter_";
	m_name += name;
	result = Str_IsAlphaNum (m_name.c_str ());
	if (! result)
		Error ("WindowSplitter::Name_Set () :  string is not alphanum");
}

// set, but do not apply
// default settings goes here
void WindowSplitter::Windows_Set (wxWindow* wnd0, wxWindow* wnd1, byte split_direction, bool is_split, bool from_end, uint offset, byte active_window)
{
	m_window_0 = wnd0;
	m_window_1 = wnd1;
	m_split_direction = split_direction;
	m_is_split = is_split;
	m_from_end = from_end;
	m_splitter_pos = offset;
	m_active_window = active_window;
}

// *uses relative path
void WindowSplitter::Config_Read (wxConfigBase* config)
{
// absolute
//	config->SetPath (_T("/"));
// relative
	config->SetPath (m_name);
	m_split_direction = !! config->Read (_T("split_direction"), (long)m_split_direction);
	m_is_split = !! config->Read (_T("is_split"), (long)m_is_split);
	m_active_window = !! config->Read (_T("active_window"), (long)m_active_window);
	m_from_end = !! config->Read (_T("from_end"), (long)m_from_end);
	m_splitter_pos = config->Read (_T("splitter_pos"), (long)m_splitter_pos);
// go back, that would be possible to request again
	config->SetPath (_T("../"));
}

// *uses relative path
void WindowSplitter::Config_Write (wxConfigBase* config)
{
	config->SetPath (m_name);
	config->Write (_T("split_direction"), (long)m_split_direction);
	config->Write (_T("is_split"), (long)m_is_split);
	config->Write (_T("active_window"), (long)m_active_window);
	config->Write (_T("from_end"), (long)m_from_end);
	config->Write (_T("splitter_pos"), (long)m_splitter_pos);
	config->SetPath (_T("../"));
}

#if 0
// split direction is not touched here; it is defined on a higher level
void WindowSplitter::Values_Get ()
{
	wxWindow* wnd;

	m_is_split = IsSplit ();
	if (m_is_split)
		goto proceed_split;
// get active window; (it is a bit hacky, that in non-split mode the active window is window 1)
	wnd = GetWindow1 ();
	m_active_window = 0;
	return;
proceed_split:
	m_from_end = 0;
	m_splitter_pos = 100;
}
#endif

//void WindowSplitter::Values_Set ()
void WindowSplitter::Values_Apply ()
{
// prevent easy occasional unsplit
	SetMinimumPaneSize (20);

	if (m_is_split)
	{
		int sash_position;

		sash_position = m_splitter_pos;
		if (m_from_end)
			sash_position = -sash_position;
		if (m_split_direction)
			SplitVertically (m_window_0, m_window_1, sash_position);
		else
			SplitHorizontally (m_window_0, m_window_1, sash_position);
//		SetSashPosition (sash_position);
	}
	else
	{
		wxWindow* wnd;

		wnd = m_active_window ? m_window_1 : m_window_0;
		Initialize (wnd);
	}
//	Update ();
}

void WindowSplitter::NewSashPosition (int pos)
{
	Assert(m_is_split);
	if (m_from_end)
	{
		int window_size;
// *gets appropriate size of the client area
		window_size = wxSplitterWindow::GetWindowSize ();
		Assert(window_size >= pos);
// in 'wxSplitterWindow::ConvertSashPosition' for negative value is 'GetWindowSize() + sashPosition'
		pos = window_size - pos;
	}
	m_splitter_pos = pos;
}

// hook the sash position
// *not always requested !
bool WindowSplitter::OnSashPositionChange (int newSashPosition)
{
	bool result;

	result = wxSplitterWindow::OnSashPositionChange (newSashPosition);
	if (result)
		NewSashPosition (newSashPosition);

	return result;
}

// hook, in case if sash will have offset from end
void WindowSplitter::OnSize (wxSizeEvent& event)
{
	event.Skip ();
	if (m_from_end)
	{
		int window_size;
		int pos;
// *gets appropriate size of the client area
		window_size = wxSplitterWindow::GetWindowSize ();
// convert to forward value
		pos = window_size - m_splitter_pos;
// hmmm, maybe not 0 ...
		if (pos < 0)
			pos = 0;
		SetSashPosition (pos);
// convert back to backward value
// in 'wxSplitterWindow::ConvertSashPosition' for negative value is 'GetWindowSize() + sashPosition'
		pos = window_size - pos;
// although possible to not store modified value, but that will require additional researches
		m_splitter_pos = pos;
	}
}

